#include <Python.h>
#include <structmember.h>
#include <lasso/lasso.h>
#include <config.h>
#include "../ghashtable.h"
#include "../../lasso/debug.h"
#include "../../lasso/utils.h"
#include "../utils.c"

#if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
typedef int Py_ssize_t;
#define PY_SSIZE_T_MAX INT_MAX
#define PY_SSIZE_T_MIN INT_MIN
#endif

GQuark lasso_wrapper_key;

PyMODINIT_FUNC init_lasso(void);
G_GNUC_UNUSED static PyObject* get_pystring_from_xml_node(xmlNode *xmlnode);
G_GNUC_UNUSED static xmlNode*  get_xml_node_from_pystring(PyObject *string);
G_GNUC_UNUSED static PyObject* get_dict_from_hashtable_of_objects(GHashTable *value);
G_GNUC_UNUSED static PyObject* get_dict_from_hashtable_of_strings(GHashTable *value);
G_GNUC_UNUSED static PyObject* PyGObjectPtr_New(GObject *obj);
G_GNUC_UNUSED static void set_hashtable_of_pygobject(GHashTable *a_hash, PyObject *dict);
G_GNUC_UNUSED static void set_hashtable_of_strings(GHashTable *a_hash, PyObject *dict);
G_GNUC_UNUSED static void set_list_of_strings(GList **a_list, PyObject *seq);
G_GNUC_UNUSED static void set_list_of_xml_nodes(GList **a_list, PyObject *seq);
G_GNUC_UNUSED static void set_list_of_pygobject(GList **a_list, PyObject *seq);
G_GNUC_UNUSED static PyObject *get_list_of_strings(const GList *a_list);
G_GNUC_UNUSED static PyObject *get_list_of_xml_nodes(const GList *a_list);
G_GNUC_UNUSED static PyObject *get_list_of_pygobject(const GList *a_list);
G_GNUC_UNUSED static gboolean valid_seq(PyObject *seq);
G_GNUC_UNUSED static void free_list(GList **a_list, GFunc free_help);
G_GNUC_UNUSED static time_t* get_time_t(PyObject *time);

typedef struct {
	PyObject_HEAD
	GObject *obj;
	PyObject *typename;
} PyGObjectPtr;
static PyTypeObject PyGObjectPtrType;

/* utility functions */
static PyObject *
noneRef() {
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
get_dict_from_hashtable_of_objects(GHashTable *value)
{
	GList *keys, *begin;
	PyObject *dict,*proxy;
	GObject *item_value;
	PyObject *item;

	dict = PyDict_New();

	begin = keys = g_hash_table_get_keys(value);
	for (; keys; keys = g_list_next(keys)) {
		item_value = g_hash_table_lookup(value, keys->data);
		if (item_value) {
			item = PyGObjectPtr_New(G_OBJECT(item_value));
			PyDict_SetItemString(dict, (char*)keys->data, item);
			Py_DECREF(item);
		} else {
			PyErr_Warn(PyExc_RuntimeWarning, "hashtable contains a null value");
		}
	}
	g_list_free(begin);

	proxy = PyDictProxy_New(dict);
	Py_DECREF(dict);
	return proxy;
}

static PyObject*
get_dict_from_hashtable_of_strings(GHashTable *value)
{
	GList *keys, *begin;
	PyObject *dict,*proxy;
	char *item_value;
	PyObject *item;

	dict = PyDict_New();

	begin = keys = g_hash_table_get_keys(value);
	for (; keys; keys = g_list_next(keys)) {
		item_value = g_hash_table_lookup(value, keys->data);
		if (item_value) {
			item = PyString_FromString(item_value);
			PyDict_SetItemString(dict, (char*)keys->data, item);
			Py_DECREF(item);
		} else {
			PyErr_Warn(PyExc_RuntimeWarning, "hashtable contains a null value");
		}
	}
	g_list_free(begin);

	proxy = PyDictProxy_New(dict);
	Py_DECREF(dict);
	return proxy;
}

static xmlBuffer*
xmlnode_to_xmlbuffer(xmlNode *node)
{
	xmlOutputBufferPtr output_buffer;
	xmlBuffer *buffer;

	if (! node)
		return NULL;

	buffer = xmlBufferCreate();
	output_buffer = xmlOutputBufferCreateBuffer(buffer, NULL);
	xmlNodeDumpOutput(output_buffer, NULL, node, 0, 0, NULL);
	xmlOutputBufferClose(output_buffer);
	xmlBufferAdd(buffer, BAD_CAST "", 1);

	return buffer;
}

static PyObject*
get_pystring_from_xml_node(xmlNode *xmlnode)
{
	PyObject *pystring = NULL;
	xmlBuffer *buffer;

	if (xmlnode == NULL) {
		return NULL;
	}
	buffer = xmlnode_to_xmlbuffer(xmlnode);

	if (buffer == NULL) {
		pystring = NULL;
	} else {
		pystring = PyString_FromString((char*)xmlBufferContent(buffer));
		xmlBufferFree(buffer);
	}

	return pystring;
}

static gboolean
valid_seq(PyObject *seq) {
	if (! seq || ( seq != Py_None && ! PyTuple_Check(seq))) {
		 PyErr_SetString(PyExc_TypeError, "value should be tuple");
		 return 0;
	}
	return 1;
}

static void
free_list(GList **a_list, GFunc free_help) {
	if (*a_list) {
		g_list_foreach(*a_list, free_help, NULL);
		g_list_free(*a_list);
	}
}

/** Remove all elements from a_hash and replace them with
 * the key-values pairs from the python dict.
 * Increase reference of new values before removeing
 * values from the hash, so if there are somme common
 * values with RefCoun = 1 they won't be deallocated.
 * */
static void
set_hashtable_of_pygobject(GHashTable *a_hash, PyObject *dict) {
	PyObject *key, *value;
	Py_ssize_t i;

	if (! a_hash) {
		 PyErr_SetString(PyExc_TypeError, "hashtable does not exist");
		 return;
	}
	if (dict != Py_None && ! PyDict_Check(dict)) {
		 PyErr_SetString(PyExc_TypeError, "value should be a frozen dict");
		 return;
	}
	i = 0;
	// Increase ref count of common object between old and new
	// value of the hashtable
	while (PyDict_Next(dict, &i, &key, &value)) {
		if (! PyString_Check(key) || ! PyObject_TypeCheck(value, &PyGObjectPtrType))
		{
			PyErr_SetString(PyExc_TypeError,
					"value should be a dict, "
					"with string keys "
					"and GObjectPtr values");
			goto failure;
		}
		g_object_ref(((PyGObjectPtr*)value)->obj);
	}
	g_hash_table_remove_all (a_hash);
	i = 0;
	while (PyDict_Next(dict, &i, &key, &value)) {
		char *ckey = g_strdup(PyString_AsString(key));
		g_hash_table_replace (a_hash, ckey, ((PyGObjectPtr*)value)->obj);
	}
	return;
failure:
	i = 0;
	while (PyDict_Next(dict, &i, &key, &value)) {
		if (! PyString_Check(key) || ! PyObject_TypeCheck(value, &PyGObjectPtrType))
			break;
		g_object_unref((PyGObjectPtr*)value);
	}
}

static void 
set_hashtable_of_strings(GHashTable *a_hash, PyObject *dict)
{
	PyObject *key, *value;
	Py_ssize_t i;

	if (! a_hash) {
		 PyErr_SetString(PyExc_TypeError, "hashtable does not exist");
		 return;
	}
	if (dict != Py_None && ! PyDict_Check(dict)) {
		 PyErr_SetString(PyExc_TypeError, "value should be a frozen dict");
		 return;
	}
	i = 0;
	// Increase ref count of common object between old and new
	// value of the hashtable
	while (PyDict_Next(dict, &i, &key, &value)) {
		if (! PyString_Check(key) || ! PyString_Check(value))
		{
			PyErr_SetString(PyExc_TypeError,
					"value should be a dict, "
					"with string keys "
					"and string values");
			goto failure;
		}
	}
	g_hash_table_remove_all (a_hash);
	i = 0;
	while (PyDict_Next(dict, &i, &key, &value)) {
		char *ckey = PyString_AsString(key);
		char *cvalue = PyString_AsString(value);
		g_hash_table_insert (a_hash, g_strdup(ckey), g_strdup(cvalue));
	}
failure:
	return;
}

/** Set the GList* pointer, pointed by a_list, to a pointer on a new GList
 * created by converting the python seq into a GList of char*.
 */
static void
set_list_of_strings(GList **a_list, PyObject *seq) {
	GList *list = NULL;
	int l = 0,i;

	lasso_return_if_fail(valid_seq(seq));
	if (seq != Py_None) {
		l = PySequence_Length(seq);
	}
	for (i=0; i<l; i++) {
		PyObject *pystr = PySequence_Fast_GET_ITEM(seq, i);
		if (! PyString_Check(pystr)) {
			PyErr_SetString(PyExc_TypeError,
					"value should be a tuple of strings");
			goto failure;
		}
		list = g_list_append(list, g_strdup(PyString_AsString(pystr)));
	}
	free_list(a_list, (GFunc)g_free);
	*a_list = list;
	return;
failure:
	free_list(&list, (GFunc)g_free);
}

/** Set the GList* pointer, pointed by a_list, to a pointer on a new GList
 * created by converting the python seq into a GList of xmlNode*.
 */
static void
set_list_of_xml_nodes(GList **a_list, PyObject *seq) {
	GList *list = NULL;
	int l = 0,i;

	lasso_return_if_fail(valid_seq(seq));
	if (seq != Py_None) {
		l = PySequence_Length(seq);
	}
	for (i=0; i<l; i++) {
		PyObject *item = PySequence_Fast_GET_ITEM(seq, i);
		xmlNode *item_node;
		if (! PyString_Check(item)) {
			PyErr_SetString(PyExc_TypeError,
					"value should be a tuple of strings");
			goto failure;
		}
		item_node = get_xml_node_from_pystring(item);
		list = g_list_append(list, item_node);
	}
	free_list(a_list, (GFunc)xmlFreeNode);
	*a_list = list;
	return;
failure:
	free_list(&list, (GFunc)xmlFreeNode);
}

/** Set the GList* pointer, pointed by a_list, to a pointer on a new GList
 * created by converting the python seq into a GList of GObject*.
 */
static void
set_list_of_pygobject(GList **a_list, PyObject *seq) {
	GList *list = NULL;
	int l = 0,i;

	lasso_return_if_fail(valid_seq(seq));
	if (seq != Py_None) {
		l = PySequence_Length(seq);
	}
	for (i=0; i<l; i++) {
		PyObject *item = PySequence_Fast_GET_ITEM(seq, i);
		GObject *gobject;
		if (! PyObject_TypeCheck(item, &PyGObjectPtrType)) {
			PyErr_SetString(PyExc_TypeError,
					"value should be a tuple of PyGobject");
			goto failure;
		}
		gobject = g_object_ref(((PyGObjectPtr*)item)->obj);
		list = g_list_append(list, gobject);
	}
	free_list(a_list, (GFunc)g_object_unref);
	*a_list = list;
	return;
failure:
	free_list(&list, (GFunc)g_object_unref);
}

static xmlNode*
get_xml_node_from_pystring(PyObject *string) {
	return lasso_string_fragment_to_xmlnode(PyString_AsString(string),
			PyString_Size(string));
}

/** Return a tuple containing the string contained in a_list */
static PyObject *
get_list_of_strings(const GList *a_list) {
	PyObject *a_tuple = NULL;
	int i = 0;

	/* Cast because g_list_length does not take const but is a const function */
	a_tuple = PyTuple_New(g_list_length((GList*)a_list));
	if (! a_tuple)
		goto failure;
	if (! a_list) {
		return a_tuple;
	}
	while (a_list) {
		if (a_list->data) {
			PyObject *str = PyString_FromString((const char*)a_list->data);
			if (!str) {
				goto failure;
			}
			PyTuple_SetItem(a_tuple, i, str);
			i++;
		} else {
			PyErr_Warn(PyExc_RuntimeWarning,
				"list contains a NULL value");
		}
		a_list = a_list->next;
	}
	if (_PyTuple_Resize(&a_tuple, i))
		goto failure;
	return a_tuple;
failure:
	PyErr_SetString(PyExc_TypeError, "Allocation problem in get_list_of_strings");
	Py_XDECREF(a_tuple);
	return noneRef();
}

static PyObject *
get_list_of_xml_nodes(const GList *a_list) {
	PyObject *a_tuple = NULL;
	int i = 0;

	/* Cast because g_list_length does not take const but is a const function */
	a_tuple = PyTuple_New(g_list_length((GList*)a_list));
	if (! a_tuple)
		goto failure;
	if (! a_list) {
		return a_tuple;
	}
	while (a_list) {
		if (a_list->data) {
			PyObject *str = get_pystring_from_xml_node((xmlNode*)a_list->data);
			if (str) {
				PyTuple_SetItem(a_tuple, i, str);
				i++;
			} else {
				PyErr_Warn(PyExc_RuntimeWarning,
					"could not convert an xmlNode to a string");
			}
		} else {
			PyErr_Warn(PyExc_RuntimeWarning,
				"list contains a NULL value");
		}
		a_list = a_list->next;
	}
	if (_PyTuple_Resize(&a_tuple, i))
		goto failure;
	return a_tuple;
failure:
	PyErr_SetString(PyExc_TypeError, "Allocation problem in get_list_of_strings");
	Py_XDECREF(a_tuple);
	return noneRef();
}

static PyObject *
get_list_of_pygobject(const GList *a_list) {
	PyObject *a_tuple = NULL;
	int i = 0;

	/* Cast because g_list_length does not take const but is a const function */
	a_tuple = PyTuple_New(g_list_length((GList*)a_list));
	if (! a_tuple)
		goto failure;
	if (! a_list) {
		return a_tuple;
	}
	while (a_list) {
		if (a_list->data) {
			PyObject *pygobject;
			pygobject = PyGObjectPtr_New((GObject*)a_list->data);
			if (pygobject) {
				PyTuple_SetItem(a_tuple, i, pygobject);
				i++;
			} else {
				PyErr_Warn(PyExc_RuntimeWarning,
					"could not convert a GObject to a PyGobject");
			}
		} else {
			PyErr_Warn(PyExc_RuntimeWarning,
				"list contains a NULL value");
		}
		a_list = a_list->next;
	}
	if (_PyTuple_Resize(&a_tuple, i))
		goto failure;
	return a_tuple;
failure:
	PyErr_SetString(PyExc_TypeError, "Allocation problem in get_list_of_strings");
	Py_XDECREF(a_tuple);
	return noneRef();
}

/**
 * get_time_t:
 * @time: a #PyInt
 *
 * Convert a python integer object to a time_t value, considering it is a unsigned 32 bit integer
 * value.
 *
 * Return: a time_t* value if time is a python integer, NULL otherwise.
 */
static time_t*
get_time_t(PyObject *time)
{
	if (time != Py_None && PyInt_Check(time)) {
		time_t *val = malloc(sizeof(time_t));

		*val = (time_t)PyInt_AS_LONG(time);
		return val;
	}
	return NULL;
}

/* wrapper around GObject */
static void
PyGObjectPtr_dealloc(PyGObjectPtr *self)
{
	if (lasso_flag_memory_debug) {
		fprintf(stderr, "dealloc (%p ptr to %p (type:%s, rc:%d))\n",
				self, self->obj,
				G_OBJECT_TYPE_NAME(self->obj),
				self->obj->ref_count);
	}
	g_object_set_qdata_full(self->obj, lasso_wrapper_key, NULL, NULL);
	g_object_unref(self->obj);
	Py_XDECREF(self->typename);
	self->ob_type->tp_free((PyObject*)self);
}

static int
startswith(const char *string, const char *prefix)
{
    return strncmp(string, prefix, strlen(prefix)) == 0;
}

static PyObject*
PyGObjectPtr_New(GObject *obj)
{
	PyGObjectPtr *self;

	if (obj == NULL) {
		return noneRef();
	}

	self = (PyGObjectPtr*)g_object_get_qdata(obj, lasso_wrapper_key);
	if (self != NULL) {
		Py_INCREF(self);
	} else {
        const char *typename;

		self = (PyGObjectPtr*)PyObject_NEW(PyGObjectPtr, &PyGObjectPtrType);
		g_object_set_qdata_full(obj, lasso_wrapper_key, self, NULL);
		self->obj = g_object_ref(obj);
        typename = G_OBJECT_TYPE_NAME(obj);
        /* XXX: Fixme !!!!! */
        if (startswith(typename, "LassoDgme")) {
    		self->typename = PyString_FromString(typename+9);
        } else if (startswith(typename, "Lasso")) {
    		self->typename = PyString_FromString(typename+5);
        } else {
            self->typename = PyString_FromString(typename);
        }
	}
	return (PyObject*)self;
}

static PyObject *
PyGObjectPtr_repr(PyGObjectPtr *obj)
{
	return PyString_FromFormat("<PyGObjectPtr to %p (type: %s, refcount: %d)>",
			obj->obj,
			G_OBJECT_TYPE_NAME(obj->obj),
			obj->obj->ref_count);
}

static PyMemberDef PyGObjectPtr_members[] = {
	{"typename", T_OBJECT, offsetof(PyGObjectPtr, typename), 0, "typename"},
	{NULL, 0, 0, 0, NULL}
};

static PyObject*
PyGObjectPtr_get_refcount(PyGObjectPtr *self, G_GNUC_UNUSED void *closure)
{
	PyObject *refcount;

	refcount = PyInt_FromLong(self->obj->ref_count);
	Py_INCREF(refcount);
	return refcount;
}

static PyGetSetDef PyGObjectPtr_getseters[] = {
	{"refcount", (getter)PyGObjectPtr_get_refcount, NULL,
		"reference count of intern GObject*", NULL},
	{NULL, NULL, NULL, NULL, NULL}  /* Sentinel */
};


static PyTypeObject PyGObjectPtrType = {
	PyObject_HEAD_INIT(NULL)
	0, /* ob_size */
	"_lasso.PyGObjectPtr", /* tp_name */
	sizeof(PyGObjectPtr),  /* tp_basicsize */
	0,                     /* tp_itemsize */
	(destructor)PyGObjectPtr_dealloc, /* tp_dealloc */
	0,       /*tp_print*/
	0,                      /*tp_getattr*/
	.tp_setattr = 0,       /*tp_setattr*/
	0,       /*tp_compare*/
	(reprfunc)PyGObjectPtr_repr,       /*tp_repr*/
	0,       /*tp_as_number*/
	0,       /*tp_as_sequence*/
	0,       /*tp_as_mapping*/
	0,       /*tp_hash */
	0,       /*tp_call*/
	0,       /*tp_str*/
	0,       /*tp_getattro*/
	0,       /*tp_setattro*/
	0,       /*tp_as_buffer*/
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,  /*tp_flags*/
	"PyGObjectPtr objects",   /* tp_doc */
	0,       /* tp_traverse */
	0,       /* tp_clear */
	0,       /* tp_richcompare */
	0,       /* tp_weaklistoffset */
	0,       /* tp_iter */
	0,       /* tp_iternext */
	0,       /* tp_methods */
	PyGObjectPtr_members,   /* tp_members */
	PyGObjectPtr_getseters, /* tp_getset */
	NULL,
	NULL
};

static void
set_object_field(GObject **a_gobject_ptr, PyGObjectPtr *a_pygobject) {
	if (*a_gobject_ptr) {
		g_object_unref(*a_gobject_ptr);
	}
	if ((PyObject*)a_pygobject == Py_None) {
		*a_gobject_ptr = NULL;
	} else {
		*a_gobject_ptr = g_object_ref(a_pygobject->obj);
	}
}


static PyObject *get_logger_object() {
	static PyObject *_logger_object = NULL;

	PyObject *logging_module = PyImport_ImportModule("lasso");

	if (logging_module) {
		_logger_object = PyObject_GetAttrString(logging_module, "logger");
		Py_DECREF(logging_module);
		if (_logger_object)
			goto exit;
	}
	/* XXX: needed so that PyImport_ImportModule("logging") always works */
	logging_module = PyImport_ImportModule("sys");
	if (logging_module) {
		Py_DECREF(logging_module);
	}
	logging_module = PyImport_ImportModule("logging");
	if (logging_module) {
		_logger_object = PyObject_CallMethod(logging_module, "getLogger",
				"s#", "lasso", sizeof("lasso")-1);
		Py_DECREF(logging_module);
	}
exit:
	if (_logger_object == Py_None) {
		Py_DECREF(_logger_object);
		_logger_object = NULL;
	}
	return _logger_object;
}

static void
lasso_python_log(G_GNUC_UNUSED const char *domain, GLogLevelFlags log_level, const gchar *message,
		G_GNUC_UNUSED gpointer user_data)
{
	PyObject *logger_object = get_logger_object(), *result;
	char *method = NULL;

	if (! logger_object) {
		PyErr_SetString(PyExc_RuntimeError, "neither lasso.logger nor "
				"logging.getLogger('lasso') did return a logger");
		return;
	}
	switch (log_level) {
		case G_LOG_LEVEL_DEBUG:
			method = "debug";
			break;
		case G_LOG_LEVEL_INFO:
		case G_LOG_LEVEL_MESSAGE:
			method = "info";
			break;
		case G_LOG_LEVEL_WARNING:
			method = "warning";
			break;
		case G_LOG_LEVEL_CRITICAL:
			method = "error";
			break;
		case G_LOG_LEVEL_ERROR:
			method = "critical";
			break;
		default:
			return;
	}
	result = PyObject_CallMethod(logger_object, method, "s#s", "%s", 2, message);
	Py_DECREF(logger_object);
	if (result) {
		Py_DECREF(result);
	} else {
		PyErr_Format(PyExc_RuntimeError, "lasso could not call method %s on its logger",
				method);
	}
}

#include <lasso/debug.h>
#include <lasso/key.h>
#include <lasso/errors.h>
#include <lasso/backward_comp.h>
#include <lasso/utils.h>
#include <lasso/export.h>
#include <lasso/lasso.h>
#include <lasso/registry.h>
#include <lasso/ctypes.h>
#include <lasso/logging.h>
#include <lasso/id-ff/federation.h>
#include <lasso/id-ff/profile.h>
#include <lasso/id-ff/server.h>
#include <lasso/id-ff/session.h>
#include <lasso/id-ff/provider.h>
#include <lasso/id-ff/name_registration.h>
#include <lasso/id-ff/login.h>
#include <lasso/id-ff/lecp.h>
#include <lasso/id-ff/identity.h>
#include <lasso/id-ff/logout.h>
#include <lasso/id-ff/name_identifier_mapping.h>
#include <lasso/id-ff/defederation.h>
#include <lasso/saml-2.0/profile.h>
#include <lasso/saml-2.0/ecp.h>
#include <lasso/saml-2.0/assertion_query.h>
#include <lasso/saml-2.0/name_id_management.h>
#include <lasso/saml-2.0/provider.h>
#include <lasso/saml-2.0/saml2_helper.h>
#include <lasso/xml/saml_statement_abstract.h>
#include <lasso/xml/lib_authn_request_envelope.h>
#include <lasso/xml/lib_authn_response_envelope.h>
#include <lasso/xml/saml_attribute_statement.h>
#include <lasso/xml/lib_name_identifier_mapping_request.h>
#include <lasso/xml/tools.h>
#include <lasso/xml/lib_subject.h>
#include <lasso/xml/saml_attribute_value.h>
#include <lasso/xml/lib_authentication_statement.h>
#include <lasso/xml/saml_subject_statement.h>
#include <lasso/xml/xml_enc.h>
#include <lasso/xml/samlp_response_abstract.h>
#include <lasso/xml/lib_register_name_identifier_response.h>
#include <lasso/xml/lib_federation_termination_notification.h>
#include <lasso/xml/saml_condition_abstract.h>
#include <lasso/xml/lib_authn_context.h>
#include <lasso/xml/saml_attribute_designator.h>
#include <lasso/xml/lib_request_authn_context.h>
#include <lasso/xml/saml_subject.h>
#include <lasso/xml/lib_authn_response.h>
#include <lasso/xml/lib_name_identifier_mapping_response.h>
#include <lasso/xml/saml_attribute.h>
#include <lasso/xml/saml_assertion.h>
#include <lasso/xml/lib_idp_list.h>
#include <lasso/xml/lib_authn_request.h>
#include <lasso/xml/lib_idp_entry.h>
#include <lasso/xml/samlp_status_code.h>
#include <lasso/xml/saml_advice.h>
#include <lasso/xml/strings.h>
#include <lasso/xml/samlp_response.h>
#include <lasso/xml/saml_conditions.h>
#include <lasso/xml/samlp_request_abstract.h>
#include <lasso/xml/lib_logout_request.h>
#include <lasso/xml/lib_assertion.h>
#include <lasso/xml/samlp_request.h>
#include <lasso/xml/saml_subject_statement_abstract.h>
#include <lasso/xml/saml_subject_locality.h>
#include <lasso/xml/lib_idp_entries.h>
#include <lasso/xml/lib_logout_response.h>
#include <lasso/xml/lib_status_response.h>
#include <lasso/xml/saml_name_identifier.h>
#include <lasso/xml/misc_text_node.h>
#include <lasso/xml/saml_audience_restriction_condition.h>
#include <lasso/xml/saml_subject_confirmation.h>
#include <lasso/xml/saml_authority_binding.h>
#include <lasso/xml/lib_scoping.h>
#include <lasso/xml/samlp_status.h>
#include <lasso/xml/lib_register_name_identifier_request.h>
#include <lasso/xml/saml_authentication_statement.h>
#include <lasso/xml/xml.h>
#include <lasso/xml/soap-1.1/soap_fault.h>
#include <lasso/xml/soap-1.1/soap_envelope.h>
#include <lasso/xml/soap-1.1/soap_header.h>
#include <lasso/xml/soap-1.1/soap_body.h>
#include <lasso/xml/soap-1.1/soap_detail.h>
#include <lasso/xml/saml-2.0/saml2_attribute_statement.h>
#include <lasso/xml/saml-2.0/saml2_subject_confirmation.h>
#include <lasso/xml/saml-2.0/saml2_authz_decision_statement.h>
#include <lasso/xml/saml-2.0/samlp2_idp_list.h>
#include <lasso/xml/saml-2.0/saml2_condition_abstract.h>
#include <lasso/xml/saml-2.0/samlp2_assertion_id_request.h>
#include <lasso/xml/saml-2.0/saml2_key_info_confirmation_data.h>
#include <lasso/xml/saml-2.0/saml2_authn_context.h>
#include <lasso/xml/saml-2.0/saml2_key_info_confirmation_data_type.h>
#include <lasso/xml/saml-2.0/samlp2_extensions.h>
#include <lasso/xml/saml-2.0/saml2_base_idabstract.h>
#include <lasso/xml/saml-2.0/saml2_subject_locality.h>
#include <lasso/xml/saml-2.0/saml2_authn_statement.h>
#include <lasso/xml/saml-2.0/samlp2_status.h>
#include <lasso/xml/saml-2.0/samlp2_request_abstract.h>
#include <lasso/xml/saml-2.0/samlp2_name_id_policy.h>
#include <lasso/xml/saml-2.0/samlp2_manage_name_id_response.h>
#include <lasso/xml/saml-2.0/samlp2_artifact_response.h>
#include <lasso/xml/saml-2.0/samlp2_authn_request.h>
#include <lasso/xml/saml-2.0/samlp2_authn_query.h>
#include <lasso/xml/saml-2.0/saml2_proxy_restriction.h>
#include <lasso/xml/saml-2.0/saml2_subject.h>
#include <lasso/xml/saml-2.0/saml2_strings.h>
#include <lasso/xml/saml-2.0/saml2_one_time_use.h>
#include <lasso/xml/saml-2.0/samlp2_requested_authn_context.h>
#include <lasso/xml/saml-2.0/samlp2_status_response.h>
#include <lasso/xml/saml-2.0/saml2_audience_restriction.h>
#include <lasso/xml/saml-2.0/saml2_attribute.h>
#include <lasso/xml/saml-2.0/samlp2_name_id_mapping_request.h>
#include <lasso/xml/saml-2.0/samlp2_manage_name_id_request.h>
#include <lasso/xml/saml-2.0/samlp2_status_code.h>
#include <lasso/xml/saml-2.0/samlp2_terminate.h>
#include <lasso/xml/saml-2.0/saml2_advice.h>
#include <lasso/xml/saml-2.0/saml2_action.h>
#include <lasso/xml/saml-2.0/samlp2_status_detail.h>
#include <lasso/xml/saml-2.0/saml2_encrypted_element.h>
#include <lasso/xml/saml-2.0/saml2_assertion.h>
#include <lasso/xml/saml-2.0/samlp2_subject_query_abstract.h>
#include <lasso/xml/saml-2.0/samlp2_logout_response.h>
#include <lasso/xml/saml-2.0/saml2_statement_abstract.h>
#include <lasso/xml/saml-2.0/samlp2_scoping.h>
#include <lasso/xml/saml-2.0/saml2_attribute_value.h>
#include <lasso/xml/saml-2.0/samlp2_idp_entry.h>
#include <lasso/xml/saml-2.0/samlp2_response.h>
#include <lasso/xml/saml-2.0/saml2_conditions.h>
#include <lasso/xml/saml-2.0/samlp2_artifact_resolve.h>
#include <lasso/xml/saml-2.0/samlp2_logout_request.h>
#include <lasso/xml/saml-2.0/samlp2_attribute_query.h>
#include <lasso/xml/saml-2.0/saml2_evidence.h>
#include <lasso/xml/saml-2.0/saml2_name_id.h>
#include <lasso/xml/saml-2.0/samlp2_name_id_mapping_response.h>
#include <lasso/xml/saml-2.0/samlp2_authz_decision_query.h>
#include <lasso/xml/saml-2.0/saml2_subject_confirmation_data.h>
#include <lasso/xml/dsig/ds_key_info.h>
#include <lasso/xml/dsig/ds_rsa_key_value.h>
#include <lasso/xml/dsig/ds_x509_data.h>
#include <lasso/xml/dsig/xml_dsig.h>
#include <lasso/xml/dsig/strings.h>
#include <lasso/xml/dsig/ds_key_value.h>

static void
register_constants(PyObject *d)
{
    PyObject *obj;

    obj = PyInt_FromLong(LASSO_ERROR_UNDEFINED);
    PyDict_SetItemString(d, "ERROR_UNDEFINED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ERROR_UNIMPLEMENTED);
    PyDict_SetItemString(d, "ERROR_UNIMPLEMENTED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ERROR_OUT_OF_MEMORY);
    PyDict_SetItemString(d, "ERROR_OUT_OF_MEMORY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ERROR_CAST_FAILED);
    PyDict_SetItemString(d, "ERROR_CAST_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_XML_ERROR_NODE_NOT_FOUND);
    PyDict_SetItemString(d, "XML_ERROR_NODE_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_XML_ERROR_NODE_CONTENT_NOT_FOUND);
    PyDict_SetItemString(d, "XML_ERROR_NODE_CONTENT_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_XML_ERROR_ATTR_NOT_FOUND);
    PyDict_SetItemString(d, "XML_ERROR_ATTR_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_XML_ERROR_ATTR_VALUE_NOT_FOUND);
    PyDict_SetItemString(d, "XML_ERROR_ATTR_VALUE_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_XML_ERROR_INVALID_FILE);
    PyDict_SetItemString(d, "XML_ERROR_INVALID_FILE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_XML_ERROR_OBJECT_CONSTRUCTION_FAILED);
    PyDict_SetItemString(d, "XML_ERROR_OBJECT_CONSTRUCTION_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_XML_ERROR_MISSING_NAMESPACE);
    PyDict_SetItemString(d, "XML_ERROR_MISSING_NAMESPACE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_XML_ERROR_SCHEMA_INVALID_FRAGMENT);
    PyDict_SetItemString(d, "XML_ERROR_SCHEMA_INVALID_FRAGMENT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_SIGNATURE_NOT_FOUND);
    PyDict_SetItemString(d, "DS_ERROR_SIGNATURE_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_INVALID_SIGNATURE);
    PyDict_SetItemString(d, "DS_ERROR_INVALID_SIGNATURE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_SIGNATURE_TMPL_CREATION_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_SIGNATURE_TMPL_CREATION_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_CONTEXT_CREATION_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_CONTEXT_CREATION_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_PUBLIC_KEY_LOAD_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_PUBLIC_KEY_LOAD_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_PRIVATE_KEY_LOAD_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_PRIVATE_KEY_LOAD_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_CERTIFICATE_LOAD_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_CERTIFICATE_LOAD_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_SIGNATURE_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_SIGNATURE_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_KEYS_MNGR_CREATION_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_KEYS_MNGR_CREATION_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_KEYS_MNGR_INIT_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_KEYS_MNGR_INIT_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_SIGNATURE_VERIFICATION_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_SIGNATURE_VERIFICATION_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_CA_CERT_CHAIN_LOAD_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_CA_CERT_CHAIN_LOAD_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_INVALID_SIGALG);
    PyDict_SetItemString(d, "DS_ERROR_INVALID_SIGALG", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_DIGEST_COMPUTE_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_DIGEST_COMPUTE_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_SIGNATURE_TEMPLATE_NOT_FOUND);
    PyDict_SetItemString(d, "DS_ERROR_SIGNATURE_TEMPLATE_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_TOO_MUCH_REFERENCES);
    PyDict_SetItemString(d, "DS_ERROR_TOO_MUCH_REFERENCES", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_INVALID_REFERENCE_FOR_SAML);
    PyDict_SetItemString(d, "DS_ERROR_INVALID_REFERENCE_FOR_SAML", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_DECRYPTION_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_DECRYPTION_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_ENCRYPTION_FAILED);
    PyDict_SetItemString(d, "DS_ERROR_ENCRYPTION_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DS_ERROR_DECRYPTION_FAILED_MISSING_PRIVATE_KEY);
    PyDict_SetItemString(d, "DS_ERROR_DECRYPTION_FAILED_MISSING_PRIVATE_KEY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SERVER_ERROR_PROVIDER_NOT_FOUND);
    PyDict_SetItemString(d, "SERVER_ERROR_PROVIDER_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SERVER_ERROR_ADD_PROVIDER_FAILED);
    PyDict_SetItemString(d, "SERVER_ERROR_ADD_PROVIDER_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SERVER_ERROR_ADD_PROVIDER_PROTOCOL_MISMATCH);
    PyDict_SetItemString(d, "SERVER_ERROR_ADD_PROVIDER_PROTOCOL_MISMATCH", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SERVER_ERROR_SET_ENCRYPTION_PRIVATE_KEY_FAILED);
    PyDict_SetItemString(d, "SERVER_ERROR_SET_ENCRYPTION_PRIVATE_KEY_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SERVER_ERROR_INVALID_XML);
    PyDict_SetItemString(d, "SERVER_ERROR_INVALID_XML", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SERVER_ERROR_NO_PROVIDER_LOADED);
    PyDict_SetItemString(d, "SERVER_ERROR_NO_PROVIDER_LOADED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGOUT_ERROR_UNSUPPORTED_PROFILE);
    PyDict_SetItemString(d, "LOGOUT_ERROR_UNSUPPORTED_PROFILE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGOUT_ERROR_REQUEST_DENIED);
    PyDict_SetItemString(d, "LOGOUT_ERROR_REQUEST_DENIED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGOUT_ERROR_FEDERATION_NOT_FOUND);
    PyDict_SetItemString(d, "LOGOUT_ERROR_FEDERATION_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGOUT_ERROR_UNKNOWN_PRINCIPAL);
    PyDict_SetItemString(d, "LOGOUT_ERROR_UNKNOWN_PRINCIPAL", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGOUT_ERROR_PARTIAL_LOGOUT);
    PyDict_SetItemString(d, "LOGOUT_ERROR_PARTIAL_LOGOUT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_INVALID_QUERY);
    PyDict_SetItemString(d, "PROFILE_ERROR_INVALID_QUERY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_INVALID_POST_MSG);
    PyDict_SetItemString(d, "PROFILE_ERROR_INVALID_POST_MSG", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_INVALID_SOAP_MSG);
    PyDict_SetItemString(d, "PROFILE_ERROR_INVALID_SOAP_MSG", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_REQUEST);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_REQUEST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_INVALID_HTTP_METHOD);
    PyDict_SetItemString(d, "PROFILE_ERROR_INVALID_HTTP_METHOD", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_INVALID_PROTOCOLPROFILE);
    PyDict_SetItemString(d, "PROFILE_ERROR_INVALID_PROTOCOLPROFILE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_INVALID_MSG);
    PyDict_SetItemString(d, "PROFILE_ERROR_INVALID_MSG", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_REMOTE_PROVIDERID);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_REMOTE_PROVIDERID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_UNSUPPORTED_PROFILE);
    PyDict_SetItemString(d, "PROFILE_ERROR_UNSUPPORTED_PROFILE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_UNKNOWN_PROFILE_URL);
    PyDict_SetItemString(d, "PROFILE_ERROR_UNKNOWN_PROFILE_URL", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_IDENTITY_NOT_FOUND);
    PyDict_SetItemString(d, "PROFILE_ERROR_IDENTITY_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_FEDERATION_NOT_FOUND);
    PyDict_SetItemString(d, "PROFILE_ERROR_FEDERATION_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_NAME_IDENTIFIER_NOT_FOUND);
    PyDict_SetItemString(d, "PROFILE_ERROR_NAME_IDENTIFIER_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_BUILDING_QUERY_FAILED);
    PyDict_SetItemString(d, "PROFILE_ERROR_BUILDING_QUERY_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_BUILDING_REQUEST_FAILED);
    PyDict_SetItemString(d, "PROFILE_ERROR_BUILDING_REQUEST_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_BUILDING_MESSAGE_FAILED);
    PyDict_SetItemString(d, "PROFILE_ERROR_BUILDING_MESSAGE_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_BUILDING_RESPONSE_FAILED);
    PyDict_SetItemString(d, "PROFILE_ERROR_BUILDING_RESPONSE_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_SESSION_NOT_FOUND);
    PyDict_SetItemString(d, "PROFILE_ERROR_SESSION_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_BAD_IDENTITY_DUMP);
    PyDict_SetItemString(d, "PROFILE_ERROR_BAD_IDENTITY_DUMP", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_BAD_SESSION_DUMP);
    PyDict_SetItemString(d, "PROFILE_ERROR_BAD_SESSION_DUMP", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_RESPONSE);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_RESPONSE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_STATUS_CODE);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_STATUS_CODE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_ARTIFACT);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_ARTIFACT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_RESOURCE_OFFERING);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_RESOURCE_OFFERING", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_SERVICE_DESCRIPTION);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_SERVICE_DESCRIPTION", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_SERVICE_TYPE);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_SERVICE_TYPE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_ASSERTION);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_ASSERTION", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_SUBJECT);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_SUBJECT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_NAME_IDENTIFIER);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_NAME_IDENTIFIER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_INVALID_ARTIFACT);
    PyDict_SetItemString(d, "PROFILE_ERROR_INVALID_ARTIFACT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_ENCRYPTION_PRIVATE_KEY);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_ENCRYPTION_PRIVATE_KEY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_STATUS_NOT_SUCCESS);
    PyDict_SetItemString(d, "PROFILE_ERROR_STATUS_NOT_SUCCESS", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_ISSUER);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_ISSUER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_SERVICE_INSTANCE);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_SERVICE_INSTANCE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_ENDPOINT_REFERENCE);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_ENDPOINT_REFERENCE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_ENDPOINT_REFERENCE_ADDRESS);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_ENDPOINT_REFERENCE_ADDRESS", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_INVALID_ISSUER);
    PyDict_SetItemString(d, "PROFILE_ERROR_INVALID_ISSUER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_MISSING_SERVER);
    PyDict_SetItemString(d, "PROFILE_ERROR_MISSING_SERVER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_UNKNOWN_PROVIDER);
    PyDict_SetItemString(d, "PROFILE_ERROR_UNKNOWN_PROVIDER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_CANNOT_VERIFY_SIGNATURE);
    PyDict_SetItemString(d, "PROFILE_ERROR_CANNOT_VERIFY_SIGNATURE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_CANNOT_FIND_A_PROVIDER);
    PyDict_SetItemString(d, "PROFILE_ERROR_CANNOT_FIND_A_PROVIDER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_RESPONSE_DOES_NOT_MATCH_REQUEST);
    PyDict_SetItemString(d, "PROFILE_ERROR_RESPONSE_DOES_NOT_MATCH_REQUEST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_INVALID_REQUEST);
    PyDict_SetItemString(d, "PROFILE_ERROR_INVALID_REQUEST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_INVALID_RESPONSE);
    PyDict_SetItemString(d, "PROFILE_ERROR_INVALID_RESPONSE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_UNSUPPORTED_BINDING);
    PyDict_SetItemString(d, "PROFILE_ERROR_UNSUPPORTED_BINDING", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_INVALID_ASSERTION_CONDITIONS);
    PyDict_SetItemString(d, "PROFILE_ERROR_INVALID_ASSERTION_CONDITIONS", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_INVALID_ASSERTION);
    PyDict_SetItemString(d, "PROFILE_ERROR_INVALID_ASSERTION", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_UNKNOWN_ISSUER);
    PyDict_SetItemString(d, "PROFILE_ERROR_UNKNOWN_ISSUER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_ISSUER_IS_NOT_AN_IDP);
    PyDict_SetItemString(d, "PROFILE_ERROR_ISSUER_IS_NOT_AN_IDP", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_REQUEST_DENIED);
    PyDict_SetItemString(d, "PROFILE_ERROR_REQUEST_DENIED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_ERROR_ENDPOINT_INDEX_NOT_FOUND);
    PyDict_SetItemString(d, "PROFILE_ERROR_ENDPOINT_INDEX_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PARAM_ERROR_BAD_TYPE_OR_NULL_OBJ);
    PyDict_SetItemString(d, "PARAM_ERROR_BAD_TYPE_OR_NULL_OBJ", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PARAM_ERROR_INVALID_VALUE);
    PyDict_SetItemString(d, "PARAM_ERROR_INVALID_VALUE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PARAM_ERROR_CHECK_FAILED);
    PyDict_SetItemString(d, "PARAM_ERROR_CHECK_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PARAM_ERROR_NON_INITIALIZED_OBJECT);
    PyDict_SetItemString(d, "PARAM_ERROR_NON_INITIALIZED_OBJECT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_ERROR_FEDERATION_NOT_FOUND);
    PyDict_SetItemString(d, "LOGIN_ERROR_FEDERATION_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_ERROR_CONSENT_NOT_OBTAINED);
    PyDict_SetItemString(d, "LOGIN_ERROR_CONSENT_NOT_OBTAINED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_ERROR_INVALID_NAMEIDPOLICY);
    PyDict_SetItemString(d, "LOGIN_ERROR_INVALID_NAMEIDPOLICY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_ERROR_REQUEST_DENIED);
    PyDict_SetItemString(d, "LOGIN_ERROR_REQUEST_DENIED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_ERROR_INVALID_SIGNATURE);
    PyDict_SetItemString(d, "LOGIN_ERROR_INVALID_SIGNATURE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_ERROR_UNSIGNED_AUTHN_REQUEST);
    PyDict_SetItemString(d, "LOGIN_ERROR_UNSIGNED_AUTHN_REQUEST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_ERROR_STATUS_NOT_SUCCESS);
    PyDict_SetItemString(d, "LOGIN_ERROR_STATUS_NOT_SUCCESS", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_ERROR_UNKNOWN_PRINCIPAL);
    PyDict_SetItemString(d, "LOGIN_ERROR_UNKNOWN_PRINCIPAL", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_ERROR_NO_DEFAULT_ENDPOINT);
    PyDict_SetItemString(d, "LOGIN_ERROR_NO_DEFAULT_ENDPOINT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_ERROR_ASSERTION_REPLAY);
    PyDict_SetItemString(d, "LOGIN_ERROR_ASSERTION_REPLAY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_ERROR_ASSERTION_DOES_NOT_MATCH_REQUEST_ID);
    PyDict_SetItemString(d, "LOGIN_ERROR_ASSERTION_DOES_NOT_MATCH_REQUEST_ID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_ERROR_INVALID_ASSERTION_SIGNATURE);
    PyDict_SetItemString(d, "LOGIN_ERROR_INVALID_ASSERTION_SIGNATURE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DEFEDERATION_ERROR_MISSING_NAME_IDENTIFIER);
    PyDict_SetItemString(d, "DEFEDERATION_ERROR_MISSING_NAME_IDENTIFIER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SOAP_ERROR_REDIRECT_REQUEST_FAULT);
    PyDict_SetItemString(d, "SOAP_ERROR_REDIRECT_REQUEST_FAULT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SOAP_FAULT_REDIRECT_REQUEST);
    PyDict_SetItemString(d, "SOAP_FAULT_REDIRECT_REQUEST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SOAP_ERROR_MISSING_ENVELOPE);
    PyDict_SetItemString(d, "SOAP_ERROR_MISSING_ENVELOPE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SOAP_ERROR_MISSING_HEADER);
    PyDict_SetItemString(d, "SOAP_ERROR_MISSING_HEADER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SOAP_ERROR_MISSING_BODY);
    PyDict_SetItemString(d, "SOAP_ERROR_MISSING_BODY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SOAP_ERROR_MISSING_SOAP_FAULT_DETAIL);
    PyDict_SetItemString(d, "SOAP_ERROR_MISSING_SOAP_FAULT_DETAIL", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_NAME_IDENTIFIER_MAPPING_ERROR_MISSING_TARGET_NAMESPACE);
    PyDict_SetItemString(d, "NAME_IDENTIFIER_MAPPING_ERROR_MISSING_TARGET_NAMESPACE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_NAME_IDENTIFIER_MAPPING_ERROR_FORBIDDEN_CALL_ON_THIS_SIDE);
    PyDict_SetItemString(d, "NAME_IDENTIFIER_MAPPING_ERROR_FORBIDDEN_CALL_ON_THIS_SIDE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_NAME_IDENTIFIER_MAPPING_ERROR_MISSING_TARGET_IDENTIFIER);
    PyDict_SetItemString(d, "NAME_IDENTIFIER_MAPPING_ERROR_MISSING_TARGET_IDENTIFIER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DATA_SERVICE_ERROR_UNREGISTERED_DST);
    PyDict_SetItemString(d, "DATA_SERVICE_ERROR_UNREGISTERED_DST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DATA_SERVICE_ERROR_CANNOT_ADD_ITEM);
    PyDict_SetItemString(d, "DATA_SERVICE_ERROR_CANNOT_ADD_ITEM", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_MISSING_CORRELATION);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_MISSING_CORRELATION", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_MISSING_SECURITY);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_MISSING_SECURITY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_MISSING_ASSERTION_ID);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_MISSING_ASSERTION_ID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_MISSING_ENDPOINT);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_MISSING_ENDPOINT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_SOAP_FAULT);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_SOAP_FAULT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_UNSUPPORTED_SECURITY_MECHANISM);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_UNSUPPORTED_SECURITY_MECHANISM", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_MISSING_DESCRIPTION);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_MISSING_DESCRIPTION", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_MISSING_RESOURCE_ID);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_MISSING_RESOURCE_ID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_MISSING_CREDENTIAL_REF);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_MISSING_CREDENTIAL_REF", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_INVALID_OR_MISSING_REFERENCE_TO_MESSAGE_ID);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_INVALID_OR_MISSING_REFERENCE_TO_MESSAGE_ID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_SECURITY_MECHANISM_CHECK_FAILED);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_SECURITY_MECHANISM_CHECK_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_UNKNOWN_STATUS_CODE);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_UNKNOWN_STATUS_CODE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_SERVER_INTERACTION_REQUIRED);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_SERVER_INTERACTION_REQUIRED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_SERVER_INTERACTION_REQUIRED_FOR_DATA);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_SERVER_INTERACTION_REQUIRED_FOR_DATA", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_REDIRECT_REQUEST);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_REDIRECT_REQUEST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_REDIRECT_REQUEST_UNSUPPORTED_BY_REQUESTER);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_REDIRECT_REQUEST_UNSUPPORTED_BY_REQUESTER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSF_PROFILE_ERROR_MISSING_SENDER_ID);
    PyDict_SetItemString(d, "WSF_PROFILE_ERROR_MISSING_SENDER_ID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DISCOVERY_ERROR_SVC_METADATA_REGISTER_FAILED);
    PyDict_SetItemString(d, "DISCOVERY_ERROR_SVC_METADATA_REGISTER_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DISCOVERY_ERROR_SVC_METADATA_ASSOCIATION_ADD_FAILED);
    PyDict_SetItemString(d, "DISCOVERY_ERROR_SVC_METADATA_ASSOCIATION_ADD_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DISCOVERY_ERROR_MISSING_REQUESTED_SERVICE);
    PyDict_SetItemString(d, "DISCOVERY_ERROR_MISSING_REQUESTED_SERVICE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DISCOVERY_ERROR_FAILED_TO_BUILD_ENDPOINT_REFERENCE);
    PyDict_SetItemString(d, "DISCOVERY_ERROR_FAILED_TO_BUILD_ENDPOINT_REFERENCE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DST_ERROR_MISSING_SERVICE_DATA);
    PyDict_SetItemString(d, "DST_ERROR_MISSING_SERVICE_DATA", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DST_ERROR_QUERY_FAILED);
    PyDict_SetItemString(d, "DST_ERROR_QUERY_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DST_ERROR_QUERY_PARTIALLY_FAILED);
    PyDict_SetItemString(d, "DST_ERROR_QUERY_PARTIALLY_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DST_ERROR_MODIFY_FAILED);
    PyDict_SetItemString(d, "DST_ERROR_MODIFY_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DST_ERROR_MODIFY_PARTIALLY_FAILED);
    PyDict_SetItemString(d, "DST_ERROR_MODIFY_PARTIALLY_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DST_ERROR_NEW_DATA_MISSING);
    PyDict_SetItemString(d, "DST_ERROR_NEW_DATA_MISSING", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DST_ERROR_QUERY_NOT_FOUND);
    PyDict_SetItemString(d, "DST_ERROR_QUERY_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DST_ERROR_NO_DATA);
    PyDict_SetItemString(d, "DST_ERROR_NO_DATA", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DST_ERROR_MALFORMED_QUERY);
    PyDict_SetItemString(d, "DST_ERROR_MALFORMED_QUERY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DST_ERROR_EMPTY_REQUEST);
    PyDict_SetItemString(d, "DST_ERROR_EMPTY_REQUEST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REGISTRY_ERROR_KEY_EXISTS);
    PyDict_SetItemString(d, "REGISTRY_ERROR_KEY_EXISTS", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROVIDER_ERROR_MISSING_PUBLIC_KEY);
    PyDict_SetItemString(d, "PROVIDER_ERROR_MISSING_PUBLIC_KEY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSSEC_ERROR_MISSING_SECURITY_TOKEN);
    PyDict_SetItemString(d, "WSSEC_ERROR_MISSING_SECURITY_TOKEN", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_WSSEC_ERROR_BAD_PASSWORD);
    PyDict_SetItemString(d, "WSSEC_ERROR_BAD_PASSWORD", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_IDWSF2_DISCOVERY_ERROR_FAILED);
    PyDict_SetItemString(d, "IDWSF2_DISCOVERY_ERROR_FAILED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_IDWSF2_DISCOVERY_ERROR_FORBIDDEN);
    PyDict_SetItemString(d, "IDWSF2_DISCOVERY_ERROR_FORBIDDEN", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_IDWSF2_DISCOVERY_ERROR_DUPLICATE);
    PyDict_SetItemString(d, "IDWSF2_DISCOVERY_ERROR_DUPLICATE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_IDWSF2_DISCOVERY_ERROR_LOGICAL_DUPLICATE);
    PyDict_SetItemString(d, "IDWSF2_DISCOVERY_ERROR_LOGICAL_DUPLICATE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_IDWSF2_DISCOVERY_ERROR_NO_RESULTS);
    PyDict_SetItemString(d, "IDWSF2_DISCOVERY_ERROR_NO_RESULTS", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_IDWSF2_DISCOVERY_ERROR_NOT_FOUND);
    PyDict_SetItemString(d, "IDWSF2_DISCOVERY_ERROR_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_IDWSF2_DST_ERROR_DUPLICATE_ITEM);
    PyDict_SetItemString(d, "IDWSF2_DST_ERROR_DUPLICATE_ITEM", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_IDWSF2_DST_ERROR_PARTIAL_FAILURE);
    PyDict_SetItemString(d, "IDWSF2_DST_ERROR_PARTIAL_FAILURE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_IDWSF2_DST_ERROR_UNKNOWN_STATUS_CODE);
    PyDict_SetItemString(d, "IDWSF2_DST_ERROR_UNKNOWN_STATUS_CODE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_IDWSF2_DST_ERROR_ITEM_NOT_FOUND);
    PyDict_SetItemString(d, "IDWSF2_DST_ERROR_ITEM_NOT_FOUND", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ASSERTION_QUERY_ERROR_ATTRIBUTE_REQUEST_ALREADY_EXIST);
    PyDict_SetItemString(d, "ASSERTION_QUERY_ERROR_ATTRIBUTE_REQUEST_ALREADY_EXIST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ASSERTION_QUERY_ERROR_NOT_AN_ATTRIBUTE_QUERY);
    PyDict_SetItemString(d, "ASSERTION_QUERY_ERROR_NOT_AN_ATTRIBUTE_QUERY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_XMLENC_ERROR_INVALID_ENCRYPTED_DATA);
    PyDict_SetItemString(d, "XMLENC_ERROR_INVALID_ENCRYPTED_DATA", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(BACKWARD_COMP_H);
    PyDict_SetItemString(d, "RD_COMP_H", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_CHECK_VERSION_EXACT);
    PyDict_SetItemString(d, "CHECK_VERSION_EXACT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_CHECK_VERSIONABI_COMPATIBLE);
    PyDict_SetItemString(d, "CHECK_VERSIONABI_COMPATIBLE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_CHECK_VERSION_NUMERIC);
    PyDict_SetItemString(d, "CHECK_VERSION_NUMERIC", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_INVALID);
    PyDict_SetItemString(d, "REQUEST_TYPE_INVALID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_LOGIN);
    PyDict_SetItemString(d, "REQUEST_TYPE_LOGIN", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_LOGOUT);
    PyDict_SetItemString(d, "REQUEST_TYPE_LOGOUT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_DEFEDERATION);
    PyDict_SetItemString(d, "REQUEST_TYPE_DEFEDERATION", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_NAME_REGISTRATION);
    PyDict_SetItemString(d, "REQUEST_TYPE_NAME_REGISTRATION", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_NAME_IDENTIFIER_MAPPING);
    PyDict_SetItemString(d, "REQUEST_TYPE_NAME_IDENTIFIER_MAPPING", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_LECP);
    PyDict_SetItemString(d, "REQUEST_TYPE_LECP", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_DISCO_QUERY);
    PyDict_SetItemString(d, "REQUEST_TYPE_DISCO_QUERY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_DISCO_MODIFY);
    PyDict_SetItemString(d, "REQUEST_TYPE_DISCO_MODIFY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_DST_QUERY);
    PyDict_SetItemString(d, "REQUEST_TYPE_DST_QUERY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_DST_MODIFY);
    PyDict_SetItemString(d, "REQUEST_TYPE_DST_MODIFY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_SASL_REQUEST);
    PyDict_SetItemString(d, "REQUEST_TYPE_SASL_REQUEST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_NAME_ID_MANAGEMENT);
    PyDict_SetItemString(d, "REQUEST_TYPE_NAME_ID_MANAGEMENT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_IDWSF2_DISCO_SVCMD_REGISTER);
    PyDict_SetItemString(d, "REQUEST_TYPE_IDWSF2_DISCO_SVCMD_REGISTER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_IDWSF2_DISCO_SVCMD_ASSOCIATION_ADD);
    PyDict_SetItemString(d, "REQUEST_TYPE_IDWSF2_DISCO_SVCMD_ASSOCIATION_ADD", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_REQUEST_TYPE_IDWSF2_DISCO_QUERY);
    PyDict_SetItemString(d, "REQUEST_TYPE_IDWSF2_DISCO_QUERY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_SIGNATURE_HINT_MAYBE);
    PyDict_SetItemString(d, "PROFILE_SIGNATURE_HINT_MAYBE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_SIGNATURE_HINT_FORCE);
    PyDict_SetItemString(d, "PROFILE_SIGNATURE_HINT_FORCE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_SIGNATURE_HINT_FORBID);
    PyDict_SetItemString(d, "PROFILE_SIGNATURE_HINT_FORBID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_SIGNATURE_VERIFY_HINT_MAYBE);
    PyDict_SetItemString(d, "PROFILE_SIGNATURE_VERIFY_HINT_MAYBE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_SIGNATURE_VERIFY_HINT_FORCE);
    PyDict_SetItemString(d, "PROFILE_SIGNATURE_VERIFY_HINT_FORCE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_SIGNATURE_VERIFY_HINT_IGNORE);
    PyDict_SetItemString(d, "PROFILE_SIGNATURE_VERIFY_HINT_IGNORE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROFILE_SIGNATURE_VERIFY_HINT_LAST);
    PyDict_SetItemString(d, "PROFILE_SIGNATURE_VERIFY_HINT_LAST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SERVER_LOAD_METADATA_FLAG_DEFAULT);
    PyDict_SetItemString(d, "SERVER_LOAD_METADATA_FLAG_DEFAULT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SERVER_LOAD_METADATA_FLAG_CHECK_ENTITIES_DESCRIPTOR_SIGNATURE);
    PyDict_SetItemString(d, "SERVER_LOAD_METADATA_FLAG_CHECK_ENTITIES_DESCRIPTOR_SIGNATURE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SERVER_LOAD_METADATA_FLAG_CHECK_ENTITY_DESCRIPTOR_SIGNATURE);
    PyDict_SetItemString(d, "SERVER_LOAD_METADATA_FLAG_CHECK_ENTITY_DESCRIPTOR_SIGNATURE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SERVER_LOAD_METADATA_FLAG_INHERIT_SIGNATURE);
    PyDict_SetItemString(d, "SERVER_LOAD_METADATA_FLAG_INHERIT_SIGNATURE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_HTTP_METHOD_NONE);
    PyDict_SetItemString(d, "HTTP_METHOD_NONE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_HTTP_METHOD_ANY);
    PyDict_SetItemString(d, "HTTP_METHOD_ANY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_HTTP_METHOD_IDP_INITIATED);
    PyDict_SetItemString(d, "HTTP_METHOD_IDP_INITIATED", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_HTTP_METHOD_GET);
    PyDict_SetItemString(d, "HTTP_METHOD_GET", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_HTTP_METHOD_POST);
    PyDict_SetItemString(d, "HTTP_METHOD_POST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_HTTP_METHOD_REDIRECT);
    PyDict_SetItemString(d, "HTTP_METHOD_REDIRECT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_HTTP_METHOD_SOAP);
    PyDict_SetItemString(d, "HTTP_METHOD_SOAP", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_HTTP_METHOD_ARTIFACT_GET);
    PyDict_SetItemString(d, "HTTP_METHOD_ARTIFACT_GET", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_HTTP_METHOD_ARTIFACT_POST);
    PyDict_SetItemString(d, "HTTP_METHOD_ARTIFACT_POST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_HTTP_METHOD_PAOS);
    PyDict_SetItemString(d, "HTTP_METHOD_PAOS", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_HTTP_METHOD_LAST);
    PyDict_SetItemString(d, "HTTP_METHOD_LAST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MD_PROTOCOL_TYPE_FEDERATION_TERMINATION);
    PyDict_SetItemString(d, "MD_PROTOCOL_TYPE_FEDERATION_TERMINATION", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MD_PROTOCOL_TYPE_NAME_IDENTIFIER_MAPPING);
    PyDict_SetItemString(d, "MD_PROTOCOL_TYPE_NAME_IDENTIFIER_MAPPING", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MD_PROTOCOL_TYPE_REGISTER_NAME_IDENTIFIER);
    PyDict_SetItemString(d, "MD_PROTOCOL_TYPE_REGISTER_NAME_IDENTIFIER", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MD_PROTOCOL_TYPE_SINGLE_LOGOUT);
    PyDict_SetItemString(d, "MD_PROTOCOL_TYPE_SINGLE_LOGOUT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MD_PROTOCOL_TYPE_SINGLE_SIGN_ON);
    PyDict_SetItemString(d, "MD_PROTOCOL_TYPE_SINGLE_SIGN_ON", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MD_PROTOCOL_TYPE_ARTIFACT_RESOLUTION);
    PyDict_SetItemString(d, "MD_PROTOCOL_TYPE_ARTIFACT_RESOLUTION", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MD_PROTOCOL_TYPE_MANAGE_NAME_ID);
    PyDict_SetItemString(d, "MD_PROTOCOL_TYPE_MANAGE_NAME_ID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MD_PROTOCOL_TYPE_ASSERTION_ID_REQUEST);
    PyDict_SetItemString(d, "MD_PROTOCOL_TYPE_ASSERTION_ID_REQUEST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MD_PROTOCOL_TYPE_AUTHN_QUERY);
    PyDict_SetItemString(d, "MD_PROTOCOL_TYPE_AUTHN_QUERY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MD_PROTOCOL_TYPE_AUTHZ);
    PyDict_SetItemString(d, "MD_PROTOCOL_TYPE_AUTHZ", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MD_PROTOCOL_TYPE_ATTRIBUTE);
    PyDict_SetItemString(d, "MD_PROTOCOL_TYPE_ATTRIBUTE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MD_PROTOCOL_TYPE_LAST);
    PyDict_SetItemString(d, "MD_PROTOCOL_TYPE_LAST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROVIDER_ROLE_ANY);
    PyDict_SetItemString(d, "PROVIDER_ROLE_ANY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROVIDER_ROLE_NONE);
    PyDict_SetItemString(d, "PROVIDER_ROLE_NONE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROVIDER_ROLE_SP);
    PyDict_SetItemString(d, "PROVIDER_ROLE_SP", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROVIDER_ROLE_IDP);
    PyDict_SetItemString(d, "PROVIDER_ROLE_IDP", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROVIDER_ROLE_BOTH);
    PyDict_SetItemString(d, "PROVIDER_ROLE_BOTH", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROVIDER_ROLE_AUTHN_AUTHORITY);
    PyDict_SetItemString(d, "PROVIDER_ROLE_AUTHN_AUTHORITY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROVIDER_ROLE_AUTHZ_AUTHORITY);
    PyDict_SetItemString(d, "PROVIDER_ROLE_AUTHZ_AUTHORITY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROVIDER_ROLE_ATTRIBUTE_AUTHORITY);
    PyDict_SetItemString(d, "PROVIDER_ROLE_ATTRIBUTE_AUTHORITY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROVIDER_ROLE_LAST);
    PyDict_SetItemString(d, "PROVIDER_ROLE_LAST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROVIDER_ROLE_ALL);
    PyDict_SetItemString(d, "PROVIDER_ROLE_ALL", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROTOCOL_NONE);
    PyDict_SetItemString(d, "PROTOCOL_NONE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROTOCOL_LIBERTY_1_0);
    PyDict_SetItemString(d, "PROTOCOL_LIBERTY_1_0", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROTOCOL_LIBERTY_1_1);
    PyDict_SetItemString(d, "PROTOCOL_LIBERTY_1_1", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROTOCOL_LIBERTY_1_2);
    PyDict_SetItemString(d, "PROTOCOL_LIBERTY_1_2", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_PROTOCOL_SAML_2_0);
    PyDict_SetItemString(d, "PROTOCOL_SAML_2_0", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ENCRYPTION_MODE_NONE);
    PyDict_SetItemString(d, "ENCRYPTION_MODE_NONE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ENCRYPTION_MODE_NAMEID);
    PyDict_SetItemString(d, "ENCRYPTION_MODE_NAMEID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ENCRYPTION_MODE_ASSERTION);
    PyDict_SetItemString(d, "ENCRYPTION_MODE_ASSERTION", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_PROTOCOL_PROFILE_BRWS_ART);
    PyDict_SetItemString(d, "LOGIN_PROTOCOL_PROFILE_BRWS_ART", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_PROTOCOL_PROFILE_BRWS_POST);
    PyDict_SetItemString(d, "LOGIN_PROTOCOL_PROFILE_BRWS_POST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_PROTOCOL_PROFILE_BRWS_LECP);
    PyDict_SetItemString(d, "LOGIN_PROTOCOL_PROFILE_BRWS_LECP", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LOGIN_PROTOCOL_PROFILE_REDIRECT);
    PyDict_SetItemString(d, "LOGIN_PROTOCOL_PROFILE_REDIRECT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ASSERTION_QUERY_REQUEST_TYPE_UNSET);
    PyDict_SetItemString(d, "ASSERTION_QUERY_REQUEST_TYPE_UNSET", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ASSERTION_QUERY_REQUEST_TYPE_ASSERTION_ID);
    PyDict_SetItemString(d, "ASSERTION_QUERY_REQUEST_TYPE_ASSERTION_ID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ASSERTION_QUERY_REQUEST_TYPE_AUTHN);
    PyDict_SetItemString(d, "ASSERTION_QUERY_REQUEST_TYPE_AUTHN", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ASSERTION_QUERY_REQUEST_TYPE_ATTRIBUTE);
    PyDict_SetItemString(d, "ASSERTION_QUERY_REQUEST_TYPE_ATTRIBUTE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ASSERTION_QUERY_REQUEST_TYPE_AUTHZ_DECISION);
    PyDict_SetItemString(d, "ASSERTION_QUERY_REQUEST_TYPE_AUTHZ_DECISION", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ASSERTION_QUERY_REQUEST_TYPE_LAST);
    PyDict_SetItemString(d, "ASSERTION_QUERY_REQUEST_TYPE_LAST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SAML2_ASSERTION_VALID);
    PyDict_SetItemString(d, "SAML2_ASSERTION_VALID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SAML2_ASSERTION_INVALID);
    PyDict_SetItemString(d, "SAML2_ASSERTION_INVALID", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SAML2_ASSERTION_INDETERMINATE);
    PyDict_SetItemString(d, "SAML2_ASSERTION_INDETERMINATE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DURATION_MINUTE);
    PyDict_SetItemString(d, "DURATION_MINUTE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DURATION_HOUR);
    PyDict_SetItemString(d, "DURATION_HOUR", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DURATION_DAY);
    PyDict_SetItemString(d, "DURATION_DAY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_DURATION_WEEK);
    PyDict_SetItemString(d, "DURATION_WEEK", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ENCRYPTION_SYM_KEY_TYPE_DEFAULT);
    PyDict_SetItemString(d, "ENCRYPTION_SYM_KEY_TYPE_DEFAULT", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ENCRYPTION_SYM_KEY_TYPE_AES_256);
    PyDict_SetItemString(d, "ENCRYPTION_SYM_KEY_TYPE_AES_256", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ENCRYPTION_SYM_KEY_TYPE_AES_128);
    PyDict_SetItemString(d, "ENCRYPTION_SYM_KEY_TYPE_AES_128", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ENCRYPTION_SYM_KEY_TYPE_3DES);
    PyDict_SetItemString(d, "ENCRYPTION_SYM_KEY_TYPE_3DES", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_ENCRYTPION_SYM_KEY_TYPE_LAST);
    PyDict_SetItemString(d, "ENCRYTPION_SYM_KEY_TYPE_LAST", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SOAP_ENV_HREF);
    PyDict_SetItemString(d, "SOAP_ENV_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SOAP_ENV_PREFIX);
    PyDict_SetItemString(d, "SOAP_ENV_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SOAP_ENV_ACTOR);
    PyDict_SetItemString(d, "SOAP_ENV_ACTOR", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SOAP_FAULT_CODE_SERVER);
    PyDict_SetItemString(d, "SOAP_FAULT_CODE_SERVER", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SOAP_FAULT_CODE_CLIENT);
    PyDict_SetItemString(d, "SOAP_FAULT_CODE_CLIENT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SOAP_FAULT_CODE_MUST_UNDERSTAND);
    PyDict_SetItemString(d, "SOAP_FAULT_CODE_MUST_UNDERSTAND", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SOAP_FAULT_CODE_VERSION_MISMATCH);
    PyDict_SetItemString(d, "SOAP_FAULT_CODE_VERSION_MISMATCH", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_PRIVATE_STATUS_CODE_FAILED_TO_RESTORE_ARTIFACT);
    PyDict_SetItemString(d, "PRIVATE_STATUS_CODE_FAILED_TO_RESTORE_ARTIFACT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_WSUTIL1_HREF);
    PyDict_SetItemString(d, "WSUTIL1_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_WSUTIL1_PREFIX);
    PyDict_SetItemString(d, "WSUTIL1_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_XMLENC_HREF);
    PyDict_SetItemString(d, "XMLENC_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_XMLENC_PREFIX);
    PyDict_SetItemString(d, "XMLENC_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LASSO_HREF);
    PyDict_SetItemString(d, "LASSO_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LASSO_PREFIX);
    PyDict_SetItemString(d, "LASSO_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_PYTHON_HREF);
    PyDict_SetItemString(d, "PYTHON_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SIGNATURE_TYPE_ATTRIBUTE);
    PyDict_SetItemString(d, "SIGNATURE_TYPE_ATTRIBUTE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SIGNATURE_METHOD_ATTRIBUTE);
    PyDict_SetItemString(d, "SIGNATURE_METHOD_ATTRIBUTE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_PRIVATE_KEY_ATTRIBUTE);
    PyDict_SetItemString(d, "PRIVATE_KEY_ATTRIBUTE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_PRIVATE_KEY_PASSWORD_ATTRIBUTE);
    PyDict_SetItemString(d, "PRIVATE_KEY_PASSWORD_ATTRIBUTE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_CERTIFICATE_ATTRIBUTE);
    PyDict_SetItemString(d, "CERTIFICATE_ATTRIBUTE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_HREF);
    PyDict_SetItemString(d, "LIB_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PREFIX);
    PyDict_SetItemString(d, "LIB_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LIB_MAJOR_VERSION_N);
    PyDict_SetItemString(d, "LIB_MAJOR_VERSION_N", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_LIB_MINOR_VERSION_N);
    PyDict_SetItemString(d, "LIB_MINOR_VERSION_N", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_NAMEID_POLICY_TYPE_NONE);
    PyDict_SetItemString(d, "LIB_NAMEID_POLICY_TYPE_NONE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_NAMEID_POLICY_TYPE_ONE_TIME);
    PyDict_SetItemString(d, "LIB_NAMEID_POLICY_TYPE_ONE_TIME", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_NAMEID_POLICY_TYPE_FEDERATED);
    PyDict_SetItemString(d, "LIB_NAMEID_POLICY_TYPE_FEDERATED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_NAMEID_POLICY_TYPE_ANY);
    PyDict_SetItemString(d, "LIB_NAMEID_POLICY_TYPE_ANY", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_INTERNET_PROTOCOL);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_INTERNET_PROTOCOL", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_INTERNET_PROTOCOL_PASSWORD);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_INTERNET_PROTOCOL_PASSWORD", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_MOBILE_ONE_FACTOR_UNREGISTERED);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_MOBILE_ONE_FACTOR_UNREGISTERED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_MOBILE_TWO_FACTOR_UNREGISTERED);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_MOBILE_TWO_FACTOR_UNREGISTERED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_MOBILE_ONE_FACTOR_CONTRACT);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_MOBILE_ONE_FACTOR_CONTRACT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_MOBILE_TWO_FACTOR_CONTRACT);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_MOBILE_TWO_FACTOR_CONTRACT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_PASSWORD);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_PASSWORD", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_PASSWORD_PROTECTED_TRANSPORT);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_PASSWORD_PROTECTED_TRANSPORT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_PREVIOUS_SESSION);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_PREVIOUS_SESSION", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_SMARTCARD);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_SMARTCARD", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_SMARTCARD_PKI);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_SMARTCARD_PKI", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_SOFTWARE_PKI);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_SOFTWARE_PKI", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_CLASS_REF_TIME_SYNC_TOKEN);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_CLASS_REF_TIME_SYNC_TOKEN", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_COMPARISON_EXACT);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_COMPARISON_EXACT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_COMPARISON_MINIMUM);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_COMPARISON_MINIMUM", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_COMPARISON_MAXIMUM);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_COMPARISON_MAXIMUM", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_AUTHN_CONTEXT_COMPARISON_BETTER);
    PyDict_SetItemString(d, "LIB_AUTHN_CONTEXT_COMPARISON_BETTER", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_STATUS_CODE_FEDERATION_DOES_NOT_EXIST);
    PyDict_SetItemString(d, "LIB_STATUS_CODE_FEDERATION_DOES_NOT_EXIST", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_STATUS_CODE_INVALID_ASSERTION_CONSUMER_SERVICE_INDEX);
    PyDict_SetItemString(d, "LIB_STATUS_CODE_INVALID_ASSERTION_CONSUMER_SERVICE_INDEX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_STATUS_CODE_INVALID_SIGNATURE);
    PyDict_SetItemString(d, "LIB_STATUS_CODE_INVALID_SIGNATURE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_STATUS_CODE_NO_AUTHN_CONTEXT);
    PyDict_SetItemString(d, "LIB_STATUS_CODE_NO_AUTHN_CONTEXT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_STATUS_CODE_NO_AVAILABLEIDP);
    PyDict_SetItemString(d, "LIB_STATUS_CODE_NO_AVAILABLEIDP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_STATUS_CODE_NO_PASSIVE);
    PyDict_SetItemString(d, "LIB_STATUS_CODE_NO_PASSIVE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_STATUS_CODE_NO_SUPPORTEDIDP);
    PyDict_SetItemString(d, "LIB_STATUS_CODE_NO_SUPPORTEDIDP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_STATUS_CODE_PROXY_COUNT_EXCEEDED);
    PyDict_SetItemString(d, "LIB_STATUS_CODE_PROXY_COUNT_EXCEEDED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_STATUS_CODE_UNKNOWN_PRINCIPAL);
    PyDict_SetItemString(d, "LIB_STATUS_CODE_UNKNOWN_PRINCIPAL", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_STATUS_CODE_UNSIGNED_AUTHN_REQUEST);
    PyDict_SetItemString(d, "LIB_STATUS_CODE_UNSIGNED_AUTHN_REQUEST", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_STATUS_CODE_UNSUPPORTED_PROFILE);
    PyDict_SetItemString(d, "LIB_STATUS_CODE_UNSUPPORTED_PROFILE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_BRWS_ART);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_BRWS_ART", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_BRWS_POST);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_BRWS_POST", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_BRWS_LECP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_BRWS_LECP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_FED_TERM_IDP_HTTP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_FED_TERM_IDP_HTTP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_FED_TERM_IDP_SOAP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_FED_TERM_IDP_SOAP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_FED_TERM_SP_HTTP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_FED_TERM_SP_HTTP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_FED_TERM_SP_SOAP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_FED_TERM_SP_SOAP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_NIM_SP_HTTP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_NIM_SP_HTTP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_RNI_IDP_HTTP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_RNI_IDP_HTTP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_RNI_IDP_SOAP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_RNI_IDP_SOAP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_RNI_SP_HTTP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_RNI_SP_HTTP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_RNI_SP_SOAP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_RNI_SP_SOAP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_SLO_SP_HTTP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_SLO_SP_HTTP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_SLO_SP_SOAP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_SLO_SP_SOAP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_SLO_IDP_HTTP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_SLO_IDP_HTTP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_PROTOCOL_PROFILE_SLO_IDP_SOAP);
    PyDict_SetItemString(d, "LIB_PROTOCOL_PROFILE_SLO_IDP_SOAP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_NAME_IDENTIFIER_FORMAT_FEDERATED);
    PyDict_SetItemString(d, "LIB_NAME_IDENTIFIER_FORMAT_FEDERATED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_NAME_IDENTIFIER_FORMAT_ONE_TIME);
    PyDict_SetItemString(d, "LIB_NAME_IDENTIFIER_FORMAT_ONE_TIME", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_NAME_IDENTIFIER_FORMAT_ENCRYPTED);
    PyDict_SetItemString(d, "LIB_NAME_IDENTIFIER_FORMAT_ENCRYPTED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_NAME_IDENTIFIER_FORMAT_ENTITYID);
    PyDict_SetItemString(d, "LIB_NAME_IDENTIFIER_FORMAT_ENTITYID", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_CONSENT_OBTAINED);
    PyDict_SetItemString(d, "LIB_CONSENT_OBTAINED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_CONSENT_OBTAINED_PRIOR);
    PyDict_SetItemString(d, "LIB_CONSENT_OBTAINED_PRIOR", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_CONSENT_OBTAINED_CURRENT_IMPLICIT);
    PyDict_SetItemString(d, "LIB_CONSENT_OBTAINED_CURRENT_IMPLICIT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_CONSENT_OBTAINED_CURRENT_EXPLICIT);
    PyDict_SetItemString(d, "LIB_CONSENT_OBTAINED_CURRENT_EXPLICIT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_CONSENT_UNAVAILABLE);
    PyDict_SetItemString(d, "LIB_CONSENT_UNAVAILABLE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_LIB_CONSENT_INAPPLICABLE);
    PyDict_SetItemString(d, "LIB_CONSENT_INAPPLICABLE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_METADATA_HREF);
    PyDict_SetItemString(d, "METADATA_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_METADATA_PREFIX);
    PyDict_SetItemString(d, "METADATA_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_ASSERTION_HREF);
    PyDict_SetItemString(d, "SAML_ASSERTION_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_ASSERTION_PREFIX);
    PyDict_SetItemString(d, "SAML_ASSERTION_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_PROTOCOL_HREF);
    PyDict_SetItemString(d, "SAML_PROTOCOL_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_PROTOCOL_PREFIX);
    PyDict_SetItemString(d, "SAML_PROTOCOL_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SAML_MAJOR_VERSION_N);
    PyDict_SetItemString(d, "SAML_MAJOR_VERSION_N", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SAML_MINOR_VERSION_N);
    PyDict_SetItemString(d, "SAML_MINOR_VERSION_N", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_STATUS_CODE_SUCCESS);
    PyDict_SetItemString(d, "SAML_STATUS_CODE_SUCCESS", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_STATUS_CODE_VERSION_MISMATCH);
    PyDict_SetItemString(d, "SAML_STATUS_CODE_VERSION_MISMATCH", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_STATUS_CODE_REQUESTER);
    PyDict_SetItemString(d, "SAML_STATUS_CODE_REQUESTER", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_STATUS_CODE_RESPONDER);
    PyDict_SetItemString(d, "SAML_STATUS_CODE_RESPONDER", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_STATUS_CODE_REQUEST_VERSION_TOO_HIGH);
    PyDict_SetItemString(d, "SAML_STATUS_CODE_REQUEST_VERSION_TOO_HIGH", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_STATUS_CODE_REQUEST_VERSION_TOO_LOW);
    PyDict_SetItemString(d, "SAML_STATUS_CODE_REQUEST_VERSION_TOO_LOW", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_STATUS_CODE_REQUEST_VERSION_DEPRECATED);
    PyDict_SetItemString(d, "SAML_STATUS_CODE_REQUEST_VERSION_DEPRECATED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_STATUS_CODE_TOO_MANY_RESPONSES);
    PyDict_SetItemString(d, "SAML_STATUS_CODE_TOO_MANY_RESPONSES", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_STATUS_CODE_RESOURCE_NOT_RECOGNIZED);
    PyDict_SetItemString(d, "SAML_STATUS_CODE_RESOURCE_NOT_RECOGNIZED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_STATUS_CODE_REQUEST_DENIED);
    PyDict_SetItemString(d, "SAML_STATUS_CODE_REQUEST_DENIED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_AUTHENTICATION_METHOD_PASSWORD);
    PyDict_SetItemString(d, "SAML_AUTHENTICATION_METHOD_PASSWORD", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_AUTHENTICATION_METHOD_KERBEROS);
    PyDict_SetItemString(d, "SAML_AUTHENTICATION_METHOD_KERBEROS", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_AUTHENTICATION_METHOD_SECURE_REMOTE_PASSWORD);
    PyDict_SetItemString(d, "SAML_AUTHENTICATION_METHOD_SECURE_REMOTE_PASSWORD", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_AUTHENTICATION_METHOD_HARDWARE_TOKEN);
    PyDict_SetItemString(d, "SAML_AUTHENTICATION_METHOD_HARDWARE_TOKEN", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_AUTHENTICATION_METHOD_SMARTCARD_PKI);
    PyDict_SetItemString(d, "SAML_AUTHENTICATION_METHOD_SMARTCARD_PKI", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_AUTHENTICATION_METHOD_SOFTWARE_PKI);
    PyDict_SetItemString(d, "SAML_AUTHENTICATION_METHOD_SOFTWARE_PKI", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_AUTHENTICATION_METHOD_PGP);
    PyDict_SetItemString(d, "SAML_AUTHENTICATION_METHOD_PGP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_AUTHENTICATION_METHODS_PKI);
    PyDict_SetItemString(d, "SAML_AUTHENTICATION_METHODS_PKI", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_AUTHENTICATION_METHOD_XKMS);
    PyDict_SetItemString(d, "SAML_AUTHENTICATION_METHOD_XKMS", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_AUTHENTICATION_METHOD_XMLD_SIG);
    PyDict_SetItemString(d, "SAML_AUTHENTICATION_METHOD_XMLD_SIG", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_AUTHENTICATION_METHOD_UNSPECIFIED);
    PyDict_SetItemString(d, "SAML_AUTHENTICATION_METHOD_UNSPECIFIED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_AUTHENTICATION_METHOD_LIBERTY);
    PyDict_SetItemString(d, "SAML_AUTHENTICATION_METHOD_LIBERTY", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_CONFIRMATION_METHOD_ARTIFACT);
    PyDict_SetItemString(d, "SAML_CONFIRMATION_METHOD_ARTIFACT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_CONFIRMATION_METHOD_ARTIFACT01);
    PyDict_SetItemString(d, "SAML_CONFIRMATION_METHOD_ARTIFACT01", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_CONFIRMATION_METHOD_BEARER);
    PyDict_SetItemString(d, "SAML_CONFIRMATION_METHOD_BEARER", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_CONFIRMATION_METHOD_HOLDER_OF_KEY);
    PyDict_SetItemString(d, "SAML_CONFIRMATION_METHOD_HOLDER_OF_KEY", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML_CONFIRMATION_METHOD_SENDER_VOUCHES);
    PyDict_SetItemString(d, "SAML_CONFIRMATION_METHOD_SENDER_VOUCHES", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_PAOS_HREF);
    PyDict_SetItemString(d, "PAOS_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_PAOS_PREFIX);
    PyDict_SetItemString(d, "PAOS_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_ECP_HREF);
    PyDict_SetItemString(d, "ECP_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_ECP_PREFIX);
    PyDict_SetItemString(d, "ECP_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_XSI_HREF);
    PyDict_SetItemString(d, "XSI_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_XSI_PREFIX);
    PyDict_SetItemString(d, "XSI_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MESSAGE_FORMAT_XSCHEMA_ERROR);
    PyDict_SetItemString(d, "MESSAGE_FORMAT_XSCHEMA_ERROR", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MESSAGE_FORMAT_ERROR);
    PyDict_SetItemString(d, "MESSAGE_FORMAT_ERROR", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MESSAGE_FORMAT_UNKNOWN);
    PyDict_SetItemString(d, "MESSAGE_FORMAT_UNKNOWN", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MESSAGE_FORMAT_XML);
    PyDict_SetItemString(d, "MESSAGE_FORMAT_XML", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MESSAGE_FORMAT_BASE64);
    PyDict_SetItemString(d, "MESSAGE_FORMAT_BASE64", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MESSAGE_FORMAT_QUERY);
    PyDict_SetItemString(d, "MESSAGE_FORMAT_QUERY", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_MESSAGE_FORMAT_SOAP);
    PyDict_SetItemString(d, "MESSAGE_FORMAT_SOAP", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SIGNATURE_TYPE_NONE);
    PyDict_SetItemString(d, "SIGNATURE_TYPE_NONE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SIGNATURE_TYPE_SIMPLE);
    PyDict_SetItemString(d, "SIGNATURE_TYPE_SIMPLE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SIGNATURE_TYPE_WITHX509);
    PyDict_SetItemString(d, "SIGNATURE_TYPE_WITHX509", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SIGNATURE_TYPE_LAST);
    PyDict_SetItemString(d, "SIGNATURE_TYPE_LAST", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SIGNATURE_METHOD_NONE);
    PyDict_SetItemString(d, "SIGNATURE_METHOD_NONE", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SIGNATURE_METHOD_RSA_SHA1);
    PyDict_SetItemString(d, "SIGNATURE_METHOD_RSA_SHA1", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SIGNATURE_METHOD_DSA_SHA1);
    PyDict_SetItemString(d, "SIGNATURE_METHOD_DSA_SHA1", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SIGNATURE_METHOD_HMAC_SHA1);
    PyDict_SetItemString(d, "SIGNATURE_METHOD_HMAC_SHA1", obj);
    Py_DECREF(obj);
    obj = PyInt_FromLong(LASSO_SIGNATURE_METHOD_LAST);
    PyDict_SetItemString(d, "SIGNATURE_METHOD_LAST", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_METADATA_HREF);
    PyDict_SetItemString(d, "SAML2_METADATA_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_METADATA_PREFIX);
    PyDict_SetItemString(d, "SAML2_METADATA_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_PROTOCOL_HREF);
    PyDict_SetItemString(d, "SAML2_PROTOCOL_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_PROTOCOL_PREFIX);
    PyDict_SetItemString(d, "SAML2_PROTOCOL_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ASSERTION_HREF);
    PyDict_SetItemString(d, "SAML2_ASSERTION_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ASSERTION_PREFIX);
    PyDict_SetItemString(d, "SAML2_ASSERTION_PREFIX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_METADATA_BINDING_SOAP);
    PyDict_SetItemString(d, "SAML2_METADATA_BINDING_SOAP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_METADATA_BINDING_REDIRECT);
    PyDict_SetItemString(d, "SAML2_METADATA_BINDING_REDIRECT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_METADATA_BINDING_POST);
    PyDict_SetItemString(d, "SAML2_METADATA_BINDING_POST", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_METADATA_BINDING_ARTIFACT);
    PyDict_SetItemString(d, "SAML2_METADATA_BINDING_ARTIFACT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_METADATA_BINDING_PAOS);
    PyDict_SetItemString(d, "SAML2_METADATA_BINDING_PAOS", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_METADATA_BINDING_URI);
    PyDict_SetItemString(d, "SAML2_METADATA_BINDING_URI", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_DEFLATE_ENCODING);
    PyDict_SetItemString(d, "SAML2_DEFLATE_ENCODING", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_NAME_IDENTIFIER_FORMAT_UNSPECIFIED);
    PyDict_SetItemString(d, "SAML2_NAME_IDENTIFIER_FORMAT_UNSPECIFIED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_NAME_IDENTIFIER_FORMAT_EMAIL);
    PyDict_SetItemString(d, "SAML2_NAME_IDENTIFIER_FORMAT_EMAIL", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_NAME_IDENTIFIER_FORMAT_X509);
    PyDict_SetItemString(d, "SAML2_NAME_IDENTIFIER_FORMAT_X509", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_NAME_IDENTIFIER_FORMAT_WINDOWS);
    PyDict_SetItemString(d, "SAML2_NAME_IDENTIFIER_FORMAT_WINDOWS", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_NAME_IDENTIFIER_FORMAT_KERBEROS);
    PyDict_SetItemString(d, "SAML2_NAME_IDENTIFIER_FORMAT_KERBEROS", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_NAME_IDENTIFIER_FORMAT_ENTITY);
    PyDict_SetItemString(d, "SAML2_NAME_IDENTIFIER_FORMAT_ENTITY", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_NAME_IDENTIFIER_FORMAT_PERSISTENT);
    PyDict_SetItemString(d, "SAML2_NAME_IDENTIFIER_FORMAT_PERSISTENT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_NAME_IDENTIFIER_FORMAT_TRANSIENT);
    PyDict_SetItemString(d, "SAML2_NAME_IDENTIFIER_FORMAT_TRANSIENT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_NAME_IDENTIFIER_FORMAT_ENCRYPTED);
    PyDict_SetItemString(d, "SAML2_NAME_IDENTIFIER_FORMAT_ENCRYPTED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ATTRIBUTE_NAME_EPR);
    PyDict_SetItemString(d, "SAML2_ATTRIBUTE_NAME_EPR", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ATTRIBUTE_NAME_FORMAT_UNSPECIFIED);
    PyDict_SetItemString(d, "SAML2_ATTRIBUTE_NAME_FORMAT_UNSPECIFIED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ATTRIBUTE_NAME_FORMAT_URI);
    PyDict_SetItemString(d, "SAML2_ATTRIBUTE_NAME_FORMAT_URI", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ATTRIBUTE_NAME_FORMAT_BASIC);
    PyDict_SetItemString(d, "SAML2_ATTRIBUTE_NAME_FORMAT_BASIC", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_NAMESPACE_RWEDC);
    PyDict_SetItemString(d, "SAML2_ACTION_NAMESPACE_RWEDC", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_NAMESPACE_RWEDC_NEGATION);
    PyDict_SetItemString(d, "SAML2_ACTION_NAMESPACE_RWEDC_NEGATION", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_NAMESPACE_GHPP);
    PyDict_SetItemString(d, "SAML2_ACTION_NAMESPACE_GHPP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_NAMESPACE_UNIX);
    PyDict_SetItemString(d, "SAML2_ACTION_NAMESPACE_UNIX", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_RWEDC_READ);
    PyDict_SetItemString(d, "SAML2_ACTION_RWEDC_READ", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_RWEDC_WRITE);
    PyDict_SetItemString(d, "SAML2_ACTION_RWEDC_WRITE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_RWEDC_EXECUTE);
    PyDict_SetItemString(d, "SAML2_ACTION_RWEDC_EXECUTE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_RWEDC_DELETE);
    PyDict_SetItemString(d, "SAML2_ACTION_RWEDC_DELETE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_RWEDC_CONTROL);
    PyDict_SetItemString(d, "SAML2_ACTION_RWEDC_CONTROL", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_RWEDC_NEGATION);
    PyDict_SetItemString(d, "SAML2_ACTION_RWEDC_NEGATION", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_GHPP_GET);
    PyDict_SetItemString(d, "SAML2_ACTION_GHPP_GET", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_GHPP_HEAD);
    PyDict_SetItemString(d, "SAML2_ACTION_GHPP_HEAD", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_GHPP_PUT);
    PyDict_SetItemString(d, "SAML2_ACTION_GHPP_PUT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ACTION_GHPP_POST);
    PyDict_SetItemString(d, "SAML2_ACTION_GHPP_POST", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_CONSENT_OBTAINED);
    PyDict_SetItemString(d, "SAML2_CONSENT_OBTAINED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_CONSENT_PRIOR);
    PyDict_SetItemString(d, "SAML2_CONSENT_PRIOR", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_CONSENT_IMPLICIT);
    PyDict_SetItemString(d, "SAML2_CONSENT_IMPLICIT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_CONSENT_EXPLICIT);
    PyDict_SetItemString(d, "SAML2_CONSENT_EXPLICIT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_CONSENT_UNAVAILABLE);
    PyDict_SetItemString(d, "SAML2_CONSENT_UNAVAILABLE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_CONSENT_INAPPLICABLE);
    PyDict_SetItemString(d, "SAML2_CONSENT_INAPPLICABLE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_SUCCESS);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_SUCCESS", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_REQUESTER);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_REQUESTER", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_RESPONDER);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_RESPONDER", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_VERSION_MISMATCH);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_VERSION_MISMATCH", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_AUTHN_FAILED);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_AUTHN_FAILED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_INVALID_ATTR_NAME);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_INVALID_ATTR_NAME", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_INVALID_NAME_ID_POLICY);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_INVALID_NAME_ID_POLICY", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_NO_AUTHN_CONTEXT);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_NO_AUTHN_CONTEXT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_NO_AVAILABLE_IDP);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_NO_AVAILABLE_IDP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_NO_PASSIVE);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_NO_PASSIVE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_NO_SUPPORTED_IDP);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_NO_SUPPORTED_IDP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_PARTIAL_LOGOUT);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_PARTIAL_LOGOUT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_PROXY_COUNT_EXCEEDED);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_PROXY_COUNT_EXCEEDED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_REQUEST_DENIED);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_REQUEST_DENIED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_REQUEST_UNSUPPORTED);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_REQUEST_UNSUPPORTED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_REQUEST_VERSION_DEPRECATED);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_REQUEST_VERSION_DEPRECATED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_REQUEST_VERSION_TOO_HIGH);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_REQUEST_VERSION_TOO_HIGH", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_REQUEST_VERSION_TOO_LOW);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_REQUEST_VERSION_TOO_LOW", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_RESOURCE_NOT_RECOGNIZED);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_RESOURCE_NOT_RECOGNIZED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_TOO_MANY_RESPONSES);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_TOO_MANY_RESPONSES", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_UNKNOWN_ATTR_PROFILE);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_UNKNOWN_ATTR_PROFILE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_UNKNOWN_PRINCIPAL);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_UNKNOWN_PRINCIPAL", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_STATUS_CODE_UNSUPPORTED_BINDING);
    PyDict_SetItemString(d, "SAML2_STATUS_CODE_UNSUPPORTED_BINDING", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_AUTHENTICATED_TELEPHONY);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_AUTHENTICATED_TELEPHONY", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_INTERNET_PROTOCOL);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_INTERNET_PROTOCOL", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_INTERNET_PROTOCOL_PASSWORD);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_INTERNET_PROTOCOL_PASSWORD", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_KERBEROS);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_KERBEROS", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_MOBILE_ONE_FACTOR_CONTRACT);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_MOBILE_ONE_FACTOR_CONTRACT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_MOBILE_ONE_FACTOR_UNREGISTERED);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_MOBILE_ONE_FACTOR_UNREGISTERED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_MOBILE_TWO_FACTOR_CONTRACT);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_MOBILE_TWO_FACTOR_CONTRACT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_MOBILE_TWO_FACTOR_UNREGISTERED);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_MOBILE_TWO_FACTOR_UNREGISTERED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_NOMAD_TELEPHONY);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_NOMAD_TELEPHONY", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_PERSONALIZED_TELEPHONY);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_PERSONALIZED_TELEPHONY", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_PGP);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_PGP", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_PASSWORD_PROTECTED_TRANSPORT);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_PASSWORD_PROTECTED_TRANSPORT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_PASSWORD);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_PASSWORD", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_PREVIOUS_SESSION);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_PREVIOUS_SESSION", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_SMARTCARD);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_SMARTCARD", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_SMARTCARD_PKI);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_SMARTCARD_PKI", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_SOFTWARE_PKI);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_SOFTWARE_PKI", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_SPKI);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_SPKI", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_SECURE_REMOTE_PASSWORD);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_SECURE_REMOTE_PASSWORD", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_TLS_CLIENT);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_TLS_CLIENT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_X509);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_X509", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_TELEPHONY);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_TELEPHONY", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_TIME_SYNC_TOKEN);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_TIME_SYNC_TOKEN", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_XMLDSIG);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_XMLDSIG", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_AUTHN_CONTEXT_UNSPECIFIED);
    PyDict_SetItemString(d, "SAML2_AUTHN_CONTEXT_UNSPECIFIED", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_CONFIRMATION_METHOD_BEARER);
    PyDict_SetItemString(d, "SAML2_CONFIRMATION_METHOD_BEARER", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_CONFIRMATION_METHOD_HOLDER_OF_KEY);
    PyDict_SetItemString(d, "SAML2_CONFIRMATION_METHOD_HOLDER_OF_KEY", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_FIELD_ENCODING);
    PyDict_SetItemString(d, "SAML2_FIELD_ENCODING", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_FIELD_RESPONSE);
    PyDict_SetItemString(d, "SAML2_FIELD_RESPONSE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_FIELD_REQUEST);
    PyDict_SetItemString(d, "SAML2_FIELD_REQUEST", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_FIELD_ARTIFACT);
    PyDict_SetItemString(d, "SAML2_FIELD_ARTIFACT", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_FIELD_RELAYSTATE);
    PyDict_SetItemString(d, "SAML2_FIELD_RELAYSTATE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_FIELD_SIGNATURE);
    PyDict_SetItemString(d, "SAML2_FIELD_SIGNATURE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_FIELD_SIGALG);
    PyDict_SetItemString(d, "SAML2_FIELD_SIGALG", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ATTRIBUTE_PROFILE_BASIC);
    PyDict_SetItemString(d, "SAML2_ATTRIBUTE_PROFILE_BASIC", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ATTRIBUTE_PROFILE_X500);
    PyDict_SetItemString(d, "SAML2_ATTRIBUTE_PROFILE_X500", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ATTRIBUTE_PROFILE_UUID);
    PyDict_SetItemString(d, "SAML2_ATTRIBUTE_PROFILE_UUID", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_SAML2_ATTRIBUTE_PROFILE_DCE);
    PyDict_SetItemString(d, "SAML2_ATTRIBUTE_PROFILE_DCE", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_DS_HREF);
    PyDict_SetItemString(d, "DS_HREF", obj);
    Py_DECREF(obj);
    obj = PyString_FromString((char*)LASSO_DS_PREFIX);
    PyDict_SetItemString(d, "DS_PREFIX", obj);
    Py_DECREF(obj);
#ifdef LASSO_WSF_ENABLED
    obj = Py_True;
#else
    obj = Py_False;
#endif
    PyDict_SetItemString(d, "WSF_ENABLED", obj);
    Py_DECREF(obj);
}

static PyObject*
key_new_for_signature_from_base64_string(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* base64_string = NULL;
    char* password = NULL;
    LassoSignatureMethod signature_method;
    char* certificate = NULL;
    LassoKey* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "ssis", &base64_string, &password, &signature_method, &certificate)) return NULL;
    return_value = (LassoKey*) lasso_key_new_for_signature_from_base64_string(base64_string, password, signature_method, certificate);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
key_new_for_signature_from_file(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* filename_or_buffer = NULL;
    char* password = NULL;
    LassoSignatureMethod signature_method;
    char* certificate = NULL;
    LassoKey* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "ssis", &filename_or_buffer, &password, &signature_method, &certificate)) return NULL;
    return_value = (LassoKey*) lasso_key_new_for_signature_from_file(filename_or_buffer, password, signature_method, certificate);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
strError(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    int error_code;
    const char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "i", &error_code)) return NULL;
    return_value = lasso_strerror(error_code);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
init(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = lasso_init();
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
shutdown(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = lasso_shutdown();
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
check_version(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    int major;
    int minor;
    int subminor;
    LassoCheckVersionMode mode = LASSO_CHECK_VERSION_NUMERIC;
    int return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "iii|i", &major, &minor, &subminor, &mode)) return NULL;
    return_value = lasso_check_version(major, minor, subminor, mode);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
set_flag(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* flag = NULL;

    if (! PyArg_ParseTuple(args, "s", &flag)) return NULL;
    lasso_set_flag(flag);
    return noneRef();
}

static PyObject*
federation_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* remote_providerID = NULL;
    LassoFederation* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &remote_providerID)) return NULL;
    return_value = (LassoFederation*) lasso_federation_new(remote_providerID);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
get_request_type_from_soap_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* soap = NULL;
    LassoRequestType return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &soap)) return NULL;
    return_value = lasso_profile_get_request_type_from_soap_msg(soap);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
isLibertyQuery(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* query = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &query)) return NULL;
    return_value = lasso_profile_is_liberty_query(query);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
server_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* metadata = NULL;
    const gchar* private_key = NULL;
    const gchar* private_key_password = NULL;
    const gchar* certificate = NULL;
    LassoServer* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "|zzzz", &metadata, &private_key, &private_key_password, &certificate)) return NULL;
    return_value = (LassoServer*) lasso_server_new(metadata, private_key, private_key_password, certificate);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
server_new_from_buffers(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* metadata = NULL;
    const gchar* private_key_content = NULL;
    const gchar* private_key_password = NULL;
    const gchar* certificate_content = NULL;
    LassoServer* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s|zzz", &metadata, &private_key_content, &private_key_password, &certificate_content)) return NULL;
    return_value = (LassoServer*) lasso_server_new_from_buffers(metadata, private_key_content, private_key_password, certificate_content);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
server_new_from_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* dump = NULL;
    LassoServer* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &dump)) return NULL;
    return_value = (LassoServer*) lasso_server_new_from_dump(dump);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
session_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSession* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSession*) lasso_session_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
session_new_from_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* dump = NULL;
    LassoSession* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &dump)) return NULL;
    return_value = (LassoSession*) lasso_session_new_from_dump(dump);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
provider_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProviderRole role;
    const char* metadata = NULL;
    const char* public_key = NULL;
    const char* ca_cert_chain = NULL;
    LassoProvider* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "is|zz", &role, &metadata, &public_key, &ca_cert_chain)) return NULL;
    return_value = (LassoProvider*) lasso_provider_new(role, metadata, public_key, ca_cert_chain);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
provider_new_from_buffer(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProviderRole role;
    const char* metadata = NULL;
    const char* public_key = NULL;
    const char* ca_cert_chain = NULL;
    LassoProvider* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "is|zz", &role, &metadata, &public_key, &ca_cert_chain)) return NULL;
    return_value = (LassoProvider*) lasso_provider_new_from_buffer(role, metadata, public_key, ca_cert_chain);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
provider_new_from_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* dump = NULL;
    LassoProvider* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &dump)) return NULL;
    return_value = (LassoProvider*) lasso_provider_new_from_dump(dump);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
name_registration_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoNameRegistration* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_server)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNameRegistration*) lasso_name_registration_new(server);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
name_registration_new_from_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    const char* dump = NULL;
    LassoNameRegistration* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_server, &dump)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNameRegistration*) lasso_name_registration_new_from_dump(server, dump);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
login_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoLogin* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_server)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoLogin*) lasso_login_new(server);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
login_new_from_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    const gchar* dump = NULL;
    LassoLogin* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_server, &dump)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoLogin*) lasso_login_new_from_dump(server, dump);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lecp_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoLecp* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_server)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoLecp*) lasso_lecp_new(server);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
identity_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoIdentity* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoIdentity*) lasso_identity_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
identity_new_from_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* dump = NULL;
    LassoIdentity* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &dump)) return NULL;
    return_value = (LassoIdentity*) lasso_identity_new_from_dump(dump);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
logout_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoLogout* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_server)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoLogout*) lasso_logout_new(server);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
logout_new_from_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    const gchar* dump = NULL;
    LassoLogout* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_server, &dump)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoLogout*) lasso_logout_new_from_dump(server, dump);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
name_identifier_mapping_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoNameIdentifierMapping* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_server)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNameIdentifierMapping*) lasso_name_identifier_mapping_new(server);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
defederation_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoDefederation* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_server)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoDefederation*) lasso_defederation_new(server);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
isSamlQuery(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* query = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &query)) return NULL;
    return_value = lasso_profile_is_saml_query(query);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
ecp_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoEcp* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_server)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoEcp*) lasso_ecp_new(server);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
assertion_query_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoAssertionQuery* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_server)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoAssertionQuery*) lasso_assertion_query_new(server);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
name_id_management_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoNameIdManagement* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_server)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNameIdManagement*) lasso_name_id_management_new(server);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
name_id_management_new_from_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    const char* dump = NULL;
    LassoNameIdManagement* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_server, &dump)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNameIdManagement*) lasso_name_id_management_new_from_dump(server, dump);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_name_id_build_persistent(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* id = NULL;
    const char* idpID = NULL;
    const char* providerID = NULL;
    LassoSaml2NameID* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "sss", &id, &idpID, &providerID)) return NULL;
    return_value = lasso_saml2_name_id_build_persistent(id, idpID, providerID);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_encrypted_element_build_encrypted_persistent_name_id(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* id = NULL;
    const char* idpID = NULL;
    const LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoSaml2EncryptedElement* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "ssO", &id, &idpID, &cvt_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (const LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_encrypted_element_build_encrypted_persistent_name_id(id, idpID, provider);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_name_id_new_with_persistent_format(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* id = NULL;
    const char* idpID = NULL;
    const char* providerID = NULL;
    LassoSaml2NameID* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "sss", &id, &idpID, &providerID)) return NULL;
    return_value = (LassoSaml2NameID*) lasso_saml2_name_id_new_with_persistent_format(id, idpID, providerID);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_authn_request_envelope_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibAuthnRequestEnvelope* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoLibAuthnRequestEnvelope*) lasso_lib_authn_request_envelope_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_authn_request_envelope_new_full(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibAuthnRequest* authnRequest = NULL;
    PyGObjectPtr *cvt_authnRequest = NULL;
    char* providerID = NULL;
    char* assertionConsumerServiceURL = NULL;
    LassoLibAuthnRequestEnvelope* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Oss", &cvt_authnRequest, &providerID, &assertionConsumerServiceURL)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_authnRequest, &PyGObjectPtrType)) {
        authnRequest = (LassoLibAuthnRequest*)cvt_authnRequest->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoLibAuthnRequestEnvelope*) lasso_lib_authn_request_envelope_new_full(authnRequest, providerID, assertionConsumerServiceURL);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_authn_response_envelope_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibAuthnResponse* response = NULL;
    PyGObjectPtr *cvt_response = NULL;
    char* assertionConsumerServiceURL = NULL;
    LassoLibAuthnResponseEnvelope* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_response, &assertionConsumerServiceURL)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_response, &PyGObjectPtrType)) {
        response = (LassoLibAuthnResponse*)cvt_response->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoLibAuthnResponseEnvelope*) lasso_lib_authn_response_envelope_new(response, assertionConsumerServiceURL);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_attribute_statement_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlAttributeStatement* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSamlAttributeStatement*) lasso_saml_attribute_statement_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_name_identifier_mapping_request_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_lib_name_identifier_mapping_request_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_name_identifier_mapping_request_new_full(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* providerID = NULL;
    LassoSamlNameIdentifier* nameIdentifier = NULL;
    PyGObjectPtr *cvt_nameIdentifier = NULL;
    const char* targetNamespace = NULL;
    LassoSignatureType sign_type;
    LassoSignatureMethod sign_method;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "sOsii", &providerID, &cvt_nameIdentifier, &targetNamespace, &sign_type, &sign_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_nameIdentifier, &PyGObjectPtrType)) {
        nameIdentifier = (LassoSamlNameIdentifier*)cvt_nameIdentifier->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNode*) lasso_lib_name_identifier_mapping_request_new_full(providerID, nameIdentifier, targetNamespace, sign_type, sign_method);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
build_unique_id(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    unsigned int size;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "i", &size)) return NULL;
    return_value = lasso_build_unique_id(size);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
lib_subject_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibSubject* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoLibSubject*) lasso_lib_subject_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_attribute_value_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlAttributeValue* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSamlAttributeValue*) lasso_saml_attribute_value_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_authentication_statement_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibAuthenticationStatement* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoLibAuthenticationStatement*) lasso_lib_authentication_statement_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_authentication_statement_new_full(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* authenticationMethod = NULL;
    const char* authenticationInstant = NULL;
    const char* reauthenticateOnOrAfter = NULL;
    LassoSamlNameIdentifier* sp_identifier = NULL;
    PyGObjectPtr *cvt_sp_identifier = NULL;
    LassoSamlNameIdentifier* idp_identifier = NULL;
    PyGObjectPtr *cvt_idp_identifier = NULL;
    LassoLibAuthenticationStatement* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s|zzOO", &authenticationMethod, &authenticationInstant, &reauthenticateOnOrAfter, &cvt_sp_identifier, &cvt_idp_identifier)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_sp_identifier, &PyGObjectPtrType)) {
        sp_identifier = (LassoSamlNameIdentifier*)cvt_sp_identifier->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_idp_identifier, &PyGObjectPtrType)) {
        idp_identifier = (LassoSamlNameIdentifier*)cvt_idp_identifier->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoLibAuthenticationStatement*) lasso_lib_authentication_statement_new_full(authenticationMethod, authenticationInstant, reauthenticateOnOrAfter, sp_identifier, idp_identifier);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_subject_statement_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml_subject_statement_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_register_name_identifier_response_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_lib_register_name_identifier_response_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_register_name_identifier_response_new_full(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* providerID = NULL;
    const char* statusCodeValue = NULL;
    LassoLibRegisterNameIdentifierRequest* request = NULL;
    PyGObjectPtr *cvt_request = NULL;
    LassoSignatureType sign_type;
    LassoSignatureMethod sign_method;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "ssOii", &providerID, &statusCodeValue, &cvt_request, &sign_type, &sign_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_request, &PyGObjectPtrType)) {
        request = (LassoLibRegisterNameIdentifierRequest*)cvt_request->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNode*) lasso_lib_register_name_identifier_response_new_full(providerID, statusCodeValue, request, sign_type, sign_method);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_federation_termination_notification_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_lib_federation_termination_notification_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_federation_termination_notification_new_full(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* providerID = NULL;
    LassoSamlNameIdentifier* nameIdentifier = NULL;
    PyGObjectPtr *cvt_nameIdentifier = NULL;
    LassoSignatureType sign_type;
    LassoSignatureMethod sign_method;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "sOii", &providerID, &cvt_nameIdentifier, &sign_type, &sign_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_nameIdentifier, &PyGObjectPtrType)) {
        nameIdentifier = (LassoSamlNameIdentifier*)cvt_nameIdentifier->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNode*) lasso_lib_federation_termination_notification_new_full(providerID, nameIdentifier, sign_type, sign_method);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_authn_context_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_lib_authn_context_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_attribute_designator_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml_attribute_designator_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_request_authn_context_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibRequestAuthnContext* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoLibRequestAuthnContext*) lasso_lib_request_authn_context_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_subject_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml_subject_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_authn_response_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* providerID = NULL;
    LassoLibAuthnRequest* request = NULL;
    PyGObjectPtr *cvt_request = NULL;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "sO", &providerID, &cvt_request)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_request, &PyGObjectPtrType)) {
        request = (LassoLibAuthnRequest*)cvt_request->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNode*) lasso_lib_authn_response_new(providerID, request);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_name_identifier_mapping_response_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_lib_name_identifier_mapping_response_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_name_identifier_mapping_response_new_full(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* provideRID = NULL;
    const char* statusCodeValue = NULL;
    LassoLibNameIdentifierMappingRequest* request = NULL;
    PyGObjectPtr *cvt_request = NULL;
    LassoSignatureType sign_type;
    LassoSignatureMethod sign_method;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "ssOii", &provideRID, &statusCodeValue, &cvt_request, &sign_type, &sign_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_request, &PyGObjectPtrType)) {
        request = (LassoLibNameIdentifierMappingRequest*)cvt_request->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNode*) lasso_lib_name_identifier_mapping_response_new_full(provideRID, statusCodeValue, request, sign_type, sign_method);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_attribute_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlAttribute* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSamlAttribute*) lasso_saml_attribute_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_assertion_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlAssertion* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSamlAssertion*) lasso_saml_assertion_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_idp_list_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_lib_idp_list_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_authn_request_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibAuthnRequest* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoLibAuthnRequest*) lasso_lib_authn_request_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_idp_entry_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_lib_idp_entry_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp_status_code_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlpStatusCode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSamlpStatusCode*) lasso_samlp_status_code_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_advice_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml_advice_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp_response_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp_response_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_conditions_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlConditions* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSamlConditions*) lasso_saml_conditions_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_logout_request_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_lib_logout_request_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_logout_request_new_full(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* providerID = NULL;
    LassoSamlNameIdentifier* nameIdentifier = NULL;
    PyGObjectPtr *cvt_nameIdentifier = NULL;
    LassoSignatureType sign_type;
    LassoSignatureMethod sign_method;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "sOii", &providerID, &cvt_nameIdentifier, &sign_type, &sign_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_nameIdentifier, &PyGObjectPtrType)) {
        nameIdentifier = (LassoSamlNameIdentifier*)cvt_nameIdentifier->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNode*) lasso_lib_logout_request_new_full(providerID, nameIdentifier, sign_type, sign_method);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_assertion_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibAssertion* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoLibAssertion*) lasso_lib_assertion_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_assertion_new_full(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* issuer = NULL;
    const char* requestID = NULL;
    const char* audience = NULL;
    const char* notBefore = NULL;
    const char* notOnOrAfter = NULL;
    LassoLibAssertion* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s|zzss", &issuer, &requestID, &audience, &notBefore, &notOnOrAfter)) return NULL;
    return_value = (LassoLibAssertion*) lasso_lib_assertion_new_full(issuer, requestID, audience, notBefore, notOnOrAfter);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp_request_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp_request_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_subject_locality_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml_subject_locality_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_idp_entries_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_lib_idp_entries_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_logout_response_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_lib_logout_response_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_logout_response_new_full(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* providerID = NULL;
    const char* statusCodeValue = NULL;
    LassoLibLogoutRequest* request = NULL;
    PyGObjectPtr *cvt_request = NULL;
    LassoSignatureType sign_type;
    LassoSignatureMethod sign_method;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "ssOii", &providerID, &statusCodeValue, &cvt_request, &sign_type, &sign_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_request, &PyGObjectPtrType)) {
        request = (LassoLibLogoutRequest*)cvt_request->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNode*) lasso_lib_logout_response_new_full(providerID, statusCodeValue, request, sign_type, sign_method);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_status_response_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_lib_status_response_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_name_identifier_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlNameIdentifier* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSamlNameIdentifier*) lasso_saml_name_identifier_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_name_identifier_new_from_xmlNode(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    xmlNode* xmlnode = NULL;
    PyObject *cvt_xmlnode = NULL;
    LassoSamlNameIdentifier* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_xmlnode)) return NULL;
    xmlnode = get_xml_node_from_pystring(cvt_xmlnode);
    return_value = (LassoSamlNameIdentifier*) lasso_saml_name_identifier_new_from_xmlNode(xmlnode);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    lasso_release_xml_node(xmlnode);
    return return_pyvalue;
}

static PyObject*
misc_text_node_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_misc_text_node_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
misc_text_node_new_with_string(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* content = NULL;
    LassoMiscTextNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &content)) return NULL;
    return_value = (LassoMiscTextNode*) lasso_misc_text_node_new_with_string(content);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
misc_text_node_new_with_xml_node(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    xmlNode* xml_node = NULL;
    PyObject *cvt_xml_node = NULL;
    LassoMiscTextNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_xml_node)) return NULL;
    xml_node = get_xml_node_from_pystring(cvt_xml_node);
    return_value = (LassoMiscTextNode*) lasso_misc_text_node_new_with_xml_node(xml_node);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    lasso_release_xml_node(xml_node);
    return return_pyvalue;
}

static PyObject*
saml_audience_restriction_condition_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlAudienceRestrictionCondition* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSamlAudienceRestrictionCondition*) lasso_saml_audience_restriction_condition_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_audience_restriction_condition_new_full(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* audience = NULL;
    LassoSamlAudienceRestrictionCondition* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &audience)) return NULL;
    return_value = (LassoSamlAudienceRestrictionCondition*) lasso_saml_audience_restriction_condition_new_full(audience);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_subject_confirmation_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlSubjectConfirmation* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSamlSubjectConfirmation*) lasso_saml_subject_confirmation_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_authority_binding_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml_authority_binding_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_scoping_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibScoping* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoLibScoping*) lasso_lib_scoping_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp_status_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlpStatus* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSamlpStatus*) lasso_samlp_status_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_register_name_identifier_request_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_lib_register_name_identifier_request_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
lib_register_name_identifier_request_new_full(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* providerID = NULL;
    LassoSamlNameIdentifier* idpNameIdentifier = NULL;
    PyGObjectPtr *cvt_idpNameIdentifier = NULL;
    LassoSamlNameIdentifier* spNameIdentifier = NULL;
    PyGObjectPtr *cvt_spNameIdentifier = NULL;
    LassoSamlNameIdentifier* oldNameIdentifier = NULL;
    PyGObjectPtr *cvt_oldNameIdentifier = NULL;
    LassoSignatureType sign_type;
    LassoSignatureMethod sign_method;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "sOOOii", &providerID, &cvt_idpNameIdentifier, &cvt_spNameIdentifier, &cvt_oldNameIdentifier, &sign_type, &sign_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_idpNameIdentifier, &PyGObjectPtrType)) {
        idpNameIdentifier = (LassoSamlNameIdentifier*)cvt_idpNameIdentifier->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_spNameIdentifier, &PyGObjectPtrType)) {
        spNameIdentifier = (LassoSamlNameIdentifier*)cvt_spNameIdentifier->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_oldNameIdentifier, &PyGObjectPtrType)) {
        oldNameIdentifier = (LassoSamlNameIdentifier*)cvt_oldNameIdentifier->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoNode*) lasso_lib_register_name_identifier_request_new_full(providerID, idpNameIdentifier, spNameIdentifier, oldNameIdentifier, sign_type, sign_method);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml_authentication_statement_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml_authentication_statement_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
node_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_node_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
node_new_from_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* dump = NULL;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &dump)) return NULL;
    return_value = (LassoNode*) lasso_node_new_from_dump(dump);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
node_new_from_soap(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* soap = NULL;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &soap)) return NULL;
    return_value = (LassoNode*) lasso_node_new_from_soap(soap);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
node_new_from_xmlNode(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    xmlNode* node = NULL;
    PyObject *cvt_node = NULL;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_node)) return NULL;
    node = get_xml_node_from_pystring(cvt_node);
    return_value = (LassoNode*) lasso_node_new_from_xmlNode(node);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    lasso_release_xml_node(node);
    return return_pyvalue;
}

static PyObject*
register_dst_service(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* prefix = NULL;
    const char* href = NULL;

    if (! PyArg_ParseTuple(args, "ss", &prefix, &href)) return NULL;
    lasso_register_dst_service(prefix, href);
    return noneRef();
}

static PyObject*
get_prefix_for_dst_service_href(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* href = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &href)) return NULL;
    return_value = lasso_get_prefix_for_dst_service_href(href);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
register_idwsf2_dst_service(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* prefix = NULL;
    const gchar* href = NULL;

    if (! PyArg_ParseTuple(args, "ss", &prefix, &href)) return NULL;
    lasso_register_idwsf2_dst_service(prefix, href);
    return noneRef();
}

static PyObject*
get_prefix_for_idwsf2_dst_service_href(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* href = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &href)) return NULL;
    return_value = lasso_get_prefix_for_idwsf2_dst_service_href(href);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
soap_fault_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSoapFault* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSoapFault*) lasso_soap_fault_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
soap_fault_new_from_message(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* message = NULL;
    LassoSoapFault* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &message)) return NULL;
    return_value = (LassoSoapFault*) lasso_soap_fault_new_from_message(message);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
soap_fault_new_full(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const char* faultcode = NULL;
    const char* faultstring = NULL;
    LassoSoapFault* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "ss", &faultcode, &faultstring)) return NULL;
    return_value = (LassoSoapFault*) lasso_soap_fault_new_full(faultcode, faultstring);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
soap_envelope_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSoapBody* body = NULL;
    PyGObjectPtr *cvt_body = NULL;
    LassoSoapEnvelope* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_body)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_body, &PyGObjectPtrType)) {
        body = (LassoSoapBody*)cvt_body->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = (LassoSoapEnvelope*) lasso_soap_envelope_new(body);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
soap_envelope_new_from_message(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* message = NULL;
    LassoSoapEnvelope* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &message)) return NULL;
    return_value = (LassoSoapEnvelope*) lasso_soap_envelope_new_from_message(message);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
soap_header_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSoapHeader* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSoapHeader*) lasso_soap_header_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
soap_header_new_from_message(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* message = NULL;
    LassoSoapHeader* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &message)) return NULL;
    return_value = (LassoSoapHeader*) lasso_soap_header_new_from_message(message);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
soap_body_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSoapBody* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSoapBody*) lasso_soap_body_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
soap_body_new_from_message(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* message = NULL;
    LassoSoapBody* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &message)) return NULL;
    return_value = (LassoSoapBody*) lasso_soap_body_new_from_message(message);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
soap_detail_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSoapDetail* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSoapDetail*) lasso_soap_detail_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
soap_detail_new_from_message(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const gchar* message = NULL;
    LassoSoapDetail* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &message)) return NULL;
    return_value = (LassoSoapDetail*) lasso_soap_detail_new_from_message(message);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_attribute_statement_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_attribute_statement_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_subject_confirmation_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_subject_confirmation_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_authz_decision_statement_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_authz_decision_statement_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_idp_list_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_idp_list_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_condition_abstract_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_condition_abstract_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_assertion_id_request_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_assertion_id_request_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_key_info_confirmation_data_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_key_info_confirmation_data_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_authn_context_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_authn_context_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_key_info_confirmation_data_type_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_key_info_confirmation_data_type_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_extensions_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_extensions_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_base_idabstract_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_base_idabstract_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_subject_locality_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_subject_locality_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_authn_statement_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_authn_statement_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_status_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_status_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_request_abstract_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_request_abstract_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_name_id_policy_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_name_id_policy_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_manage_name_id_response_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_manage_name_id_response_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_artifact_response_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_artifact_response_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_authn_request_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_authn_request_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_authn_query_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_authn_query_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_proxy_restriction_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_proxy_restriction_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_subject_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_subject_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_one_time_use_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_one_time_use_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_requested_authn_context_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_requested_authn_context_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_status_response_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_status_response_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_audience_restriction_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_audience_restriction_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_attribute_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_attribute_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_name_id_mapping_request_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_name_id_mapping_request_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_manage_name_id_request_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_manage_name_id_request_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_status_code_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_status_code_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_terminate_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_terminate_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_advice_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_advice_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_action_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_action_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_action_new_with_string(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* content = NULL;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &content)) return NULL;
    return_value = (LassoNode*) lasso_saml2_action_new_with_string(content);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_status_detail_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_status_detail_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_encrypted_element_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_encrypted_element_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_assertion_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_assertion_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_subject_query_abstract_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_subject_query_abstract_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_logout_response_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_logout_response_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_statement_abstract_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_statement_abstract_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_scoping_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_scoping_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_attribute_value_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2AttributeValue* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoSaml2AttributeValue*) lasso_saml2_attribute_value_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_idp_entry_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_idp_entry_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_response_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_response_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_conditions_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_conditions_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_artifact_resolve_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_artifact_resolve_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_logout_request_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_logout_request_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_attribute_query_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_attribute_query_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_evidence_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_evidence_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_name_id_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_name_id_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_name_id_new_with_string(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* content = NULL;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "s", &content)) return NULL;
    return_value = (LassoNode*) lasso_saml2_name_id_new_with_string(content);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_name_id_mapping_response_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_name_id_mapping_response_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_authz_decision_query_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_samlp2_authz_decision_query_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_subject_confirmation_data_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoNode*) lasso_saml2_subject_confirmation_data_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
ds_key_info_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsKeyInfo* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoDsKeyInfo*) lasso_ds_key_info_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
ds_rsa_key_value_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsRsaKeyValue* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoDsRsaKeyValue*) lasso_ds_rsa_key_value_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
ds_x509_data_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsX509Data* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoDsX509Data*) lasso_ds_x509_data_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
ds_key_value_new(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsKeyValue* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "")) return NULL;
    return_value = (LassoDsKeyValue*) lasso_ds_key_value_new();
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
node_cleanup_original_xmlnodes(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_node_cleanup_original_xmlnodes(node);
    return noneRef();
}

static PyObject*
node_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_dump(node);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
node_export_to_base64(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_export_to_base64(node);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
node_export_to_query(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    LassoSignatureMethod sign_method = 1;
    const char* private_key_file = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|iz", &cvt_node, &sign_method, &private_key_file)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_export_to_query(node, sign_method, private_key_file);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
node_export_to_query_with_password(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    LassoSignatureMethod sign_method = 1;
    const char* private_key_file = NULL;
    const char* private_key_file_password = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|izz", &cvt_node, &sign_method, &private_key_file, &private_key_file_password)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_export_to_query_with_password(node, sign_method, private_key_file, private_key_file_password);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
node_export_to_soap(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_export_to_soap(node);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
node_export_to_xml(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_export_to_xml(node);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
node_export_to_paos_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    const char* issuer = NULL;
    const char* responseConsumerURL = NULL;
    const char* relay_state = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Osss", &cvt_node, &issuer, &responseConsumerURL, &relay_state)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_export_to_paos_request(node, issuer, responseConsumerURL, relay_state);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
node_export_to_ecp_soap_response(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    const char* assertionConsumerURL = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_node, &assertionConsumerURL)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_export_to_ecp_soap_response(node, assertionConsumerURL);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
node_get_xmlNode(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    gboolean lasso_dump;
    xmlNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Oi", &cvt_node, &lasso_dump)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_get_xmlNode(node, lasso_dump);
    if (return_value) {
        return_pyvalue = get_pystring_from_xml_node(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_xml_node(return_value);
    return return_pyvalue;
}

static PyObject*
node_get_original_xmlnode(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    xmlNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_get_original_xmlnode(node);
    if (return_value) {
        return_pyvalue = get_pystring_from_xml_node(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
node_set_original_xmlnode(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    xmlNode* xmlnode = NULL;
    PyObject *cvt_xmlnode = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_node, &cvt_xmlnode)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    xmlnode = get_xml_node_from_pystring(cvt_xmlnode);
    lasso_node_set_original_xmlnode(node, xmlnode);
    lasso_release_xml_node(xmlnode);
    return noneRef();
}

static PyObject*
node_set_custom_namespace(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    const char* prefix = NULL;
    const char* href = NULL;

    if (! PyArg_ParseTuple(args, "Oss", &cvt_node, &prefix, &href)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_node_set_custom_namespace(node, prefix, href);
    return noneRef();
}

static PyObject*
node_set_custom_nodename(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    const char* nodename = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_node, &nodename)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_node_set_custom_nodename(node, nodename);
    return noneRef();
}

static PyObject*
node_get_name(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    const char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_get_name(node);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
node_get_namespace(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    const char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_get_namespace(node);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
node_init_from_message(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    const char* message = NULL;
    LassoMessageFormat return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_node, &message)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_init_from_message(node, message);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
node_init_from_query(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    const char* query = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_node, &query)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_init_from_query(node, query);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
node_init_from_xml(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    xmlNode* xmlnode = NULL;
    PyObject *cvt_xmlnode = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_node, &cvt_xmlnode)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    xmlnode = get_xml_node_from_pystring(cvt_xmlnode);
    return_value = lasso_node_init_from_xml(node, xmlnode);
    return_pyvalue = PyInt_FromLong(return_value);
    lasso_release_xml_node(xmlnode);
    return return_pyvalue;
}

static PyObject*
node_debug(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    int level = 10;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|i", &cvt_node, &level)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_node_debug(node, level);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
key_query_verify(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoKey* key = NULL;
    PyGObjectPtr *cvt_key = NULL;
    const char* query = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_key, &query)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_key, &PyGObjectPtrType)) {
        key = (LassoKey*)cvt_key->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_key_query_verify(key, query);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
key_query_sign(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoKey* key = NULL;
    PyGObjectPtr *cvt_key = NULL;
    const char* query = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_key, &query)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_key, &PyGObjectPtrType)) {
        key = (LassoKey*)cvt_key->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_key_query_sign(key, query);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
key_saml2_xml_verify(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoKey* key = NULL;
    PyGObjectPtr *cvt_key = NULL;
    char* id = NULL;
    xmlNode* document = NULL;
    PyObject *cvt_document = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OsO", &cvt_key, &id, &cvt_document)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_key, &PyGObjectPtrType)) {
        key = (LassoKey*)cvt_key->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    document = get_xml_node_from_pystring(cvt_document);
    return_value = lasso_key_saml2_xml_verify(key, id, document);
    return_pyvalue = PyInt_FromLong(return_value);
    lasso_release_xml_node(document);
    return return_pyvalue;
}

static PyObject*
key_saml2_xml_sign(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoKey* key = NULL;
    PyGObjectPtr *cvt_key = NULL;
    const char* id = NULL;
    xmlNode* document = NULL;
    PyObject *cvt_document = NULL;
    xmlNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OsO", &cvt_key, &id, &cvt_document)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_key, &PyGObjectPtrType)) {
        key = (LassoKey*)cvt_key->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    document = get_xml_node_from_pystring(cvt_document);
    return_value = lasso_key_saml2_xml_sign(key, id, document);
    if (return_value) {
        return_pyvalue = get_pystring_from_xml_node(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_xml_node(return_value);
    lasso_release_xml_node(document);
    return return_pyvalue;
}

static PyObject*
Federation_remoteProviderId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoFederation* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoFederation*)cvt_this->obj;
    return_value = this->remote_providerID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Federation_remoteProviderId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoFederation* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoFederation*)cvt_this->obj;
    lasso_assign_string(this->remote_providerID, value);
    return noneRef();
}

static PyObject*
Federation_localNameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoFederation* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoFederation*)cvt_this->obj;
    return_value = this->local_nameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Federation_localNameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoFederation* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoFederation*)cvt_this->obj;
    set_object_field((GObject**)&this->local_nameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
Federation_remoteNameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoFederation* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoFederation*)cvt_this->obj;
    return_value = this->remote_nameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Federation_remoteNameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoFederation* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoFederation*)cvt_this->obj;
    set_object_field((GObject**)&this->remote_nameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
federation_build_local_name_identifier(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoFederation* federation = NULL;
    PyGObjectPtr *cvt_federation = NULL;
    const gchar* nameQualifier = NULL;
    const gchar* format = NULL;
    const gchar* content = NULL;

    if (! PyArg_ParseTuple(args, "Osss", &cvt_federation, &nameQualifier, &format, &content)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_federation, &PyGObjectPtrType)) {
        federation = (LassoFederation*)cvt_federation->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_federation_build_local_name_identifier(federation, nameQualifier, format, content);
    return noneRef();
}

static PyObject*
federation_verify_name_identifier(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoFederation* federation = NULL;
    PyGObjectPtr *cvt_federation = NULL;
    LassoNode* name_identifier = NULL;
    PyGObjectPtr *cvt_name_identifier = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_federation, &cvt_name_identifier)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_federation, &PyGObjectPtrType)) {
        federation = (LassoFederation*)cvt_federation->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_name_identifier, &PyGObjectPtrType)) {
        name_identifier = (LassoNode*)cvt_name_identifier->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_federation_verify_name_identifier(federation, name_identifier);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
Profile_server_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProfile* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    return_value = this->server;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Profile_server_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProfile* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    set_object_field((GObject**)&this->server, cvt_value);
    return noneRef();
}

static PyObject*
Profile_request_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProfile* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    return_value = this->request;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Profile_request_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProfile* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    set_object_field((GObject**)&this->request, cvt_value);
    return noneRef();
}

static PyObject*
Profile_response_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProfile* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    return_value = this->response;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Profile_response_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProfile* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    set_object_field((GObject**)&this->response, cvt_value);
    return noneRef();
}

static PyObject*
Profile_nameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProfile* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    return_value = this->nameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Profile_nameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProfile* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    set_object_field((GObject**)&this->nameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
Profile_remoteProviderId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProfile* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    return_value = this->remote_providerID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Profile_remoteProviderId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProfile* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    lasso_assign_string(this->remote_providerID, value);
    return noneRef();
}

static PyObject*
Profile_msgUrl_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProfile* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    return_value = this->msg_url;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Profile_msgUrl_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProfile* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    lasso_assign_string(this->msg_url, value);
    return noneRef();
}

static PyObject*
Profile_msgBody_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProfile* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    return_value = this->msg_body;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Profile_msgBody_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProfile* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    lasso_assign_string(this->msg_body, value);
    return noneRef();
}

static PyObject*
Profile_msgRelayState_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProfile* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    return_value = this->msg_relayState;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Profile_msgRelayState_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProfile* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoProfile*)cvt_this->obj;
    lasso_assign_string(this->msg_relayState, value);
    return noneRef();
}

static PyObject*
profile_get_identity(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    LassoIdentity* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_profile)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_get_identity(profile);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
profile_get_session(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    LassoSession* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_profile)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_get_session(profile);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
profile_has_dirty_identity(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_profile)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_is_identity_dirty(profile);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
profile_has_dirty_session(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_profile)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_is_session_dirty(profile);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
profile_set_identity_from_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    const gchar* dump = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_profile, &dump)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_set_identity_from_dump(profile, dump);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
profile_set_session_from_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    const gchar* dump = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_profile, &dump)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_set_session_from_dump(profile, dump);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
profile_get_federation_nameIdentifier(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_profile)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_get_nameIdentifier(profile);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
profile_get_artifact(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_profile)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_get_artifact(profile);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
profile_get_artifact_message(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_profile)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_get_artifact_message(profile);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
profile_set_artifact_message(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    const char* message = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_profile, &message)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_profile_set_artifact_message(profile, message);
    return noneRef();
}

static PyObject*
profile_set_signature_hint(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    LassoProfileSignatureHint signature_hint;

    if (! PyArg_ParseTuple(args, "Oi", &cvt_profile, &signature_hint)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_profile_set_signature_hint(profile, signature_hint);
    return noneRef();
}

static PyObject*
profile_get_signature_hint(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    LassoProfileSignatureHint return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_profile)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_get_signature_hint(profile);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
profile_set_soap_fault_response(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    const char* faultcode = NULL;
    const char* faultstring = NULL;
    GList* details = NULL;
    PyObject *cvt_details = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os|zO", &cvt_profile, &faultcode, &faultstring, &cvt_details)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    set_list_of_pygobject(&details, cvt_details);
    return_value = lasso_profile_set_soap_fault_response(profile, faultcode, faultstring, details);
    return_pyvalue = PyInt_FromLong(return_value);
    free_list(&details, (GFunc)g_object_unref);
    return return_pyvalue;
}

static PyObject*
profile_set_signature_verify_hint(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    LassoProfileSignatureVerifyHint signature_verify_hint;

    if (! PyArg_ParseTuple(args, "Oi", &cvt_profile, &signature_verify_hint)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_profile_set_signature_verify_hint(profile, signature_verify_hint);
    return noneRef();
}

static PyObject*
profile_get_signature_verify_hint(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    LassoProfileSignatureVerifyHint return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_profile)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_get_signature_verify_hint(profile);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
profile_sso_role_with(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    const char* remote_provider_id = NULL;
    LassoProviderRole return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_profile, &remote_provider_id)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_sso_role_with(profile, remote_provider_id);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
profile_get_signature_status(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProfile* profile = NULL;
    PyGObjectPtr *cvt_profile = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_profile)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_profile, &PyGObjectPtrType)) {
        profile = (LassoProfile*)cvt_profile->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_profile_get_signature_status(profile);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
Session_assertions_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GHashTable* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSession* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSession*)cvt_this->obj;
    return_value = this->assertions;
    return_pyvalue = get_dict_from_hashtable_of_objects(return_value);
    return return_pyvalue;
}

static PyObject*
Session_assertions_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSession* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSession*)cvt_this->obj;
    set_hashtable_of_pygobject(this->assertions, cvt_value);
    return noneRef();
}

static PyObject*
Session_isDirty_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    gboolean return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSession* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSession*)cvt_this->obj;
    return_value = this->is_dirty;
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
Session_isDirty_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSession* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSession*)cvt_this->obj;
    this->is_dirty = PyInt_AS_LONG(cvt_value) ? TRUE : FALSE;
    return noneRef();
}

static PyObject*
session_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSession* session = NULL;
    PyGObjectPtr *cvt_session = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_session)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_session, &PyGObjectPtrType)) {
        session = (LassoSession*)cvt_session->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_session_dump(session);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
session_get_assertions(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSession* session = NULL;
    PyGObjectPtr *cvt_session = NULL;
    const char* provider_id = NULL;
    GList* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|z", &cvt_session, &provider_id)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_session, &PyGObjectPtrType)) {
        session = (LassoSession*)cvt_session->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_session_get_assertions(session, provider_id);
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
session_get_provider_index(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSession* session = NULL;
    PyGObjectPtr *cvt_session = NULL;
    gint index;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Oi", &cvt_session, &index)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_session, &PyGObjectPtrType)) {
        session = (LassoSession*)cvt_session->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_session_get_provider_index(session, index);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
session_is_empty(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSession* session = NULL;
    PyGObjectPtr *cvt_session = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_session)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_session, &PyGObjectPtrType)) {
        session = (LassoSession*)cvt_session->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_session_is_empty(session);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
session_remove_assertion(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSession* session = NULL;
    PyGObjectPtr *cvt_session = NULL;
    const gchar* providerID = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_session, &providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_session, &PyGObjectPtrType)) {
        session = (LassoSession*)cvt_session->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_session_remove_assertion(session, providerID);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
session_get_assertion(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSession* session = NULL;
    PyGObjectPtr *cvt_session = NULL;
    const gchar* providerID = NULL;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_session, &providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_session, &PyGObjectPtrType)) {
        session = (LassoSession*)cvt_session->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_session_get_assertion(session, providerID);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
session_add_assertion(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSession* session = NULL;
    PyGObjectPtr *cvt_session = NULL;
    const char* providerID = NULL;
    LassoNode* assertion = NULL;
    PyGObjectPtr *cvt_assertion = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OsO", &cvt_session, &providerID, &cvt_assertion)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_session, &PyGObjectPtrType)) {
        session = (LassoSession*)cvt_session->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_assertion, &PyGObjectPtrType)) {
        assertion = (LassoNode*)cvt_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_session_add_assertion(session, providerID, assertion);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
session_get_name_ids(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSession* session = NULL;
    PyGObjectPtr *cvt_session = NULL;
    const gchar* providerID = NULL;
    GList* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_session, &providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_session, &PyGObjectPtrType)) {
        session = (LassoSession*)cvt_session->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_session_get_name_ids(session, providerID);
    return_pyvalue = get_list_of_pygobject(return_value);
    lasso_release_list_of_gobjects(return_value);
    return return_pyvalue;
}

static PyObject*
session_get_session_indexes(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSession* session = NULL;
    PyGObjectPtr *cvt_session = NULL;
    const gchar* providerID = NULL;
    LassoNode* name_id = NULL;
    PyGObjectPtr *cvt_name_id = NULL;
    GList* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OsO", &cvt_session, &providerID, &cvt_name_id)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_session, &PyGObjectPtrType)) {
        session = (LassoSession*)cvt_session->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_name_id, &PyGObjectPtrType)) {
        name_id = (LassoNode*)cvt_name_id->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_session_get_session_indexes(session, providerID, name_id);
    return_pyvalue = get_list_of_strings(return_value);
    lasso_release_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
session_get_assertion_ids(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSession* session = NULL;
    PyGObjectPtr *cvt_session = NULL;
    const gchar* providerID = NULL;
    GList* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_session, &providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_session, &PyGObjectPtrType)) {
        session = (LassoSession*)cvt_session->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_session_get_assertion_ids(session, providerID);
    return_pyvalue = get_list_of_strings(return_value);
    lasso_release_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
Provider_providerId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProvider* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProvider*)cvt_this->obj;
    return_value = this->ProviderID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Provider_providerId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProvider* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoProvider*)cvt_this->obj;
    lasso_assign_string(this->ProviderID, value);
    return noneRef();
}

static PyObject*
Provider_role_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProviderRole return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProvider* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProvider*)cvt_this->obj;
    return_value = this->role;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
Provider_role_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProvider* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoProvider*)cvt_this->obj;
    this->role = value;
    return noneRef();
}

static PyObject*
Provider_metadataFilename_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProvider* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProvider*)cvt_this->obj;
    return_value = this->metadata_filename;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Provider_metadataFilename_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProvider* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoProvider*)cvt_this->obj;
    lasso_assign_string(this->metadata_filename, value);
    return noneRef();
}

static PyObject*
Provider_publicKey_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProvider* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProvider*)cvt_this->obj;
    return_value = this->public_key;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Provider_publicKey_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProvider* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoProvider*)cvt_this->obj;
    lasso_assign_string(this->public_key, value);
    return noneRef();
}

static PyObject*
Provider_caCertChain_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoProvider* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoProvider*)cvt_this->obj;
    return_value = this->ca_cert_chain;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Provider_caCertChain_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoProvider* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoProvider*)cvt_this->obj;
    lasso_assign_string(this->ca_cert_chain, value);
    return noneRef();
}

static PyObject*
provider_get_assertion_consumer_service_url(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    const char* service_id = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|z", &cvt_provider, &service_id)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_assertion_consumer_service_url(provider, service_id);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
provider_get_metadata_one(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    const char* name = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_provider, &name)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_metadata_one(provider, name);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
provider_get_metadata_list(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    const char* name = NULL;
    GList* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_provider, &name)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_metadata_list(provider, name);
    return_pyvalue = get_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
provider_get_first_http_method(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoProvider* remote_provider = NULL;
    PyGObjectPtr *cvt_remote_provider = NULL;
    LassoMdProtocolType protocol_type;
    LassoHttpMethod return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OOi", &cvt_provider, &cvt_remote_provider, &protocol_type)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_remote_provider, &PyGObjectPtrType)) {
        remote_provider = (LassoProvider*)cvt_remote_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_first_http_method(provider, remote_provider, protocol_type);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
provider_accept_http_method(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoProvider* remote_provider = NULL;
    PyGObjectPtr *cvt_remote_provider = NULL;
    LassoMdProtocolType protocol_type;
    LassoHttpMethod http_method;
    gboolean initiate_profile;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OOiii", &cvt_provider, &cvt_remote_provider, &protocol_type, &http_method, &initiate_profile)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_remote_provider, &PyGObjectPtrType)) {
        remote_provider = (LassoProvider*)cvt_remote_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_accept_http_method(provider, remote_provider, protocol_type, http_method, initiate_profile);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
provider_has_protocol_profile(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoMdProtocolType protocol_type;
    const char* protocol_profile = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Ois", &cvt_provider, &protocol_type, &protocol_profile)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_has_protocol_profile(provider, protocol_type, protocol_profile);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
provider_get_base64_succinct_id(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (const LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_base64_succinct_id(provider);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
provider_get_organization(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    xmlNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (const LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_organization(provider);
    if (return_value) {
        return_pyvalue = get_pystring_from_xml_node(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_xml_node(return_value);
    return return_pyvalue;
}

static PyObject*
provider_get_protocol_conformance(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoProtocolConformance return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (const LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_protocol_conformance(provider);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
provider_set_encryption_mode(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoEncryptionMode encryption_mode;

    if (! PyArg_ParseTuple(args, "Oi", &cvt_provider, &encryption_mode)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_provider_set_encryption_mode(provider, encryption_mode);
    return noneRef();
}

static PyObject*
provider_get_encryption_mode(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoEncryptionMode return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_encryption_mode(provider);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
provider_set_encryption_sym_key_type(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoEncryptionSymKeyType encryption_sym_key_type;

    if (! PyArg_ParseTuple(args, "Oi", &cvt_provider, &encryption_sym_key_type)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_provider_set_encryption_sym_key_type(provider, encryption_sym_key_type);
    return noneRef();
}

static PyObject*
provider_get_default_name_id_format(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_default_name_id_format(provider);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
provider_get_sp_name_qualifier(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    const char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_sp_name_qualifier(provider);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
provider_verify_single_node_signature(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;
    const char* id_attr_name = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OOs", &cvt_provider, &cvt_node, &id_attr_name)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_verify_single_node_signature(provider, node, id_attr_name);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
provider_get_idp_supported_attributes(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    GList* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_idp_supported_attributes(provider);
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
provider_get_valid_until(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_valid_until(provider);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
provider_get_cache_duration(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_cache_duration(provider);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
provider_get_metadata_one_for_role(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoProviderRole role;
    const char* name = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Ois", &cvt_provider, &role, &name)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_metadata_one_for_role(provider, role, name);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
provider_get_metadata_list_for_role(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoProviderRole role;
    const char* name = NULL;
    GList* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Ois", &cvt_provider, &role, &name)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (const LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_metadata_list_for_role(provider, role, name);
    return_pyvalue = get_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
provider_get_metadata_keys_for_role(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoProviderRole role;
    GList* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Oi", &cvt_provider, &role)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_metadata_keys_for_role(provider, role);
    return_pyvalue = get_list_of_strings(return_value);
    lasso_release_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
provider_get_roles(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoProviderRole return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_get_roles(provider);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
provider_match_conformance(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoProvider* another_provider = NULL;
    PyGObjectPtr *cvt_another_provider = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_provider, &cvt_another_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_another_provider, &PyGObjectPtrType)) {
        another_provider = (LassoProvider*)cvt_another_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_match_conformance(provider, another_provider);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
provider_set_server_signing_key(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoKey* key = NULL;
    PyGObjectPtr *cvt_key = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_provider, &cvt_key)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_key, &PyGObjectPtrType)) {
        key = (LassoKey*)cvt_key->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_set_server_signing_key(provider, key);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
provider_add_key(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoKey* key = NULL;
    PyGObjectPtr *cvt_key = NULL;
    gboolean after = FALSE;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OO|i", &cvt_provider, &cvt_key, &after)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_key, &PyGObjectPtrType)) {
        key = (LassoKey*)cvt_key->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_add_key(provider, key, after);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
provider_verify_signature(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    const char* message = NULL;
    const char* id_attr_name = NULL;
    LassoMessageFormat format;
    int return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Ossi", &cvt_provider, &message, &id_attr_name, &format)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_verify_signature(provider, message, id_attr_name, format);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
provider_saml2_node_encrypt(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    LassoNode* lasso_node = NULL;
    PyGObjectPtr *cvt_lasso_node = NULL;
    LassoSaml2EncryptedElement* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_provider, &cvt_lasso_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (const LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_lasso_node, &PyGObjectPtrType)) {
        lasso_node = (LassoNode*)cvt_lasso_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_provider_saml2_node_encrypt(provider, lasso_node);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
Server_providers_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GHashTable* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoServer* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoServer*)cvt_this->obj;
    return_value = this->providers;
    return_pyvalue = get_dict_from_hashtable_of_objects(return_value);
    return return_pyvalue;
}

static PyObject*
Server_providers_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoServer* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoServer*)cvt_this->obj;
    set_hashtable_of_pygobject(this->providers, cvt_value);
    return noneRef();
}

static PyObject*
Server_privateKey_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoServer* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoServer*)cvt_this->obj;
    return_value = this->private_key;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Server_privateKey_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoServer* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoServer*)cvt_this->obj;
    lasso_assign_string(this->private_key, value);
    return noneRef();
}

static PyObject*
Server_privateKeyPassword_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoServer* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoServer*)cvt_this->obj;
    return_value = this->private_key_password;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Server_privateKeyPassword_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoServer* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoServer*)cvt_this->obj;
    lasso_assign_string(this->private_key_password, value);
    return noneRef();
}

static PyObject*
Server_certificate_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoServer* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoServer*)cvt_this->obj;
    return_value = this->certificate;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Server_certificate_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoServer* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoServer*)cvt_this->obj;
    lasso_assign_string(this->certificate, value);
    return noneRef();
}

static PyObject*
Server_signatureMethod_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSignatureMethod return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoServer* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoServer*)cvt_this->obj;
    return_value = this->signature_method;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
Server_signatureMethod_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoServer* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoServer*)cvt_this->obj;
    this->signature_method = value;
    return noneRef();
}

static PyObject*
server_add_provider(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoProviderRole role;
    const gchar* metadata = NULL;
    const gchar* public_key = NULL;
    const gchar* ca_cert_chain = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Ois|zz", &cvt_server, &role, &metadata, &public_key, &ca_cert_chain)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_server_add_provider(server, role, metadata, public_key, ca_cert_chain);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
server_add_provider_from_buffer(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoProviderRole role;
    const gchar* metadata = NULL;
    const gchar* public_key = NULL;
    const gchar* ca_cert_chain = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Ois|zz", &cvt_server, &role, &metadata, &public_key, &ca_cert_chain)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_server_add_provider_from_buffer(server, role, metadata, public_key, ca_cert_chain);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
server_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_server)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_server_dump(server);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
server_get_provider(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    const gchar* providerID = NULL;
    LassoProvider* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_server, &providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (const LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_server_get_provider(server, providerID);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
server_set_encryption_private_key(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    const gchar* filename_or_buffer = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|z", &cvt_server, &filename_or_buffer)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_server_set_encryption_private_key(server, filename_or_buffer);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
server_load_affiliation(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    const gchar* filename = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_server, &filename)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_server_load_affiliation(server, filename);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
server_set_encryption_private_key_with_password(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    const gchar* filename_or_buffer = NULL;
    const gchar* password = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|zz", &cvt_server, &filename_or_buffer, &password)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_server_set_encryption_private_key_with_password(server, filename_or_buffer, password);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
server_load_metadata(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoProviderRole role;
    const gchar* federation_file = NULL;
    const gchar* trusted_roots = NULL;
    GList* blacklisted_entity_ids = NULL;
    PyObject *cvt_blacklisted_entity_ids = NULL;
    PyObject *out_pyvalue = NULL;
    GList* loaded_entity_ids = NULL;
    PyObject *cvt_loaded_entity_ids_out = NULL;
    LassoServerLoadMetadataFlag flags;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Ois|zOOi", &cvt_server, &role, &federation_file, &trusted_roots, &cvt_blacklisted_entity_ids, &cvt_loaded_entity_ids_out, &flags)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    set_list_of_strings(&blacklisted_entity_ids, cvt_blacklisted_entity_ids);
    return_value = lasso_server_load_metadata(server, role, federation_file, trusted_roots, blacklisted_entity_ids, &loaded_entity_ids, flags);
    return_pyvalue = PyInt_FromLong(return_value);
    out_pyvalue = get_list_of_strings(loaded_entity_ids);
    PyList_SetItem(cvt_loaded_entity_ids_out, 0, out_pyvalue);
    return return_pyvalue;
}

static PyObject*
server_add_provider2(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoProvider* provider = NULL;
    PyGObjectPtr *cvt_provider = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_server, &cvt_provider)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_provider, &PyGObjectPtrType)) {
        provider = (LassoProvider*)cvt_provider->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_server_add_provider2(server, provider);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
server_saml2_assertion_setup_signature(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_server, &cvt_saml2_assertion)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_server_saml2_assertion_setup_signature(server, saml2_assertion);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
NameRegistration_oldNameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlNameIdentifier* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoNameRegistration* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoNameRegistration*)cvt_this->obj;
    return_value = this->oldNameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
NameRegistration_oldNameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoNameRegistration* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoNameRegistration*)cvt_this->obj;
    set_object_field((GObject**)&this->oldNameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
name_registration_build_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameRegistration* name_registration = NULL;
    PyGObjectPtr *cvt_name_registration = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_name_registration)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_registration, &PyGObjectPtrType)) {
        name_registration = (LassoNameRegistration*)cvt_name_registration->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_registration_build_request_msg(name_registration);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_registration_build_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameRegistration* name_registration = NULL;
    PyGObjectPtr *cvt_name_registration = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_name_registration)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_registration, &PyGObjectPtrType)) {
        name_registration = (LassoNameRegistration*)cvt_name_registration->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_registration_build_response_msg(name_registration);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_registration_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameRegistration* name_registration = NULL;
    PyGObjectPtr *cvt_name_registration = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_name_registration)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_registration, &PyGObjectPtrType)) {
        name_registration = (LassoNameRegistration*)cvt_name_registration->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_registration_dump(name_registration);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
name_registration_init_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameRegistration* name_registration = NULL;
    PyGObjectPtr *cvt_name_registration = NULL;
    char* remote_providerID = NULL;
    LassoHttpMethod http_method = LASSO_HTTP_METHOD_ANY;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os|i", &cvt_name_registration, &remote_providerID, &http_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_registration, &PyGObjectPtrType)) {
        name_registration = (LassoNameRegistration*)cvt_name_registration->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_registration_init_request(name_registration, remote_providerID, http_method);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_registration_process_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameRegistration* name_registration = NULL;
    PyGObjectPtr *cvt_name_registration = NULL;
    gchar* request_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_name_registration, &request_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_registration, &PyGObjectPtrType)) {
        name_registration = (LassoNameRegistration*)cvt_name_registration->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_registration_process_request_msg(name_registration, request_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_registration_process_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameRegistration* name_registration = NULL;
    PyGObjectPtr *cvt_name_registration = NULL;
    gchar* response_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_name_registration, &response_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_registration, &PyGObjectPtrType)) {
        name_registration = (LassoNameRegistration*)cvt_name_registration->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_registration_process_response_msg(name_registration, response_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_registration_validate_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameRegistration* name_registration = NULL;
    PyGObjectPtr *cvt_name_registration = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_name_registration)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_registration, &PyGObjectPtrType)) {
        name_registration = (LassoNameRegistration*)cvt_name_registration->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_registration_validate_request(name_registration);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
Login_protocolProfile_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLoginProtocolProfile return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLogin* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLogin*)cvt_this->obj;
    return_value = this->protocolProfile;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
Login_protocolProfile_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLogin* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoLogin*)cvt_this->obj;
    this->protocolProfile = value;
    return noneRef();
}

static PyObject*
Login_assertionArtifact_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLogin* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLogin*)cvt_this->obj;
    return_value = this->assertionArtifact;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Login_assertionArtifact_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLogin* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLogin*)cvt_this->obj;
    lasso_assign_string(this->assertionArtifact, value);
    return noneRef();
}

static PyObject*
login_accept_sso(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_login)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_accept_sso(login);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_build_artifact_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    LassoHttpMethod http_method;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Oi", &cvt_login, &http_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_build_artifact_msg(login, http_method);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_build_assertion(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    const char* authenticationMethod = NULL;
    const char* authenticationInstant = NULL;
    const char* reauthenticateOnOrAfter = NULL;
    const char* notBefore = NULL;
    const char* notOnOrAfter = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|zzzzz", &cvt_login, &authenticationMethod, &authenticationInstant, &reauthenticateOnOrAfter, &notBefore, &notOnOrAfter)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_build_assertion(login, authenticationMethod, authenticationInstant, reauthenticateOnOrAfter, notBefore, notOnOrAfter);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_build_authn_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_login)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_build_authn_request_msg(login);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_build_authn_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_login)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_build_authn_response_msg(login);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_build_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_login)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_build_request_msg(login);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_build_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    gchar* remote_providerID = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|z", &cvt_login, &remote_providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_build_response_msg(login, remote_providerID);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_login)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_dump(login);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
login_init_authn_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    const gchar* remote_providerID = NULL;
    LassoHttpMethod http_method = LASSO_HTTP_METHOD_REDIRECT;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|zi", &cvt_login, &remote_providerID, &http_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_init_authn_request(login, remote_providerID, http_method);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_init_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    gchar* response_msg = NULL;
    LassoHttpMethod response_http_method = LASSO_HTTP_METHOD_REDIRECT;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os|i", &cvt_login, &response_msg, &response_http_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_init_request(login, response_msg, response_http_method);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_init_idp_initiated_authn_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    const gchar* remote_providerID = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_login, &remote_providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_init_idp_initiated_authn_request(login, remote_providerID);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_must_ask_for_consent(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_login)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_must_ask_for_consent(login);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
login_must_authenticate(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_login)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_must_authenticate(login);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
login_process_authn_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    const char* authn_request_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|z", &cvt_login, &authn_request_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_process_authn_request_msg(login, authn_request_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_process_authn_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    gchar* authn_response_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_login, &authn_response_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_process_authn_response_msg(login, authn_response_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_process_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    gchar* request_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_login, &request_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_process_request_msg(login, request_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_process_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    gchar* response_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_login, &response_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_process_response_msg(login, response_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_validate_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    gboolean authentication_result;
    gboolean is_consent_obtained;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Oii", &cvt_login, &authentication_result, &is_consent_obtained)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_validate_request_msg(login, authentication_result, is_consent_obtained);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_process_paos_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    gchar* msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_login, &msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_process_paos_response_msg(login, msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
login_get_assertion(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogin* login = NULL;
    PyGObjectPtr *cvt_login = NULL;
    LassoNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_login)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_login, &PyGObjectPtrType)) {
        login = (LassoLogin*)cvt_login->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_login_get_assertion(login);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
Lecp_authnRequestEnvelope_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibAuthnRequestEnvelope* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLecp* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLecp*)cvt_this->obj;
    return_value = this->authnRequestEnvelope;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Lecp_authnRequestEnvelope_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLecp* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLecp*)cvt_this->obj;
    set_object_field((GObject**)&this->authnRequestEnvelope, cvt_value);
    return noneRef();
}

static PyObject*
Lecp_authnResponseEnvelope_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibAuthnResponseEnvelope* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLecp* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLecp*)cvt_this->obj;
    return_value = this->authnResponseEnvelope;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Lecp_authnResponseEnvelope_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLecp* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLecp*)cvt_this->obj;
    set_object_field((GObject**)&this->authnResponseEnvelope, cvt_value);
    return noneRef();
}

static PyObject*
Lecp_assertionConsumerServiceURL_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLecp* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLecp*)cvt_this->obj;
    return_value = this->assertionConsumerServiceURL;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Lecp_assertionConsumerServiceURL_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLecp* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLecp*)cvt_this->obj;
    lasso_assign_string(this->assertionConsumerServiceURL, value);
    return noneRef();
}

static PyObject*
lecp_build_authn_request_envelope_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLecp* lecp = NULL;
    PyGObjectPtr *cvt_lecp = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_lecp)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_lecp, &PyGObjectPtrType)) {
        lecp = (LassoLecp*)cvt_lecp->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_lecp_build_authn_request_envelope_msg(lecp);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
lecp_build_authn_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLecp* lecp = NULL;
    PyGObjectPtr *cvt_lecp = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_lecp)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_lecp, &PyGObjectPtrType)) {
        lecp = (LassoLecp*)cvt_lecp->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_lecp_build_authn_request_msg(lecp);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
lecp_build_authn_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLecp* lecp = NULL;
    PyGObjectPtr *cvt_lecp = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_lecp)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_lecp, &PyGObjectPtrType)) {
        lecp = (LassoLecp*)cvt_lecp->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_lecp_build_authn_response_msg(lecp);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
lecp_build_authn_response_envelope_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLecp* lecp = NULL;
    PyGObjectPtr *cvt_lecp = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_lecp)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_lecp, &PyGObjectPtrType)) {
        lecp = (LassoLecp*)cvt_lecp->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_lecp_build_authn_response_envelope_msg(lecp);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
lecp_init_authn_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLecp* lecp = NULL;
    PyGObjectPtr *cvt_lecp = NULL;
    const char* remote_providerID = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|z", &cvt_lecp, &remote_providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_lecp, &PyGObjectPtrType)) {
        lecp = (LassoLecp*)cvt_lecp->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_lecp_init_authn_request(lecp, remote_providerID);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
lecp_process_authn_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLecp* lecp = NULL;
    PyGObjectPtr *cvt_lecp = NULL;
    const char* authn_request_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_lecp, &authn_request_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_lecp, &PyGObjectPtrType)) {
        lecp = (LassoLecp*)cvt_lecp->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_lecp_process_authn_request_msg(lecp, authn_request_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
lecp_process_authn_request_envelope_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLecp* lecp = NULL;
    PyGObjectPtr *cvt_lecp = NULL;
    const char* request_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_lecp, &request_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_lecp, &PyGObjectPtrType)) {
        lecp = (LassoLecp*)cvt_lecp->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_lecp_process_authn_request_envelope_msg(lecp, request_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
lecp_process_authn_response_envelope_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLecp* lecp = NULL;
    PyGObjectPtr *cvt_lecp = NULL;
    const char* response_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_lecp, &response_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_lecp, &PyGObjectPtrType)) {
        lecp = (LassoLecp*)cvt_lecp->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_lecp_process_authn_response_envelope_msg(lecp, response_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
Identity_federations_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GHashTable* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoIdentity* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoIdentity*)cvt_this->obj;
    return_value = this->federations;
    return_pyvalue = get_dict_from_hashtable_of_objects(return_value);
    return return_pyvalue;
}

static PyObject*
Identity_federations_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoIdentity* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoIdentity*)cvt_this->obj;
    set_hashtable_of_pygobject(this->federations, cvt_value);
    return noneRef();
}

static PyObject*
Identity_isDirty_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    gboolean return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoIdentity* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoIdentity*)cvt_this->obj;
    return_value = this->is_dirty;
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
Identity_isDirty_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoIdentity* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoIdentity*)cvt_this->obj;
    this->is_dirty = PyInt_AS_LONG(cvt_value) ? TRUE : FALSE;
    return noneRef();
}

static PyObject*
identity_get_federation(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoIdentity* identity = NULL;
    PyGObjectPtr *cvt_identity = NULL;
    const char* providerID = NULL;
    LassoFederation* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_identity, &providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_identity, &PyGObjectPtrType)) {
        identity = (LassoIdentity*)cvt_identity->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_identity_get_federation(identity, providerID);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
identity_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoIdentity* identity = NULL;
    PyGObjectPtr *cvt_identity = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_identity)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_identity, &PyGObjectPtrType)) {
        identity = (LassoIdentity*)cvt_identity->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_identity_dump(identity);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
logout_build_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogout* logout = NULL;
    PyGObjectPtr *cvt_logout = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_logout)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_logout, &PyGObjectPtrType)) {
        logout = (LassoLogout*)cvt_logout->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_logout_build_request_msg(logout);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
logout_build_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogout* logout = NULL;
    PyGObjectPtr *cvt_logout = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_logout)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_logout, &PyGObjectPtrType)) {
        logout = (LassoLogout*)cvt_logout->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_logout_build_response_msg(logout);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
logout_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogout* logout = NULL;
    PyGObjectPtr *cvt_logout = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_logout)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_logout, &PyGObjectPtrType)) {
        logout = (LassoLogout*)cvt_logout->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_logout_dump(logout);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
logout_get_next_providerID(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogout* logout = NULL;
    PyGObjectPtr *cvt_logout = NULL;
    gchar* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_logout)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_logout, &PyGObjectPtrType)) {
        logout = (LassoLogout*)cvt_logout->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_logout_get_next_providerID(logout);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
logout_init_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogout* logout = NULL;
    PyGObjectPtr *cvt_logout = NULL;
    gchar* remote_providerID = NULL;
    LassoHttpMethod request_method = LASSO_HTTP_METHOD_ANY;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|zi", &cvt_logout, &remote_providerID, &request_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_logout, &PyGObjectPtrType)) {
        logout = (LassoLogout*)cvt_logout->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_logout_init_request(logout, remote_providerID, request_method);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
logout_process_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogout* logout = NULL;
    PyGObjectPtr *cvt_logout = NULL;
    gchar* request_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_logout, &request_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_logout, &PyGObjectPtrType)) {
        logout = (LassoLogout*)cvt_logout->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_logout_process_request_msg(logout, request_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
logout_process_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogout* logout = NULL;
    PyGObjectPtr *cvt_logout = NULL;
    gchar* response_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_logout, &response_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_logout, &PyGObjectPtrType)) {
        logout = (LassoLogout*)cvt_logout->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_logout_process_response_msg(logout, response_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
logout_reset_providerID_index(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogout* logout = NULL;
    PyGObjectPtr *cvt_logout = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_logout)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_logout, &PyGObjectPtrType)) {
        logout = (LassoLogout*)cvt_logout->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_logout_reset_providerID_index(logout);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
logout_validate_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLogout* logout = NULL;
    PyGObjectPtr *cvt_logout = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_logout)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_logout, &PyGObjectPtrType)) {
        logout = (LassoLogout*)cvt_logout->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_logout_validate_request(logout);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
NameIdentifierMapping_targetNameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoNameIdentifierMapping* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoNameIdentifierMapping*)cvt_this->obj;
    return_value = this->targetNameIdentifier;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
NameIdentifierMapping_targetNameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoNameIdentifierMapping* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoNameIdentifierMapping*)cvt_this->obj;
    lasso_assign_string(this->targetNameIdentifier, value);
    return noneRef();
}

static PyObject*
name_identifier_mapping_build_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdentifierMapping* mapping = NULL;
    PyGObjectPtr *cvt_mapping = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_mapping)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_mapping, &PyGObjectPtrType)) {
        mapping = (LassoNameIdentifierMapping*)cvt_mapping->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_identifier_mapping_build_request_msg(mapping);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_identifier_mapping_build_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdentifierMapping* mapping = NULL;
    PyGObjectPtr *cvt_mapping = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_mapping)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_mapping, &PyGObjectPtrType)) {
        mapping = (LassoNameIdentifierMapping*)cvt_mapping->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_identifier_mapping_build_response_msg(mapping);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_identifier_mapping_init_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdentifierMapping* mapping = NULL;
    PyGObjectPtr *cvt_mapping = NULL;
    gchar* targetNamespace = NULL;
    gchar* remote_providerID = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os|z", &cvt_mapping, &targetNamespace, &remote_providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_mapping, &PyGObjectPtrType)) {
        mapping = (LassoNameIdentifierMapping*)cvt_mapping->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_identifier_mapping_init_request(mapping, targetNamespace, remote_providerID);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_identifier_mapping_process_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdentifierMapping* mapping = NULL;
    PyGObjectPtr *cvt_mapping = NULL;
    gchar* request_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_mapping, &request_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_mapping, &PyGObjectPtrType)) {
        mapping = (LassoNameIdentifierMapping*)cvt_mapping->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_identifier_mapping_process_request_msg(mapping, request_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_identifier_mapping_process_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdentifierMapping* mapping = NULL;
    PyGObjectPtr *cvt_mapping = NULL;
    gchar* response_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_mapping, &response_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_mapping, &PyGObjectPtrType)) {
        mapping = (LassoNameIdentifierMapping*)cvt_mapping->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_identifier_mapping_process_response_msg(mapping, response_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_identifier_mapping_validate_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdentifierMapping* mapping = NULL;
    PyGObjectPtr *cvt_mapping = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_mapping)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_mapping, &PyGObjectPtrType)) {
        mapping = (LassoNameIdentifierMapping*)cvt_mapping->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_identifier_mapping_validate_request(mapping);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
defederation_build_notification_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDefederation* defederation = NULL;
    PyGObjectPtr *cvt_defederation = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_defederation)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_defederation, &PyGObjectPtrType)) {
        defederation = (LassoDefederation*)cvt_defederation->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_defederation_build_notification_msg(defederation);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
defederation_init_notification(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDefederation* defederation = NULL;
    PyGObjectPtr *cvt_defederation = NULL;
    gchar* remote_providerID = NULL;
    LassoHttpMethod http_method = LASSO_HTTP_METHOD_ANY;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|zi", &cvt_defederation, &remote_providerID, &http_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_defederation, &PyGObjectPtrType)) {
        defederation = (LassoDefederation*)cvt_defederation->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_defederation_init_notification(defederation, remote_providerID, http_method);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
defederation_process_notification_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDefederation* defederation = NULL;
    PyGObjectPtr *cvt_defederation = NULL;
    gchar* notification_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_defederation, &notification_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_defederation, &PyGObjectPtrType)) {
        defederation = (LassoDefederation*)cvt_defederation->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_defederation_process_notification_msg(defederation, notification_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
defederation_validate_notification(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDefederation* defederation = NULL;
    PyGObjectPtr *cvt_defederation = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_defederation)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_defederation, &PyGObjectPtrType)) {
        defederation = (LassoDefederation*)cvt_defederation->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_defederation_validate_notification(defederation);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
Ecp_assertionConsumerURL_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoEcp* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoEcp*)cvt_this->obj;
    return_value = this->assertionConsumerURL;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Ecp_assertionConsumerURL_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoEcp* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoEcp*)cvt_this->obj;
    lasso_assign_string(this->assertionConsumerURL, value);
    return noneRef();
}

static PyObject*
ecp_process_authn_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoEcp* ecp = NULL;
    PyGObjectPtr *cvt_ecp = NULL;
    const char* authn_request_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_ecp, &authn_request_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_ecp, &PyGObjectPtrType)) {
        ecp = (LassoEcp*)cvt_ecp->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_ecp_process_authn_request_msg(ecp, authn_request_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
ecp_process_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoEcp* ecp = NULL;
    PyGObjectPtr *cvt_ecp = NULL;
    const char* response_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_ecp, &response_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_ecp, &PyGObjectPtrType)) {
        ecp = (LassoEcp*)cvt_ecp->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_ecp_process_response_msg(ecp, response_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
assertion_query_init_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoAssertionQuery* assertion_query = NULL;
    PyGObjectPtr *cvt_assertion_query = NULL;
    char* remote_provider_id = NULL;
    LassoHttpMethod http_method;
    LassoAssertionQueryRequestType query_request_type;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|zii", &cvt_assertion_query, &remote_provider_id, &http_method, &query_request_type)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_assertion_query, &PyGObjectPtrType)) {
        assertion_query = (LassoAssertionQuery*)cvt_assertion_query->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_assertion_query_init_request(assertion_query, remote_provider_id, http_method, query_request_type);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
assertion_query_validate_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoAssertionQuery* assertion_query = NULL;
    PyGObjectPtr *cvt_assertion_query = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_assertion_query)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_assertion_query, &PyGObjectPtrType)) {
        assertion_query = (LassoAssertionQuery*)cvt_assertion_query->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_assertion_query_validate_request(assertion_query);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
assertion_query_build_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoAssertionQuery* assertion_query = NULL;
    PyGObjectPtr *cvt_assertion_query = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_assertion_query)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_assertion_query, &PyGObjectPtrType)) {
        assertion_query = (LassoAssertionQuery*)cvt_assertion_query->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_assertion_query_build_request_msg(assertion_query);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
assertion_query_process_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoAssertionQuery* assertion_query = NULL;
    PyGObjectPtr *cvt_assertion_query = NULL;
    gchar* request_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_assertion_query, &request_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_assertion_query, &PyGObjectPtrType)) {
        assertion_query = (LassoAssertionQuery*)cvt_assertion_query->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_assertion_query_process_request_msg(assertion_query, request_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
assertion_query_build_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoAssertionQuery* assertion_query = NULL;
    PyGObjectPtr *cvt_assertion_query = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_assertion_query)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_assertion_query, &PyGObjectPtrType)) {
        assertion_query = (LassoAssertionQuery*)cvt_assertion_query->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_assertion_query_build_response_msg(assertion_query);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
assertion_query_process_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoAssertionQuery* assertion_query = NULL;
    PyGObjectPtr *cvt_assertion_query = NULL;
    gchar* response_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_assertion_query, &response_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_assertion_query, &PyGObjectPtrType)) {
        assertion_query = (LassoAssertionQuery*)cvt_assertion_query->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_assertion_query_process_response_msg(assertion_query, response_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
assertion_query_add_attribute_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoAssertionQuery* assertion_query = NULL;
    PyGObjectPtr *cvt_assertion_query = NULL;
    char* format = NULL;
    char* name = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Oss", &cvt_assertion_query, &format, &name)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_assertion_query, &PyGObjectPtrType)) {
        assertion_query = (LassoAssertionQuery*)cvt_assertion_query->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_assertion_query_add_attribute_request(assertion_query, format, name);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
assertion_query_get_request_type(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoAssertionQuery* assertion_query = NULL;
    PyGObjectPtr *cvt_assertion_query = NULL;
    LassoAssertionQueryRequestType return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_assertion_query)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_assertion_query, &PyGObjectPtrType)) {
        assertion_query = (LassoAssertionQuery*)cvt_assertion_query->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_assertion_query_get_request_type(assertion_query);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_id_management_dump(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdManagement* name_id_management = NULL;
    PyGObjectPtr *cvt_name_id_management = NULL;
    char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_name_id_management)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_id_management, &PyGObjectPtrType)) {
        name_id_management = (LassoNameIdManagement*)cvt_name_id_management->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_id_management_dump(name_id_management);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    lasso_release_string(return_value);
    return return_pyvalue;
}

static PyObject*
name_id_management_init_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdManagement* name_id_management = NULL;
    PyGObjectPtr *cvt_name_id_management = NULL;
    char* remote_provider_id = NULL;
    char* new_name_id = NULL;
    LassoHttpMethod http_method = LASSO_HTTP_METHOD_ANY;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|zzi", &cvt_name_id_management, &remote_provider_id, &new_name_id, &http_method)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_id_management, &PyGObjectPtrType)) {
        name_id_management = (LassoNameIdManagement*)cvt_name_id_management->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_id_management_init_request(name_id_management, remote_provider_id, new_name_id, http_method);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_id_management_build_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdManagement* name_id_management = NULL;
    PyGObjectPtr *cvt_name_id_management = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_name_id_management)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_id_management, &PyGObjectPtrType)) {
        name_id_management = (LassoNameIdManagement*)cvt_name_id_management->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_id_management_build_request_msg(name_id_management);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_id_management_process_request_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdManagement* name_id_management = NULL;
    PyGObjectPtr *cvt_name_id_management = NULL;
    gchar* request_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_name_id_management, &request_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_id_management, &PyGObjectPtrType)) {
        name_id_management = (LassoNameIdManagement*)cvt_name_id_management->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_id_management_process_request_msg(name_id_management, request_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_id_management_validate_request(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdManagement* name_id_management = NULL;
    PyGObjectPtr *cvt_name_id_management = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_name_id_management)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_id_management, &PyGObjectPtrType)) {
        name_id_management = (LassoNameIdManagement*)cvt_name_id_management->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_id_management_validate_request(name_id_management);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_id_management_build_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdManagement* name_id_management = NULL;
    PyGObjectPtr *cvt_name_id_management = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_name_id_management)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_id_management, &PyGObjectPtrType)) {
        name_id_management = (LassoNameIdManagement*)cvt_name_id_management->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_id_management_build_response_msg(name_id_management);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
name_id_management_process_response_msg(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNameIdManagement* name_id_management = NULL;
    PyGObjectPtr *cvt_name_id_management = NULL;
    gchar* response_msg = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_name_id_management, &response_msg)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_id_management, &PyGObjectPtrType)) {
        name_id_management = (LassoNameIdManagement*)cvt_name_id_management->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_name_id_management_process_response_msg(name_id_management, response_msg);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
LibAuthnRequestEnvelope_extension_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    return_value = this->Extension;
    return_pyvalue = get_list_of_xml_nodes(return_value);
    return return_pyvalue;
}

static PyObject*
LibAuthnRequestEnvelope_extension_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    set_list_of_xml_nodes(&this->Extension, cvt_value);
    return noneRef();
}

static PyObject*
LibAuthnRequestEnvelope_authnRequest_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibAuthnRequest* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    return_value = this->AuthnRequest;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibAuthnRequestEnvelope_authnRequest_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    set_object_field((GObject**)&this->AuthnRequest, cvt_value);
    return noneRef();
}

static PyObject*
LibAuthnRequestEnvelope_providerId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    return_value = this->ProviderID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequestEnvelope_providerId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    lasso_assign_string(this->ProviderID, value);
    return noneRef();
}

static PyObject*
LibAuthnRequestEnvelope_providerName_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    return_value = this->ProviderName;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequestEnvelope_providerName_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    lasso_assign_string(this->ProviderName, value);
    return noneRef();
}

static PyObject*
LibAuthnRequestEnvelope_assertionConsumerServiceUrl_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    return_value = this->AssertionConsumerServiceURL;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequestEnvelope_assertionConsumerServiceUrl_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    lasso_assign_string(this->AssertionConsumerServiceURL, value);
    return noneRef();
}

static PyObject*
LibAuthnRequestEnvelope_idpList_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibIDPList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    return_value = this->IDPList;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibAuthnRequestEnvelope_idpList_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    set_object_field((GObject**)&this->IDPList, cvt_value);
    return noneRef();
}

static PyObject*
LibAuthnRequestEnvelope_isPassive_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    gboolean return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    return_value = this->IsPassive;
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequestEnvelope_isPassive_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequestEnvelope* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthnRequestEnvelope*)cvt_this->obj;
    this->IsPassive = PyInt_AS_LONG(cvt_value) ? TRUE : FALSE;
    return noneRef();
}

static PyObject*
LibAuthnResponseEnvelope_extension_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponseEnvelope* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnResponseEnvelope*)cvt_this->obj;
    return_value = this->Extension;
    return_pyvalue = get_list_of_xml_nodes(return_value);
    return return_pyvalue;
}

static PyObject*
LibAuthnResponseEnvelope_extension_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponseEnvelope* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthnResponseEnvelope*)cvt_this->obj;
    set_list_of_xml_nodes(&this->Extension, cvt_value);
    return noneRef();
}

static PyObject*
LibAuthnResponseEnvelope_authnResponse_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibAuthnResponse* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponseEnvelope* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnResponseEnvelope*)cvt_this->obj;
    return_value = this->AuthnResponse;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibAuthnResponseEnvelope_authnResponse_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponseEnvelope* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthnResponseEnvelope*)cvt_this->obj;
    set_object_field((GObject**)&this->AuthnResponse, cvt_value);
    return noneRef();
}

static PyObject*
LibAuthnResponseEnvelope_assertionConsumerServiceUrl_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponseEnvelope* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnResponseEnvelope*)cvt_this->obj;
    return_value = this->AssertionConsumerServiceURL;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnResponseEnvelope_assertionConsumerServiceUrl_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponseEnvelope* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnResponseEnvelope*)cvt_this->obj;
    lasso_assign_string(this->AssertionConsumerServiceURL, value);
    return noneRef();
}

static PyObject*
SamlAttributeValue_any_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAttributeValue* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAttributeValue*)cvt_this->obj;
    return_value = this->any;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
SamlAttributeValue_any_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAttributeValue* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAttributeValue*)cvt_this->obj;
    set_list_of_pygobject(&this->any, cvt_value);
    return noneRef();
}

static PyObject*
SamlpResponseAbstract_responseId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    return_value = this->ResponseID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpResponseAbstract_responseId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    lasso_assign_string(this->ResponseID, value);
    return noneRef();
}

static PyObject*
SamlpResponseAbstract_inResponseTo_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    return_value = this->InResponseTo;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpResponseAbstract_inResponseTo_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    lasso_assign_string(this->InResponseTo, value);
    return noneRef();
}

static PyObject*
SamlpResponseAbstract_majorVersion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    int return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    return_value = this->MajorVersion;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
SamlpResponseAbstract_majorVersion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    this->MajorVersion = value;
    return noneRef();
}

static PyObject*
SamlpResponseAbstract_minorVersion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    int return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    return_value = this->MinorVersion;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
SamlpResponseAbstract_minorVersion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    this->MinorVersion = value;
    return noneRef();
}

static PyObject*
SamlpResponseAbstract_issueInstant_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    return_value = this->IssueInstant;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpResponseAbstract_issueInstant_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    lasso_assign_string(this->IssueInstant, value);
    return noneRef();
}

static PyObject*
SamlpResponseAbstract_recipient_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    return_value = this->Recipient;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpResponseAbstract_recipient_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    lasso_assign_string(this->Recipient, value);
    return noneRef();
}

static PyObject*
SamlpResponseAbstract_signType_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSignatureType return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    return_value = this->sign_type;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
SamlpResponseAbstract_signType_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    this->sign_type = value;
    return noneRef();
}

static PyObject*
SamlpResponseAbstract_signMethod_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSignatureMethod return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    return_value = this->sign_method;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
SamlpResponseAbstract_signMethod_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    this->sign_method = value;
    return noneRef();
}

static PyObject*
SamlpResponseAbstract_privateKeyFile_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    return_value = this->private_key_file;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpResponseAbstract_privateKeyFile_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    lasso_assign_string(this->private_key_file, value);
    return noneRef();
}

static PyObject*
SamlpResponseAbstract_certificateFile_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    return_value = this->certificate_file;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpResponseAbstract_certificateFile_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpResponseAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpResponseAbstract*)cvt_this->obj;
    lasso_assign_string(this->certificate_file, value);
    return noneRef();
}

static PyObject*
samlp_response_abstract_fill(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlpResponseAbstract* response = NULL;
    PyGObjectPtr *cvt_response = NULL;
    const char* InResponseTo = NULL;
    const char* Recipient = NULL;

    if (! PyArg_ParseTuple(args, "Oss", &cvt_response, &InResponseTo, &Recipient)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_response, &PyGObjectPtrType)) {
        response = (LassoSamlpResponseAbstract*)cvt_response->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_samlp_response_abstract_fill(response, InResponseTo, Recipient);
    return noneRef();
}

static PyObject*
LibAuthnContext_authnContextClassRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnContext* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnContext*)cvt_this->obj;
    return_value = this->AuthnContextClassRef;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnContext_authnContextClassRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnContext* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnContext*)cvt_this->obj;
    lasso_assign_string(this->AuthnContextClassRef, value);
    return noneRef();
}

static PyObject*
LibAuthnContext_authnContextStatementRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnContext* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnContext*)cvt_this->obj;
    return_value = this->AuthnContextStatementRef;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnContext_authnContextStatementRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnContext* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnContext*)cvt_this->obj;
    lasso_assign_string(this->AuthnContextStatementRef, value);
    return noneRef();
}

static PyObject*
SamlAttributeDesignator_attributeName_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAttributeDesignator* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAttributeDesignator*)cvt_this->obj;
    return_value = this->AttributeName;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAttributeDesignator_attributeName_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAttributeDesignator* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAttributeDesignator*)cvt_this->obj;
    lasso_assign_string(this->AttributeName, value);
    return noneRef();
}

static PyObject*
SamlAttributeDesignator_attributeNamespace_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAttributeDesignator* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAttributeDesignator*)cvt_this->obj;
    return_value = this->AttributeNamespace;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAttributeDesignator_attributeNamespace_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAttributeDesignator* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAttributeDesignator*)cvt_this->obj;
    lasso_assign_string(this->AttributeNamespace, value);
    return noneRef();
}

static PyObject*
LibRequestAuthnContext_authnContextClassRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibRequestAuthnContext* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibRequestAuthnContext*)cvt_this->obj;
    return_value = this->AuthnContextClassRef;
    return_pyvalue = get_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
LibRequestAuthnContext_authnContextClassRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibRequestAuthnContext* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibRequestAuthnContext*)cvt_this->obj;
    set_list_of_strings(&this->AuthnContextClassRef, cvt_value);
    return noneRef();
}

static PyObject*
LibRequestAuthnContext_authnContextStatementRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibRequestAuthnContext* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibRequestAuthnContext*)cvt_this->obj;
    return_value = this->AuthnContextStatementRef;
    return_pyvalue = get_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
LibRequestAuthnContext_authnContextStatementRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibRequestAuthnContext* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibRequestAuthnContext*)cvt_this->obj;
    set_list_of_strings(&this->AuthnContextStatementRef, cvt_value);
    return noneRef();
}

static PyObject*
LibRequestAuthnContext_authnContextComparison_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibRequestAuthnContext* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibRequestAuthnContext*)cvt_this->obj;
    return_value = this->AuthnContextComparison;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibRequestAuthnContext_authnContextComparison_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibRequestAuthnContext* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibRequestAuthnContext*)cvt_this->obj;
    lasso_assign_string(this->AuthnContextComparison, value);
    return noneRef();
}

static PyObject*
SamlSubject_nameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlNameIdentifier* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlSubject* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlSubject*)cvt_this->obj;
    return_value = this->NameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlSubject_nameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlSubject* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlSubject*)cvt_this->obj;
    set_object_field((GObject**)&this->NameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
SamlSubject_subjectConfirmation_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlSubjectConfirmation* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlSubject* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlSubject*)cvt_this->obj;
    return_value = this->SubjectConfirmation;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlSubject_subjectConfirmation_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlSubject* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlSubject*)cvt_this->obj;
    set_object_field((GObject**)&this->SubjectConfirmation, cvt_value);
    return noneRef();
}

static PyObject*
SamlSubject_encryptedNameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2EncryptedElement* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlSubject* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlSubject*)cvt_this->obj;
    return_value = this->EncryptedNameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlSubject_encryptedNameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlSubject* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlSubject*)cvt_this->obj;
    set_object_field((GObject**)&this->EncryptedNameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
LibSubject_idpProvidedNameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlNameIdentifier* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibSubject* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibSubject*)cvt_this->obj;
    return_value = this->IDPProvidedNameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibSubject_idpProvidedNameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibSubject* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibSubject*)cvt_this->obj;
    set_object_field((GObject**)&this->IDPProvidedNameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
LibNameIdentifierMappingResponse_extension_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibNameIdentifierMappingResponse*)cvt_this->obj;
    return_value = this->Extension;
    return_pyvalue = get_list_of_xml_nodes(return_value);
    return return_pyvalue;
}

static PyObject*
LibNameIdentifierMappingResponse_extension_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingResponse* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibNameIdentifierMappingResponse*)cvt_this->obj;
    set_list_of_xml_nodes(&this->Extension, cvt_value);
    return noneRef();
}

static PyObject*
LibNameIdentifierMappingResponse_providerId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibNameIdentifierMappingResponse*)cvt_this->obj;
    return_value = this->ProviderID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibNameIdentifierMappingResponse_providerId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingResponse* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibNameIdentifierMappingResponse*)cvt_this->obj;
    lasso_assign_string(this->ProviderID, value);
    return noneRef();
}

static PyObject*
LibNameIdentifierMappingResponse_status_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlpStatus* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibNameIdentifierMappingResponse*)cvt_this->obj;
    return_value = this->Status;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibNameIdentifierMappingResponse_status_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingResponse* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibNameIdentifierMappingResponse*)cvt_this->obj;
    set_object_field((GObject**)&this->Status, cvt_value);
    return noneRef();
}

static PyObject*
LibNameIdentifierMappingResponse_nameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlNameIdentifier* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibNameIdentifierMappingResponse*)cvt_this->obj;
    return_value = this->NameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibNameIdentifierMappingResponse_nameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingResponse* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibNameIdentifierMappingResponse*)cvt_this->obj;
    set_object_field((GObject**)&this->NameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
SamlAttribute_attributeName_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAttribute* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAttribute*)cvt_this->obj;
    return_value = this->attributeName;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAttribute_attributeName_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAttribute* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAttribute*)cvt_this->obj;
    lasso_assign_string(this->attributeName, value);
    return noneRef();
}

static PyObject*
SamlAttribute_attributeNameSpace_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAttribute* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAttribute*)cvt_this->obj;
    return_value = this->attributeNameSpace;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAttribute_attributeNameSpace_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAttribute* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAttribute*)cvt_this->obj;
    lasso_assign_string(this->attributeNameSpace, value);
    return noneRef();
}

static PyObject*
SamlAttribute_attributeValue_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAttribute* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAttribute*)cvt_this->obj;
    return_value = this->AttributeValue;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
SamlAttribute_attributeValue_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAttribute* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAttribute*)cvt_this->obj;
    set_list_of_pygobject(&this->AttributeValue, cvt_value);
    return noneRef();
}

static PyObject*
SamlAssertion_conditions_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlConditions* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->Conditions;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlAssertion_conditions_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    set_object_field((GObject**)&this->Conditions, cvt_value);
    return noneRef();
}

static PyObject*
SamlAssertion_advice_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlAdvice* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->Advice;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlAssertion_advice_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    set_object_field((GObject**)&this->Advice, cvt_value);
    return noneRef();
}

static PyObject*
SamlAssertion_subjectStatement_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlSubjectStatement* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->SubjectStatement;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlAssertion_subjectStatement_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    set_object_field((GObject**)&this->SubjectStatement, cvt_value);
    return noneRef();
}

static PyObject*
SamlAssertion_authenticationStatement_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlAuthenticationStatement* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->AuthenticationStatement;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlAssertion_authenticationStatement_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    set_object_field((GObject**)&this->AuthenticationStatement, cvt_value);
    return noneRef();
}

static PyObject*
SamlAssertion_attributeStatement_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlAttributeStatement* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->AttributeStatement;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlAssertion_attributeStatement_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    set_object_field((GObject**)&this->AttributeStatement, cvt_value);
    return noneRef();
}

static PyObject*
SamlAssertion_majorVersion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    int return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->MajorVersion;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
SamlAssertion_majorVersion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    this->MajorVersion = value;
    return noneRef();
}

static PyObject*
SamlAssertion_minorVersion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    int return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->MinorVersion;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
SamlAssertion_minorVersion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    this->MinorVersion = value;
    return noneRef();
}

static PyObject*
SamlAssertion_assertionId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->AssertionID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAssertion_assertionId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    lasso_assign_string(this->AssertionID, value);
    return noneRef();
}

static PyObject*
SamlAssertion_issuer_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->Issuer;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAssertion_issuer_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    lasso_assign_string(this->Issuer, value);
    return noneRef();
}

static PyObject*
SamlAssertion_issueInstant_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->IssueInstant;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAssertion_issueInstant_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    lasso_assign_string(this->IssueInstant, value);
    return noneRef();
}

static PyObject*
SamlAssertion_signType_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSignatureType return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->sign_type;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
SamlAssertion_signType_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    this->sign_type = value;
    return noneRef();
}

static PyObject*
SamlAssertion_signMethod_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSignatureMethod return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->sign_method;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
SamlAssertion_signMethod_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    this->sign_method = value;
    return noneRef();
}

static PyObject*
SamlAssertion_privateKeyFile_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->private_key_file;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAssertion_privateKeyFile_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    lasso_assign_string(this->private_key_file, value);
    return noneRef();
}

static PyObject*
SamlAssertion_certificateFile_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    return_value = this->certificate_file;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAssertion_certificateFile_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAssertion* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAssertion*)cvt_this->obj;
    lasso_assign_string(this->certificate_file, value);
    return noneRef();
}

static PyObject*
LibIDPList_idpEntries_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibIDPEntries* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibIDPList* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibIDPList*)cvt_this->obj;
    return_value = this->IDPEntries;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibIDPList_idpEntries_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibIDPList* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibIDPList*)cvt_this->obj;
    set_object_field((GObject**)&this->IDPEntries, cvt_value);
    return noneRef();
}

static PyObject*
LibIDPList_getComplete_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibIDPList* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibIDPList*)cvt_this->obj;
    return_value = this->GetComplete;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibIDPList_getComplete_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibIDPList* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibIDPList*)cvt_this->obj;
    lasso_assign_string(this->GetComplete, value);
    return noneRef();
}

static PyObject*
LibIDPEntry_providerId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibIDPEntry* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibIDPEntry*)cvt_this->obj;
    return_value = this->ProviderID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibIDPEntry_providerId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibIDPEntry* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibIDPEntry*)cvt_this->obj;
    lasso_assign_string(this->ProviderID, value);
    return noneRef();
}

static PyObject*
LibIDPEntry_providerName_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibIDPEntry* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibIDPEntry*)cvt_this->obj;
    return_value = this->ProviderName;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibIDPEntry_providerName_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibIDPEntry* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibIDPEntry*)cvt_this->obj;
    lasso_assign_string(this->ProviderName, value);
    return noneRef();
}

static PyObject*
LibIDPEntry_loc_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibIDPEntry* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibIDPEntry*)cvt_this->obj;
    return_value = this->Loc;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibIDPEntry_loc_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibIDPEntry* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibIDPEntry*)cvt_this->obj;
    lasso_assign_string(this->Loc, value);
    return noneRef();
}

static PyObject*
SamlpStatusCode_statusCode_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlpStatusCode* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpStatusCode* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpStatusCode*)cvt_this->obj;
    return_value = this->StatusCode;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlpStatusCode_statusCode_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpStatusCode* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlpStatusCode*)cvt_this->obj;
    set_object_field((GObject**)&this->StatusCode, cvt_value);
    return noneRef();
}

static PyObject*
SamlpStatusCode_value_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpStatusCode* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpStatusCode*)cvt_this->obj;
    return_value = this->Value;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpStatusCode_value_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpStatusCode* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpStatusCode*)cvt_this->obj;
    lasso_assign_string(this->Value, value);
    return noneRef();
}

static PyObject*
SamlAdvice_assertionIdReference_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAdvice* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAdvice*)cvt_this->obj;
    return_value = this->AssertionIDReference;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
SamlAdvice_assertionIdReference_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAdvice* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAdvice*)cvt_this->obj;
    set_list_of_pygobject(&this->AssertionIDReference, cvt_value);
    return noneRef();
}

static PyObject*
SamlAdvice_assertion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlAssertion* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAdvice* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAdvice*)cvt_this->obj;
    return_value = this->Assertion;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlAdvice_assertion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAdvice* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAdvice*)cvt_this->obj;
    set_object_field((GObject**)&this->Assertion, cvt_value);
    return noneRef();
}

static PyObject*
SamlpResponse_status_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlpStatus* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpResponse*)cvt_this->obj;
    return_value = this->Status;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlpResponse_status_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpResponse* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlpResponse*)cvt_this->obj;
    set_object_field((GObject**)&this->Status, cvt_value);
    return noneRef();
}

static PyObject*
SamlpResponse_assertion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpResponse*)cvt_this->obj;
    return_value = this->Assertion;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
SamlpResponse_assertion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpResponse* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlpResponse*)cvt_this->obj;
    set_list_of_pygobject(&this->Assertion, cvt_value);
    return noneRef();
}

static PyObject*
LibAuthnResponse_extension_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnResponse*)cvt_this->obj;
    return_value = this->Extension;
    return_pyvalue = get_list_of_xml_nodes(return_value);
    return return_pyvalue;
}

static PyObject*
LibAuthnResponse_extension_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponse* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthnResponse*)cvt_this->obj;
    set_list_of_xml_nodes(&this->Extension, cvt_value);
    return noneRef();
}

static PyObject*
LibAuthnResponse_providerId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnResponse*)cvt_this->obj;
    return_value = this->ProviderID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnResponse_providerId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponse* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnResponse*)cvt_this->obj;
    lasso_assign_string(this->ProviderID, value);
    return noneRef();
}

static PyObject*
LibAuthnResponse_relayState_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnResponse*)cvt_this->obj;
    return_value = this->RelayState;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnResponse_relayState_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponse* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnResponse*)cvt_this->obj;
    lasso_assign_string(this->RelayState, value);
    return noneRef();
}

static PyObject*
LibAuthnResponse_consent_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnResponse*)cvt_this->obj;
    return_value = this->consent;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnResponse_consent_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnResponse* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnResponse*)cvt_this->obj;
    lasso_assign_string(this->consent, value);
    return noneRef();
}

static PyObject*
SamlConditions_condition_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlConditions* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlConditions*)cvt_this->obj;
    return_value = this->Condition;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
SamlConditions_condition_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlConditions* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlConditions*)cvt_this->obj;
    set_list_of_pygobject(&this->Condition, cvt_value);
    return noneRef();
}

static PyObject*
SamlConditions_audienceRestrictionCondition_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlConditions* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlConditions*)cvt_this->obj;
    return_value = this->AudienceRestrictionCondition;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
SamlConditions_audienceRestrictionCondition_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlConditions* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlConditions*)cvt_this->obj;
    set_list_of_pygobject(&this->AudienceRestrictionCondition, cvt_value);
    return noneRef();
}

static PyObject*
SamlConditions_notBefore_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlConditions* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlConditions*)cvt_this->obj;
    return_value = this->NotBefore;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlConditions_notBefore_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlConditions* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlConditions*)cvt_this->obj;
    lasso_assign_string(this->NotBefore, value);
    return noneRef();
}

static PyObject*
SamlConditions_notOnOrAfter_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlConditions* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlConditions*)cvt_this->obj;
    return_value = this->NotOnOrAfter;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlConditions_notOnOrAfter_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlConditions* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlConditions*)cvt_this->obj;
    lasso_assign_string(this->NotOnOrAfter, value);
    return noneRef();
}

static PyObject*
SamlpRequestAbstract_respondWith_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    return_value = this->RespondWith;
    return_pyvalue = get_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
SamlpRequestAbstract_respondWith_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    set_list_of_strings(&this->RespondWith, cvt_value);
    return noneRef();
}

static PyObject*
SamlpRequestAbstract_requestId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    return_value = this->RequestID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpRequestAbstract_requestId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    lasso_assign_string(this->RequestID, value);
    return noneRef();
}

static PyObject*
SamlpRequestAbstract_majorVersion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    int return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    return_value = this->MajorVersion;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
SamlpRequestAbstract_majorVersion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    this->MajorVersion = value;
    return noneRef();
}

static PyObject*
SamlpRequestAbstract_minorVersion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    int return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    return_value = this->MinorVersion;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
SamlpRequestAbstract_minorVersion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    this->MinorVersion = value;
    return noneRef();
}

static PyObject*
SamlpRequestAbstract_issueInstant_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    return_value = this->IssueInstant;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpRequestAbstract_issueInstant_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    lasso_assign_string(this->IssueInstant, value);
    return noneRef();
}

static PyObject*
SamlpRequestAbstract_signType_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSignatureType return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    return_value = this->sign_type;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
SamlpRequestAbstract_signType_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    this->sign_type = value;
    return noneRef();
}

static PyObject*
SamlpRequestAbstract_signMethod_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSignatureMethod return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    return_value = this->sign_method;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
SamlpRequestAbstract_signMethod_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    this->sign_method = value;
    return noneRef();
}

static PyObject*
SamlpRequestAbstract_privateKeyFile_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    return_value = this->private_key_file;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpRequestAbstract_privateKeyFile_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    lasso_assign_string(this->private_key_file, value);
    return noneRef();
}

static PyObject*
SamlpRequestAbstract_certificateFile_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    return_value = this->certificate_file;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpRequestAbstract_certificateFile_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpRequestAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpRequestAbstract*)cvt_this->obj;
    lasso_assign_string(this->certificate_file, value);
    return noneRef();
}

static PyObject*
LibNameIdentifierMappingRequest_extension_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibNameIdentifierMappingRequest*)cvt_this->obj;
    return_value = this->Extension;
    return_pyvalue = get_list_of_xml_nodes(return_value);
    return return_pyvalue;
}

static PyObject*
LibNameIdentifierMappingRequest_extension_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingRequest* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibNameIdentifierMappingRequest*)cvt_this->obj;
    set_list_of_xml_nodes(&this->Extension, cvt_value);
    return noneRef();
}

static PyObject*
LibNameIdentifierMappingRequest_providerId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibNameIdentifierMappingRequest*)cvt_this->obj;
    return_value = this->ProviderID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibNameIdentifierMappingRequest_providerId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibNameIdentifierMappingRequest*)cvt_this->obj;
    lasso_assign_string(this->ProviderID, value);
    return noneRef();
}

static PyObject*
LibNameIdentifierMappingRequest_nameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlNameIdentifier* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibNameIdentifierMappingRequest*)cvt_this->obj;
    return_value = this->NameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibNameIdentifierMappingRequest_nameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibNameIdentifierMappingRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->NameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
LibNameIdentifierMappingRequest_targetNamespace_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibNameIdentifierMappingRequest*)cvt_this->obj;
    return_value = this->TargetNamespace;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibNameIdentifierMappingRequest_targetNamespace_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibNameIdentifierMappingRequest*)cvt_this->obj;
    lasso_assign_string(this->TargetNamespace, value);
    return noneRef();
}

static PyObject*
LibNameIdentifierMappingRequest_consent_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibNameIdentifierMappingRequest*)cvt_this->obj;
    return_value = this->consent;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibNameIdentifierMappingRequest_consent_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibNameIdentifierMappingRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibNameIdentifierMappingRequest*)cvt_this->obj;
    lasso_assign_string(this->consent, value);
    return noneRef();
}

static PyObject*
LibFederationTerminationNotification_extension_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibFederationTerminationNotification* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibFederationTerminationNotification*)cvt_this->obj;
    return_value = this->Extension;
    return_pyvalue = get_list_of_xml_nodes(return_value);
    return return_pyvalue;
}

static PyObject*
LibFederationTerminationNotification_extension_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibFederationTerminationNotification* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibFederationTerminationNotification*)cvt_this->obj;
    set_list_of_xml_nodes(&this->Extension, cvt_value);
    return noneRef();
}

static PyObject*
LibFederationTerminationNotification_providerId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibFederationTerminationNotification* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibFederationTerminationNotification*)cvt_this->obj;
    return_value = this->ProviderID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibFederationTerminationNotification_providerId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibFederationTerminationNotification* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibFederationTerminationNotification*)cvt_this->obj;
    lasso_assign_string(this->ProviderID, value);
    return noneRef();
}

static PyObject*
LibFederationTerminationNotification_nameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlNameIdentifier* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibFederationTerminationNotification* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibFederationTerminationNotification*)cvt_this->obj;
    return_value = this->NameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibFederationTerminationNotification_nameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibFederationTerminationNotification* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibFederationTerminationNotification*)cvt_this->obj;
    set_object_field((GObject**)&this->NameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
LibFederationTerminationNotification_consent_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibFederationTerminationNotification* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibFederationTerminationNotification*)cvt_this->obj;
    return_value = this->consent;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibFederationTerminationNotification_consent_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibFederationTerminationNotification* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibFederationTerminationNotification*)cvt_this->obj;
    lasso_assign_string(this->consent, value);
    return noneRef();
}

static PyObject*
LibFederationTerminationNotification_relayState_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibFederationTerminationNotification* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibFederationTerminationNotification*)cvt_this->obj;
    return_value = this->RelayState;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibFederationTerminationNotification_relayState_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibFederationTerminationNotification* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibFederationTerminationNotification*)cvt_this->obj;
    lasso_assign_string(this->RelayState, value);
    return noneRef();
}

static PyObject*
LibAuthnRequest_extension_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    return_value = this->Extension;
    return_pyvalue = get_list_of_xml_nodes(return_value);
    return return_pyvalue;
}

static PyObject*
LibAuthnRequest_extension_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    set_list_of_xml_nodes(&this->Extension, cvt_value);
    return noneRef();
}

static PyObject*
LibAuthnRequest_providerId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    return_value = this->ProviderID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequest_providerId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    lasso_assign_string(this->ProviderID, value);
    return noneRef();
}

static PyObject*
LibAuthnRequest_affiliationId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    return_value = this->AffiliationID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequest_affiliationId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    lasso_assign_string(this->AffiliationID, value);
    return noneRef();
}

static PyObject*
LibAuthnRequest_nameIdPolicy_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    return_value = this->NameIDPolicy;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequest_nameIdPolicy_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    lasso_assign_string(this->NameIDPolicy, value);
    return noneRef();
}

static PyObject*
LibAuthnRequest_forceAuthn_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    gboolean return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    return_value = this->ForceAuthn;
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequest_forceAuthn_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    this->ForceAuthn = PyInt_AS_LONG(cvt_value) ? TRUE : FALSE;
    return noneRef();
}

static PyObject*
LibAuthnRequest_isPassive_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    gboolean return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    return_value = this->IsPassive;
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequest_isPassive_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    this->IsPassive = PyInt_AS_LONG(cvt_value) ? TRUE : FALSE;
    return noneRef();
}

static PyObject*
LibAuthnRequest_protocolProfile_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    return_value = this->ProtocolProfile;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequest_protocolProfile_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    lasso_assign_string(this->ProtocolProfile, value);
    return noneRef();
}

static PyObject*
LibAuthnRequest_assertionConsumerServiceId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    return_value = this->AssertionConsumerServiceID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequest_assertionConsumerServiceId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    lasso_assign_string(this->AssertionConsumerServiceID, value);
    return noneRef();
}

static PyObject*
LibAuthnRequest_requestAuthnContext_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibRequestAuthnContext* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    return_value = this->RequestAuthnContext;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibAuthnRequest_requestAuthnContext_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->RequestAuthnContext, cvt_value);
    return noneRef();
}

static PyObject*
LibAuthnRequest_relayState_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    return_value = this->RelayState;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequest_relayState_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    lasso_assign_string(this->RelayState, value);
    return noneRef();
}

static PyObject*
LibAuthnRequest_scoping_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibScoping* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    return_value = this->Scoping;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibAuthnRequest_scoping_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->Scoping, cvt_value);
    return noneRef();
}

static PyObject*
LibAuthnRequest_consent_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    return_value = this->consent;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthnRequest_consent_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthnRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthnRequest*)cvt_this->obj;
    lasso_assign_string(this->consent, value);
    return noneRef();
}

static PyObject*
LibLogoutRequest_extension_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    return_value = this->Extension;
    return_pyvalue = get_list_of_xml_nodes(return_value);
    return return_pyvalue;
}

static PyObject*
LibLogoutRequest_extension_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    set_list_of_xml_nodes(&this->Extension, cvt_value);
    return noneRef();
}

static PyObject*
LibLogoutRequest_providerId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    return_value = this->ProviderID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibLogoutRequest_providerId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    lasso_assign_string(this->ProviderID, value);
    return noneRef();
}

static PyObject*
LibLogoutRequest_nameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlNameIdentifier* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    return_value = this->NameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibLogoutRequest_nameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->NameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
LibLogoutRequest_sessionIndex_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    return_value = this->SessionIndex;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibLogoutRequest_sessionIndex_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    lasso_assign_string(this->SessionIndex, value);
    return noneRef();
}

static PyObject*
LibLogoutRequest_relayState_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    return_value = this->RelayState;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibLogoutRequest_relayState_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    lasso_assign_string(this->RelayState, value);
    return noneRef();
}

static PyObject*
LibLogoutRequest_consent_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    return_value = this->consent;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibLogoutRequest_consent_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    lasso_assign_string(this->consent, value);
    return noneRef();
}

static PyObject*
LibLogoutRequest_notOnOrAfter_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    return_value = this->NotOnOrAfter;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibLogoutRequest_notOnOrAfter_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibLogoutRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibLogoutRequest*)cvt_this->obj;
    lasso_assign_string(this->NotOnOrAfter, value);
    return noneRef();
}

static PyObject*
lib_logout_request_set_session_indexes(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibLogoutRequest* lib_logout_request = NULL;
    PyGObjectPtr *cvt_lib_logout_request = NULL;
    GList* session_indexes = NULL;
    PyObject *cvt_session_indexes = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_lib_logout_request, &cvt_session_indexes)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_lib_logout_request, &PyGObjectPtrType)) {
        lib_logout_request = (LassoLibLogoutRequest*)cvt_lib_logout_request->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    set_list_of_strings(&session_indexes, cvt_session_indexes);
    lasso_lib_logout_request_set_session_indexes(lib_logout_request, session_indexes);
    return noneRef();
}

static PyObject*
lib_logout_request_get_session_indexes(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibLogoutRequest* lib_logout_request = NULL;
    PyGObjectPtr *cvt_lib_logout_request = NULL;
    GList* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_lib_logout_request)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_lib_logout_request, &PyGObjectPtrType)) {
        lib_logout_request = (LassoLibLogoutRequest*)cvt_lib_logout_request->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_lib_logout_request_get_session_indexes(lib_logout_request);
    return_pyvalue = get_list_of_strings(return_value);
    lasso_release_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
LibAssertion_inResponseTo_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAssertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAssertion*)cvt_this->obj;
    return_value = this->InResponseTo;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAssertion_inResponseTo_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAssertion* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAssertion*)cvt_this->obj;
    lasso_assign_string(this->InResponseTo, value);
    return noneRef();
}

static PyObject*
SamlpRequest_assertionArtifact_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpRequest*)cvt_this->obj;
    return_value = this->AssertionArtifact;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpRequest_assertionArtifact_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpRequest*)cvt_this->obj;
    lasso_assign_string(this->AssertionArtifact, value);
    return noneRef();
}

static PyObject*
SamlSubjectStatementAbstract_subject_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlSubject* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlSubjectStatementAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlSubjectStatementAbstract*)cvt_this->obj;
    return_value = this->Subject;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlSubjectStatementAbstract_subject_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlSubjectStatementAbstract* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlSubjectStatementAbstract*)cvt_this->obj;
    set_object_field((GObject**)&this->Subject, cvt_value);
    return noneRef();
}

static PyObject*
SamlAttributeStatement_attribute_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAttributeStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAttributeStatement*)cvt_this->obj;
    return_value = this->Attribute;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
SamlAttributeStatement_attribute_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAttributeStatement* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAttributeStatement*)cvt_this->obj;
    set_list_of_pygobject(&this->Attribute, cvt_value);
    return noneRef();
}

static PyObject*
SamlSubjectLocality_ipAddress_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlSubjectLocality* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlSubjectLocality*)cvt_this->obj;
    return_value = this->IPAddress;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlSubjectLocality_ipAddress_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlSubjectLocality* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlSubjectLocality*)cvt_this->obj;
    lasso_assign_string(this->IPAddress, value);
    return noneRef();
}

static PyObject*
SamlSubjectLocality_dnsAddress_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlSubjectLocality* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlSubjectLocality*)cvt_this->obj;
    return_value = this->DNSAddress;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlSubjectLocality_dnsAddress_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlSubjectLocality* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlSubjectLocality*)cvt_this->obj;
    lasso_assign_string(this->DNSAddress, value);
    return noneRef();
}

static PyObject*
LibIDPEntries_idpEntry_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibIDPEntries* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibIDPEntries*)cvt_this->obj;
    return_value = this->IDPEntry;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
LibIDPEntries_idpEntry_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibIDPEntries* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibIDPEntries*)cvt_this->obj;
    set_list_of_pygobject(&this->IDPEntry, cvt_value);
    return noneRef();
}

static PyObject*
LibStatusResponse_extension_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibStatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibStatusResponse*)cvt_this->obj;
    return_value = this->Extension;
    return_pyvalue = get_list_of_xml_nodes(return_value);
    return return_pyvalue;
}

static PyObject*
LibStatusResponse_extension_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibStatusResponse* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibStatusResponse*)cvt_this->obj;
    set_list_of_xml_nodes(&this->Extension, cvt_value);
    return noneRef();
}

static PyObject*
LibStatusResponse_providerId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibStatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibStatusResponse*)cvt_this->obj;
    return_value = this->ProviderID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibStatusResponse_providerId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibStatusResponse* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibStatusResponse*)cvt_this->obj;
    lasso_assign_string(this->ProviderID, value);
    return noneRef();
}

static PyObject*
LibStatusResponse_status_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlpStatus* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibStatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibStatusResponse*)cvt_this->obj;
    return_value = this->Status;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibStatusResponse_status_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibStatusResponse* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibStatusResponse*)cvt_this->obj;
    set_object_field((GObject**)&this->Status, cvt_value);
    return noneRef();
}

static PyObject*
LibStatusResponse_relayState_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibStatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibStatusResponse*)cvt_this->obj;
    return_value = this->RelayState;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibStatusResponse_relayState_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibStatusResponse* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibStatusResponse*)cvt_this->obj;
    lasso_assign_string(this->RelayState, value);
    return noneRef();
}

static PyObject*
SamlNameIdentifier_nameQualifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlNameIdentifier* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlNameIdentifier*)cvt_this->obj;
    return_value = this->NameQualifier;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlNameIdentifier_nameQualifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlNameIdentifier* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlNameIdentifier*)cvt_this->obj;
    lasso_assign_string(this->NameQualifier, value);
    return noneRef();
}

static PyObject*
SamlNameIdentifier_format_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlNameIdentifier* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlNameIdentifier*)cvt_this->obj;
    return_value = this->Format;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlNameIdentifier_format_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlNameIdentifier* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlNameIdentifier*)cvt_this->obj;
    lasso_assign_string(this->Format, value);
    return noneRef();
}

static PyObject*
SamlNameIdentifier_content_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlNameIdentifier* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlNameIdentifier*)cvt_this->obj;
    return_value = this->content;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlNameIdentifier_content_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlNameIdentifier* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlNameIdentifier*)cvt_this->obj;
    lasso_assign_string(this->content, value);
    return noneRef();
}

static PyObject*
saml_name_identifier_equals(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlNameIdentifier* a = NULL;
    PyGObjectPtr *cvt_a = NULL;
    LassoSamlNameIdentifier* b = NULL;
    PyGObjectPtr *cvt_b = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_a, &cvt_b)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_a, &PyGObjectPtrType)) {
        a = (LassoSamlNameIdentifier*)cvt_a->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_b, &PyGObjectPtrType)) {
        b = (LassoSamlNameIdentifier*)cvt_b->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml_name_identifier_equals(a, b);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
MiscTextNode_content_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoMiscTextNode* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoMiscTextNode*)cvt_this->obj;
    return_value = this->content;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
MiscTextNode_content_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoMiscTextNode* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoMiscTextNode*)cvt_this->obj;
    lasso_assign_string(this->content, value);
    return noneRef();
}

static PyObject*
MiscTextNode_name_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoMiscTextNode* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoMiscTextNode*)cvt_this->obj;
    return_value = this->name;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
MiscTextNode_name_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoMiscTextNode* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoMiscTextNode*)cvt_this->obj;
    lasso_assign_string(this->name, value);
    return noneRef();
}

static PyObject*
MiscTextNode_nsHref_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoMiscTextNode* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoMiscTextNode*)cvt_this->obj;
    return_value = this->ns_href;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
MiscTextNode_nsHref_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoMiscTextNode* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoMiscTextNode*)cvt_this->obj;
    lasso_assign_string(this->ns_href, value);
    return noneRef();
}

static PyObject*
MiscTextNode_nsPrefix_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoMiscTextNode* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoMiscTextNode*)cvt_this->obj;
    return_value = this->ns_prefix;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
MiscTextNode_nsPrefix_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoMiscTextNode* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoMiscTextNode*)cvt_this->obj;
    lasso_assign_string(this->ns_prefix, value);
    return noneRef();
}

static PyObject*
MiscTextNode_textChild_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    gboolean return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoMiscTextNode* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoMiscTextNode*)cvt_this->obj;
    return_value = this->text_child;
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
MiscTextNode_textChild_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoMiscTextNode* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoMiscTextNode*)cvt_this->obj;
    this->text_child = PyInt_AS_LONG(cvt_value) ? TRUE : FALSE;
    return noneRef();
}

static PyObject*
misc_text_node_set_xml_content(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoMiscTextNode* misc_text_node = NULL;
    PyGObjectPtr *cvt_misc_text_node = NULL;
    xmlNode* node = NULL;
    PyObject *cvt_node = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_misc_text_node, &cvt_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_misc_text_node, &PyGObjectPtrType)) {
        misc_text_node = (LassoMiscTextNode*)cvt_misc_text_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    node = get_xml_node_from_pystring(cvt_node);
    lasso_misc_text_node_set_xml_content(misc_text_node, node);
    lasso_release_xml_node(node);
    return noneRef();
}

static PyObject*
misc_text_node_get_xml_content(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoMiscTextNode* misc_text_node = NULL;
    PyGObjectPtr *cvt_misc_text_node = NULL;
    xmlNode* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_misc_text_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_misc_text_node, &PyGObjectPtrType)) {
        misc_text_node = (LassoMiscTextNode*)cvt_misc_text_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_misc_text_node_get_xml_content(misc_text_node);
    if (return_value) {
        return_pyvalue = get_pystring_from_xml_node(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAudienceRestrictionCondition_audience_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAudienceRestrictionCondition* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAudienceRestrictionCondition*)cvt_this->obj;
    return_value = this->Audience;
    return_pyvalue = get_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
SamlAudienceRestrictionCondition_audience_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAudienceRestrictionCondition* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAudienceRestrictionCondition*)cvt_this->obj;
    set_list_of_strings(&this->Audience, cvt_value);
    return noneRef();
}

static PyObject*
SamlSubjectConfirmation_confirmationMethod_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlSubjectConfirmation* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlSubjectConfirmation*)cvt_this->obj;
    return_value = this->ConfirmationMethod;
    return_pyvalue = get_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
SamlSubjectConfirmation_confirmationMethod_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlSubjectConfirmation* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlSubjectConfirmation*)cvt_this->obj;
    set_list_of_strings(&this->ConfirmationMethod, cvt_value);
    return noneRef();
}

static PyObject*
SamlSubjectConfirmation_subjectConfirmationData_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlSubjectConfirmation* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlSubjectConfirmation*)cvt_this->obj;
    return_value = this->SubjectConfirmationData;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlSubjectConfirmation_subjectConfirmationData_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlSubjectConfirmation* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlSubjectConfirmation*)cvt_this->obj;
    lasso_assign_string(this->SubjectConfirmationData, value);
    return noneRef();
}

static PyObject*
SamlSubjectConfirmation_keyInfo_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsKeyInfo* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlSubjectConfirmation* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlSubjectConfirmation*)cvt_this->obj;
    return_value = this->KeyInfo;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlSubjectConfirmation_keyInfo_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlSubjectConfirmation* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlSubjectConfirmation*)cvt_this->obj;
    set_object_field((GObject**)&this->KeyInfo, cvt_value);
    return noneRef();
}

static PyObject*
SamlAuthorityBinding_authorityKind_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAuthorityBinding* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAuthorityBinding*)cvt_this->obj;
    return_value = this->AuthorityKind;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAuthorityBinding_authorityKind_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAuthorityBinding* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAuthorityBinding*)cvt_this->obj;
    lasso_assign_string(this->AuthorityKind, value);
    return noneRef();
}

static PyObject*
SamlAuthorityBinding_location_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAuthorityBinding* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAuthorityBinding*)cvt_this->obj;
    return_value = this->Location;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAuthorityBinding_location_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAuthorityBinding* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAuthorityBinding*)cvt_this->obj;
    lasso_assign_string(this->Location, value);
    return noneRef();
}

static PyObject*
SamlAuthorityBinding_binding_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAuthorityBinding* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAuthorityBinding*)cvt_this->obj;
    return_value = this->Binding;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAuthorityBinding_binding_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAuthorityBinding* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAuthorityBinding*)cvt_this->obj;
    lasso_assign_string(this->Binding, value);
    return noneRef();
}

static PyObject*
LibScoping_proxyCount_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    int return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibScoping* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibScoping*)cvt_this->obj;
    return_value = this->ProxyCount;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
LibScoping_proxyCount_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibScoping* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoLibScoping*)cvt_this->obj;
    this->ProxyCount = value;
    return noneRef();
}

static PyObject*
LibScoping_idpList_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibIDPList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibScoping* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibScoping*)cvt_this->obj;
    return_value = this->IDPList;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibScoping_idpList_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibScoping* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibScoping*)cvt_this->obj;
    set_object_field((GObject**)&this->IDPList, cvt_value);
    return noneRef();
}

static PyObject*
SamlpStatus_statusCode_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlpStatusCode* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpStatus* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpStatus*)cvt_this->obj;
    return_value = this->StatusCode;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlpStatus_statusCode_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpStatus* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlpStatus*)cvt_this->obj;
    set_object_field((GObject**)&this->StatusCode, cvt_value);
    return noneRef();
}

static PyObject*
SamlpStatus_statusMessage_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlpStatus* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlpStatus*)cvt_this->obj;
    return_value = this->StatusMessage;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlpStatus_statusMessage_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlpStatus* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlpStatus*)cvt_this->obj;
    lasso_assign_string(this->StatusMessage, value);
    return noneRef();
}

static PyObject*
LibRegisterNameIdentifierRequest_extension_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibRegisterNameIdentifierRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibRegisterNameIdentifierRequest*)cvt_this->obj;
    return_value = this->Extension;
    return_pyvalue = get_list_of_xml_nodes(return_value);
    return return_pyvalue;
}

static PyObject*
LibRegisterNameIdentifierRequest_extension_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibRegisterNameIdentifierRequest* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibRegisterNameIdentifierRequest*)cvt_this->obj;
    set_list_of_xml_nodes(&this->Extension, cvt_value);
    return noneRef();
}

static PyObject*
LibRegisterNameIdentifierRequest_providerId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibRegisterNameIdentifierRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibRegisterNameIdentifierRequest*)cvt_this->obj;
    return_value = this->ProviderID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibRegisterNameIdentifierRequest_providerId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibRegisterNameIdentifierRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibRegisterNameIdentifierRequest*)cvt_this->obj;
    lasso_assign_string(this->ProviderID, value);
    return noneRef();
}

static PyObject*
LibRegisterNameIdentifierRequest_idpProvidedNameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlNameIdentifier* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibRegisterNameIdentifierRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibRegisterNameIdentifierRequest*)cvt_this->obj;
    return_value = this->IDPProvidedNameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibRegisterNameIdentifierRequest_idpProvidedNameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibRegisterNameIdentifierRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibRegisterNameIdentifierRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->IDPProvidedNameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
LibRegisterNameIdentifierRequest_spProvidedNameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlNameIdentifier* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibRegisterNameIdentifierRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibRegisterNameIdentifierRequest*)cvt_this->obj;
    return_value = this->SPProvidedNameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibRegisterNameIdentifierRequest_spProvidedNameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibRegisterNameIdentifierRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibRegisterNameIdentifierRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->SPProvidedNameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
LibRegisterNameIdentifierRequest_oldProvidedNameIdentifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlNameIdentifier* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibRegisterNameIdentifierRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibRegisterNameIdentifierRequest*)cvt_this->obj;
    return_value = this->OldProvidedNameIdentifier;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibRegisterNameIdentifierRequest_oldProvidedNameIdentifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibRegisterNameIdentifierRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibRegisterNameIdentifierRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->OldProvidedNameIdentifier, cvt_value);
    return noneRef();
}

static PyObject*
LibRegisterNameIdentifierRequest_relayState_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibRegisterNameIdentifierRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibRegisterNameIdentifierRequest*)cvt_this->obj;
    return_value = this->RelayState;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibRegisterNameIdentifierRequest_relayState_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibRegisterNameIdentifierRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibRegisterNameIdentifierRequest*)cvt_this->obj;
    lasso_assign_string(this->RelayState, value);
    return noneRef();
}

static PyObject*
SamlAuthenticationStatement_subjectLocality_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlSubjectLocality* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAuthenticationStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAuthenticationStatement*)cvt_this->obj;
    return_value = this->SubjectLocality;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SamlAuthenticationStatement_subjectLocality_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAuthenticationStatement* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAuthenticationStatement*)cvt_this->obj;
    set_object_field((GObject**)&this->SubjectLocality, cvt_value);
    return noneRef();
}

static PyObject*
SamlAuthenticationStatement_authorityBinding_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAuthenticationStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAuthenticationStatement*)cvt_this->obj;
    return_value = this->AuthorityBinding;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
SamlAuthenticationStatement_authorityBinding_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAuthenticationStatement* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlAuthenticationStatement*)cvt_this->obj;
    set_list_of_pygobject(&this->AuthorityBinding, cvt_value);
    return noneRef();
}

static PyObject*
SamlAuthenticationStatement_authenticationMethod_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAuthenticationStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAuthenticationStatement*)cvt_this->obj;
    return_value = this->AuthenticationMethod;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAuthenticationStatement_authenticationMethod_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAuthenticationStatement* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAuthenticationStatement*)cvt_this->obj;
    lasso_assign_string(this->AuthenticationMethod, value);
    return noneRef();
}

static PyObject*
SamlAuthenticationStatement_authenticationInstant_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlAuthenticationStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlAuthenticationStatement*)cvt_this->obj;
    return_value = this->AuthenticationInstant;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SamlAuthenticationStatement_authenticationInstant_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlAuthenticationStatement* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlAuthenticationStatement*)cvt_this->obj;
    lasso_assign_string(this->AuthenticationInstant, value);
    return noneRef();
}

static PyObject*
LibAuthenticationStatement_authnContext_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoLibAuthnContext* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthenticationStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthenticationStatement*)cvt_this->obj;
    return_value = this->AuthnContext;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
LibAuthenticationStatement_authnContext_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthenticationStatement* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoLibAuthenticationStatement*)cvt_this->obj;
    set_object_field((GObject**)&this->AuthnContext, cvt_value);
    return noneRef();
}

static PyObject*
LibAuthenticationStatement_reauthenticateOnOrAfter_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthenticationStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthenticationStatement*)cvt_this->obj;
    return_value = this->ReauthenticateOnOrAfter;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthenticationStatement_reauthenticateOnOrAfter_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthenticationStatement* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthenticationStatement*)cvt_this->obj;
    lasso_assign_string(this->ReauthenticateOnOrAfter, value);
    return noneRef();
}

static PyObject*
LibAuthenticationStatement_sessionIndex_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoLibAuthenticationStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoLibAuthenticationStatement*)cvt_this->obj;
    return_value = this->SessionIndex;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
LibAuthenticationStatement_sessionIndex_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoLibAuthenticationStatement* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoLibAuthenticationStatement*)cvt_this->obj;
    lasso_assign_string(this->SessionIndex, value);
    return noneRef();
}

static PyObject*
SoapFault_faultcode_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSoapFault* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSoapFault*)cvt_this->obj;
    return_value = this->faultcode;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SoapFault_faultcode_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSoapFault* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSoapFault*)cvt_this->obj;
    lasso_assign_string(this->faultcode, value);
    return noneRef();
}

static PyObject*
SoapFault_faultstring_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSoapFault* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSoapFault*)cvt_this->obj;
    return_value = this->faultstring;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SoapFault_faultstring_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSoapFault* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSoapFault*)cvt_this->obj;
    lasso_assign_string(this->faultstring, value);
    return noneRef();
}

static PyObject*
SoapFault_faultactor_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSoapFault* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSoapFault*)cvt_this->obj;
    return_value = this->faultactor;
    return_pyvalue = get_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
SoapFault_faultactor_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSoapFault* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSoapFault*)cvt_this->obj;
    set_list_of_strings(&this->faultactor, cvt_value);
    return noneRef();
}

static PyObject*
SoapFault_detail_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSoapDetail* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSoapFault* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSoapFault*)cvt_this->obj;
    return_value = this->Detail;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SoapFault_detail_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSoapFault* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSoapFault*)cvt_this->obj;
    set_object_field((GObject**)&this->Detail, cvt_value);
    return noneRef();
}

static PyObject*
SoapEnvelope_header_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSoapHeader* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSoapEnvelope* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSoapEnvelope*)cvt_this->obj;
    return_value = this->Header;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SoapEnvelope_header_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSoapEnvelope* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSoapEnvelope*)cvt_this->obj;
    set_object_field((GObject**)&this->Header, cvt_value);
    return noneRef();
}

static PyObject*
SoapEnvelope_body_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSoapBody* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSoapEnvelope* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSoapEnvelope*)cvt_this->obj;
    return_value = this->Body;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
SoapEnvelope_body_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSoapEnvelope* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSoapEnvelope*)cvt_this->obj;
    set_object_field((GObject**)&this->Body, cvt_value);
    return noneRef();
}

static PyObject*
SoapHeader_other_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSoapHeader* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSoapHeader*)cvt_this->obj;
    return_value = this->Other;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
SoapHeader_other_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSoapHeader* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSoapHeader*)cvt_this->obj;
    set_list_of_pygobject(&this->Other, cvt_value);
    return noneRef();
}

static PyObject*
SoapBody_any_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSoapBody* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSoapBody*)cvt_this->obj;
    return_value = this->any;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
SoapBody_any_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSoapBody* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSoapBody*)cvt_this->obj;
    set_list_of_pygobject(&this->any, cvt_value);
    return noneRef();
}

static PyObject*
SoapBody_id_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSoapBody* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSoapBody*)cvt_this->obj;
    return_value = this->Id;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
SoapBody_id_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSoapBody* this;
    gchar* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSoapBody*)cvt_this->obj;
    lasso_assign_string(this->Id, value);
    return noneRef();
}

static PyObject*
SoapDetail_any_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSoapDetail* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSoapDetail*)cvt_this->obj;
    return_value = this->any;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
SoapDetail_any_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSoapDetail* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSoapDetail*)cvt_this->obj;
    set_list_of_pygobject(&this->any, cvt_value);
    return noneRef();
}

static PyObject*
Saml2SubjectConfirmation_baseId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2BaseIDAbstract* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmation* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2SubjectConfirmation*)cvt_this->obj;
    return_value = this->BaseID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2SubjectConfirmation_baseId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmation* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2SubjectConfirmation*)cvt_this->obj;
    set_object_field((GObject**)&this->BaseID, cvt_value);
    return noneRef();
}

static PyObject*
Saml2SubjectConfirmation_nameId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2NameID* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmation* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2SubjectConfirmation*)cvt_this->obj;
    return_value = this->NameID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2SubjectConfirmation_nameId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmation* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2SubjectConfirmation*)cvt_this->obj;
    set_object_field((GObject**)&this->NameID, cvt_value);
    return noneRef();
}

static PyObject*
Saml2SubjectConfirmation_encryptedId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2EncryptedElement* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmation* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2SubjectConfirmation*)cvt_this->obj;
    return_value = this->EncryptedID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2SubjectConfirmation_encryptedId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmation* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2SubjectConfirmation*)cvt_this->obj;
    set_object_field((GObject**)&this->EncryptedID, cvt_value);
    return noneRef();
}

static PyObject*
Saml2SubjectConfirmation_subjectConfirmationData_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2SubjectConfirmationData* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmation* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2SubjectConfirmation*)cvt_this->obj;
    return_value = this->SubjectConfirmationData;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2SubjectConfirmation_subjectConfirmationData_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmation* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2SubjectConfirmation*)cvt_this->obj;
    set_object_field((GObject**)&this->SubjectConfirmationData, cvt_value);
    return noneRef();
}

static PyObject*
Saml2SubjectConfirmation_method_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmation* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2SubjectConfirmation*)cvt_this->obj;
    return_value = this->Method;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2SubjectConfirmation_method_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmation* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2SubjectConfirmation*)cvt_this->obj;
    lasso_assign_string(this->Method, value);
    return noneRef();
}

static PyObject*
Samlp2IDPList_idpEntry_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2IDPEntry* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2IDPList* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2IDPList*)cvt_this->obj;
    return_value = this->IDPEntry;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2IDPList_idpEntry_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2IDPList* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2IDPList*)cvt_this->obj;
    set_object_field((GObject**)&this->IDPEntry, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2IDPList_getComplete_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2IDPList* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2IDPList*)cvt_this->obj;
    return_value = this->GetComplete;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2IDPList_getComplete_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2IDPList* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2IDPList*)cvt_this->obj;
    lasso_assign_string(this->GetComplete, value);
    return noneRef();
}

static PyObject*
Saml2AuthnContext_authnContextClassRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnContext* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AuthnContext*)cvt_this->obj;
    return_value = this->AuthnContextClassRef;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2AuthnContext_authnContextClassRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnContext* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2AuthnContext*)cvt_this->obj;
    lasso_assign_string(this->AuthnContextClassRef, value);
    return noneRef();
}

static PyObject*
Saml2AuthnContext_authnContextDeclRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnContext* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AuthnContext*)cvt_this->obj;
    return_value = this->AuthnContextDeclRef;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2AuthnContext_authnContextDeclRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnContext* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2AuthnContext*)cvt_this->obj;
    lasso_assign_string(this->AuthnContextDeclRef, value);
    return noneRef();
}

static PyObject*
Saml2AuthnContext_authenticatingAuthority_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnContext* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AuthnContext*)cvt_this->obj;
    return_value = this->AuthenticatingAuthority;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2AuthnContext_authenticatingAuthority_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnContext* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2AuthnContext*)cvt_this->obj;
    lasso_assign_string(this->AuthenticatingAuthority, value);
    return noneRef();
}

static PyObject*
Saml2BaseIDAbstract_nameQualifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2BaseIDAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2BaseIDAbstract*)cvt_this->obj;
    return_value = this->NameQualifier;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2BaseIDAbstract_nameQualifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2BaseIDAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2BaseIDAbstract*)cvt_this->obj;
    lasso_assign_string(this->NameQualifier, value);
    return noneRef();
}

static PyObject*
Saml2BaseIDAbstract_spNameQualifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2BaseIDAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2BaseIDAbstract*)cvt_this->obj;
    return_value = this->SPNameQualifier;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2BaseIDAbstract_spNameQualifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2BaseIDAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2BaseIDAbstract*)cvt_this->obj;
    lasso_assign_string(this->SPNameQualifier, value);
    return noneRef();
}

static PyObject*
Saml2SubjectLocality_address_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectLocality* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2SubjectLocality*)cvt_this->obj;
    return_value = this->Address;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2SubjectLocality_address_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectLocality* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2SubjectLocality*)cvt_this->obj;
    lasso_assign_string(this->Address, value);
    return noneRef();
}

static PyObject*
Saml2SubjectLocality_dnsName_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectLocality* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2SubjectLocality*)cvt_this->obj;
    return_value = this->DNSName;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2SubjectLocality_dnsName_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectLocality* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2SubjectLocality*)cvt_this->obj;
    lasso_assign_string(this->DNSName, value);
    return noneRef();
}

static PyObject*
Samlp2Status_statusCode_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2StatusCode* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2Status* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2Status*)cvt_this->obj;
    return_value = this->StatusCode;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2Status_statusCode_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2Status* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2Status*)cvt_this->obj;
    set_object_field((GObject**)&this->StatusCode, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2Status_statusMessage_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2Status* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2Status*)cvt_this->obj;
    return_value = this->StatusMessage;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2Status_statusMessage_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2Status* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2Status*)cvt_this->obj;
    lasso_assign_string(this->StatusMessage, value);
    return noneRef();
}

static PyObject*
Samlp2Status_statusDetail_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2StatusDetail* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2Status* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2Status*)cvt_this->obj;
    return_value = this->StatusDetail;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2Status_statusDetail_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2Status* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2Status*)cvt_this->obj;
    set_object_field((GObject**)&this->StatusDetail, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2RequestAbstract_issuer_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2NameID* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    return_value = this->Issuer;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2RequestAbstract_issuer_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    set_object_field((GObject**)&this->Issuer, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2RequestAbstract_extensions_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2Extensions* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    return_value = this->Extensions;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2RequestAbstract_extensions_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    set_object_field((GObject**)&this->Extensions, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2RequestAbstract_id_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    return_value = this->ID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2RequestAbstract_id_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    lasso_assign_string(this->ID, value);
    return noneRef();
}

static PyObject*
Samlp2RequestAbstract_version_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    return_value = this->Version;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2RequestAbstract_version_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    lasso_assign_string(this->Version, value);
    return noneRef();
}

static PyObject*
Samlp2RequestAbstract_issueInstant_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    return_value = this->IssueInstant;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2RequestAbstract_issueInstant_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    lasso_assign_string(this->IssueInstant, value);
    return noneRef();
}

static PyObject*
Samlp2RequestAbstract_destination_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    return_value = this->Destination;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2RequestAbstract_destination_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    lasso_assign_string(this->Destination, value);
    return noneRef();
}

static PyObject*
Samlp2RequestAbstract_consent_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    return_value = this->Consent;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2RequestAbstract_consent_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestAbstract* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2RequestAbstract*)cvt_this->obj;
    lasso_assign_string(this->Consent, value);
    return noneRef();
}

static PyObject*
Samlp2AssertionIDRequest_assertionIdRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AssertionIDRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AssertionIDRequest*)cvt_this->obj;
    return_value = this->AssertionIDRef;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2AssertionIDRequest_assertionIdRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AssertionIDRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2AssertionIDRequest*)cvt_this->obj;
    lasso_assign_string(this->AssertionIDRef, value);
    return noneRef();
}

static PyObject*
Samlp2NameIDPolicy_format_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDPolicy* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2NameIDPolicy*)cvt_this->obj;
    return_value = this->Format;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2NameIDPolicy_format_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDPolicy* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2NameIDPolicy*)cvt_this->obj;
    lasso_assign_string(this->Format, value);
    return noneRef();
}

static PyObject*
Samlp2NameIDPolicy_spNameQualifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDPolicy* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2NameIDPolicy*)cvt_this->obj;
    return_value = this->SPNameQualifier;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2NameIDPolicy_spNameQualifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDPolicy* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2NameIDPolicy*)cvt_this->obj;
    lasso_assign_string(this->SPNameQualifier, value);
    return noneRef();
}

static PyObject*
Samlp2NameIDPolicy_allowCreate_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    gboolean return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDPolicy* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2NameIDPolicy*)cvt_this->obj;
    return_value = this->AllowCreate;
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
Samlp2NameIDPolicy_allowCreate_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDPolicy* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2NameIDPolicy*)cvt_this->obj;
    this->AllowCreate = PyInt_AS_LONG(cvt_value) ? TRUE : FALSE;
    return noneRef();
}

static PyObject*
Samlp2AuthnRequest_subject_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Subject* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    return_value = this->Subject;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2AuthnRequest_subject_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->Subject, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2AuthnRequest_nameIdPolicy_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2NameIDPolicy* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    return_value = this->NameIDPolicy;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2AuthnRequest_nameIdPolicy_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->NameIDPolicy, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2AuthnRequest_conditions_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Conditions* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    return_value = this->Conditions;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2AuthnRequest_conditions_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->Conditions, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2AuthnRequest_requestedAuthnContext_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2RequestedAuthnContext* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    return_value = this->RequestedAuthnContext;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2AuthnRequest_requestedAuthnContext_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->RequestedAuthnContext, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2AuthnRequest_scoping_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2Scoping* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    return_value = this->Scoping;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2AuthnRequest_scoping_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->Scoping, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2AuthnRequest_forceAuthn_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    gboolean return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    return_value = this->ForceAuthn;
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
Samlp2AuthnRequest_forceAuthn_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    this->ForceAuthn = PyInt_AS_LONG(cvt_value) ? TRUE : FALSE;
    return noneRef();
}

static PyObject*
Samlp2AuthnRequest_isPassive_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    gboolean return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    return_value = this->IsPassive;
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
Samlp2AuthnRequest_isPassive_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    this->IsPassive = PyInt_AS_LONG(cvt_value) ? TRUE : FALSE;
    return noneRef();
}

static PyObject*
Samlp2AuthnRequest_protocolBinding_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    return_value = this->ProtocolBinding;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2AuthnRequest_protocolBinding_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    lasso_assign_string(this->ProtocolBinding, value);
    return noneRef();
}

static PyObject*
Samlp2AuthnRequest_assertionConsumerServiceIndex_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    int return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    return_value = this->AssertionConsumerServiceIndex;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
Samlp2AuthnRequest_assertionConsumerServiceIndex_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    this->AssertionConsumerServiceIndex = value;
    return noneRef();
}

static PyObject*
Samlp2AuthnRequest_assertionConsumerServiceUrl_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    return_value = this->AssertionConsumerServiceURL;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2AuthnRequest_assertionConsumerServiceUrl_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    lasso_assign_string(this->AssertionConsumerServiceURL, value);
    return noneRef();
}

static PyObject*
Samlp2AuthnRequest_attributeConsumingServiceIndex_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    int return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    return_value = this->AttributeConsumingServiceIndex;
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
Samlp2AuthnRequest_attributeConsumingServiceIndex_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;
    long value;
    if (! PyArg_ParseTuple(args, "Ol", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    this->AttributeConsumingServiceIndex = value;
    return noneRef();
}

static PyObject*
Samlp2AuthnRequest_providerName_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    return_value = this->ProviderName;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2AuthnRequest_providerName_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2AuthnRequest*)cvt_this->obj;
    lasso_assign_string(this->ProviderName, value);
    return noneRef();
}

static PyObject*
Saml2ProxyRestriction_audience_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2ProxyRestriction* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2ProxyRestriction*)cvt_this->obj;
    return_value = this->Audience;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2ProxyRestriction_audience_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2ProxyRestriction* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2ProxyRestriction*)cvt_this->obj;
    lasso_assign_string(this->Audience, value);
    return noneRef();
}

static PyObject*
Saml2ProxyRestriction_count_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2ProxyRestriction* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2ProxyRestriction*)cvt_this->obj;
    return_value = this->Count;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2ProxyRestriction_count_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2ProxyRestriction* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2ProxyRestriction*)cvt_this->obj;
    lasso_assign_string(this->Count, value);
    return noneRef();
}

static PyObject*
Saml2Subject_baseId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2BaseIDAbstract* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Subject* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Subject*)cvt_this->obj;
    return_value = this->BaseID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2Subject_baseId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Subject* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Subject*)cvt_this->obj;
    set_object_field((GObject**)&this->BaseID, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Subject_nameId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2NameID* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Subject* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Subject*)cvt_this->obj;
    return_value = this->NameID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2Subject_nameId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Subject* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Subject*)cvt_this->obj;
    set_object_field((GObject**)&this->NameID, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Subject_encryptedId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2EncryptedElement* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Subject* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Subject*)cvt_this->obj;
    return_value = this->EncryptedID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2Subject_encryptedId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Subject* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Subject*)cvt_this->obj;
    set_object_field((GObject**)&this->EncryptedID, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Subject_subjectConfirmation_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2SubjectConfirmation* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Subject* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Subject*)cvt_this->obj;
    return_value = this->SubjectConfirmation;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2Subject_subjectConfirmation_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Subject* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Subject*)cvt_this->obj;
    set_object_field((GObject**)&this->SubjectConfirmation, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2RequestedAuthnContext_authnContextClassRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestedAuthnContext* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2RequestedAuthnContext*)cvt_this->obj;
    return_value = this->AuthnContextClassRef;
    return_pyvalue = get_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
Samlp2RequestedAuthnContext_authnContextClassRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestedAuthnContext* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2RequestedAuthnContext*)cvt_this->obj;
    set_list_of_strings(&this->AuthnContextClassRef, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2RequestedAuthnContext_authnContextDeclRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestedAuthnContext* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2RequestedAuthnContext*)cvt_this->obj;
    return_value = this->AuthnContextDeclRef;
    return_pyvalue = get_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
Samlp2RequestedAuthnContext_authnContextDeclRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestedAuthnContext* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2RequestedAuthnContext*)cvt_this->obj;
    set_list_of_strings(&this->AuthnContextDeclRef, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2RequestedAuthnContext_comparison_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestedAuthnContext* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2RequestedAuthnContext*)cvt_this->obj;
    return_value = this->Comparison;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2RequestedAuthnContext_comparison_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2RequestedAuthnContext* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2RequestedAuthnContext*)cvt_this->obj;
    lasso_assign_string(this->Comparison, value);
    return noneRef();
}

static PyObject*
Samlp2StatusResponse_issuer_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2NameID* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    return_value = this->Issuer;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2StatusResponse_issuer_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    set_object_field((GObject**)&this->Issuer, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2StatusResponse_extensions_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2Extensions* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    return_value = this->Extensions;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2StatusResponse_extensions_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    set_object_field((GObject**)&this->Extensions, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2StatusResponse_status_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2Status* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    return_value = this->Status;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2StatusResponse_status_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    set_object_field((GObject**)&this->Status, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2StatusResponse_id_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    return_value = this->ID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2StatusResponse_id_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    lasso_assign_string(this->ID, value);
    return noneRef();
}

static PyObject*
Samlp2StatusResponse_inResponseTo_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    return_value = this->InResponseTo;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2StatusResponse_inResponseTo_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    lasso_assign_string(this->InResponseTo, value);
    return noneRef();
}

static PyObject*
Samlp2StatusResponse_version_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    return_value = this->Version;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2StatusResponse_version_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    lasso_assign_string(this->Version, value);
    return noneRef();
}

static PyObject*
Samlp2StatusResponse_issueInstant_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    return_value = this->IssueInstant;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2StatusResponse_issueInstant_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    lasso_assign_string(this->IssueInstant, value);
    return noneRef();
}

static PyObject*
Samlp2StatusResponse_destination_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    return_value = this->Destination;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2StatusResponse_destination_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    lasso_assign_string(this->Destination, value);
    return noneRef();
}

static PyObject*
Samlp2StatusResponse_consent_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    return_value = this->Consent;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2StatusResponse_consent_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusResponse* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2StatusResponse*)cvt_this->obj;
    lasso_assign_string(this->Consent, value);
    return noneRef();
}

static PyObject*
Samlp2ArtifactResponse_any_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2ArtifactResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2ArtifactResponse*)cvt_this->obj;
    return_value = this->any;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2ArtifactResponse_any_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2ArtifactResponse* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2ArtifactResponse*)cvt_this->obj;
    set_object_field((GObject**)&this->any, cvt_value);
    return noneRef();
}

static PyObject*
Saml2AudienceRestriction_audience_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AudienceRestriction* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AudienceRestriction*)cvt_this->obj;
    return_value = this->Audience;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2AudienceRestriction_audience_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AudienceRestriction* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2AudienceRestriction*)cvt_this->obj;
    lasso_assign_string(this->Audience, value);
    return noneRef();
}

static PyObject*
Saml2Attribute_attributeValue_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Attribute* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Attribute*)cvt_this->obj;
    return_value = this->AttributeValue;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Attribute_attributeValue_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Attribute* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Attribute*)cvt_this->obj;
    set_list_of_pygobject(&this->AttributeValue, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Attribute_name_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Attribute* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Attribute*)cvt_this->obj;
    return_value = this->Name;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2Attribute_name_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Attribute* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2Attribute*)cvt_this->obj;
    lasso_assign_string(this->Name, value);
    return noneRef();
}

static PyObject*
Saml2Attribute_nameFormat_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Attribute* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Attribute*)cvt_this->obj;
    return_value = this->NameFormat;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2Attribute_nameFormat_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Attribute* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2Attribute*)cvt_this->obj;
    lasso_assign_string(this->NameFormat, value);
    return noneRef();
}

static PyObject*
Saml2Attribute_friendlyName_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Attribute* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Attribute*)cvt_this->obj;
    return_value = this->FriendlyName;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2Attribute_friendlyName_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Attribute* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2Attribute*)cvt_this->obj;
    lasso_assign_string(this->FriendlyName, value);
    return noneRef();
}

static PyObject*
Samlp2NameIDMappingRequest_baseId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2BaseIDAbstract* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDMappingRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2NameIDMappingRequest*)cvt_this->obj;
    return_value = this->BaseID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2NameIDMappingRequest_baseId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDMappingRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2NameIDMappingRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->BaseID, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2NameIDMappingRequest_nameId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2NameID* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDMappingRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2NameIDMappingRequest*)cvt_this->obj;
    return_value = this->NameID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2NameIDMappingRequest_nameId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDMappingRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2NameIDMappingRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->NameID, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2NameIDMappingRequest_encryptedId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2EncryptedElement* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDMappingRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2NameIDMappingRequest*)cvt_this->obj;
    return_value = this->EncryptedID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2NameIDMappingRequest_encryptedId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDMappingRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2NameIDMappingRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->EncryptedID, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2NameIDMappingRequest_nameIdPolicy_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2NameIDPolicy* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDMappingRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2NameIDMappingRequest*)cvt_this->obj;
    return_value = this->NameIDPolicy;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2NameIDMappingRequest_nameIdPolicy_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDMappingRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2NameIDMappingRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->NameIDPolicy, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2ManageNameIDRequest_nameId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2NameID* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2ManageNameIDRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2ManageNameIDRequest*)cvt_this->obj;
    return_value = this->NameID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2ManageNameIDRequest_nameId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2ManageNameIDRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2ManageNameIDRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->NameID, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2ManageNameIDRequest_encryptedId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2EncryptedElement* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2ManageNameIDRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2ManageNameIDRequest*)cvt_this->obj;
    return_value = this->EncryptedID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2ManageNameIDRequest_encryptedId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2ManageNameIDRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2ManageNameIDRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->EncryptedID, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2ManageNameIDRequest_newId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2ManageNameIDRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2ManageNameIDRequest*)cvt_this->obj;
    return_value = this->NewID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2ManageNameIDRequest_newId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2ManageNameIDRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2ManageNameIDRequest*)cvt_this->obj;
    lasso_assign_string(this->NewID, value);
    return noneRef();
}

static PyObject*
Samlp2ManageNameIDRequest_newEncryptedId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2EncryptedElement* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2ManageNameIDRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2ManageNameIDRequest*)cvt_this->obj;
    return_value = this->NewEncryptedID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2ManageNameIDRequest_newEncryptedId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2ManageNameIDRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2ManageNameIDRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->NewEncryptedID, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2ManageNameIDRequest_terminate_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2Terminate* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2ManageNameIDRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2ManageNameIDRequest*)cvt_this->obj;
    return_value = this->Terminate;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2ManageNameIDRequest_terminate_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2ManageNameIDRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2ManageNameIDRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->Terminate, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2StatusCode_statusCode_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2StatusCode* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusCode* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2StatusCode*)cvt_this->obj;
    return_value = this->StatusCode;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2StatusCode_statusCode_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusCode* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2StatusCode*)cvt_this->obj;
    set_object_field((GObject**)&this->StatusCode, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2StatusCode_value_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusCode* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2StatusCode*)cvt_this->obj;
    return_value = this->Value;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2StatusCode_value_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2StatusCode* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2StatusCode*)cvt_this->obj;
    lasso_assign_string(this->Value, value);
    return noneRef();
}

static PyObject*
Saml2Advice_assertionIdRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Advice* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Advice*)cvt_this->obj;
    return_value = this->AssertionIDRef;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Advice_assertionIdRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Advice* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Advice*)cvt_this->obj;
    set_list_of_pygobject(&this->AssertionIDRef, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Advice_assertionUriRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Advice* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Advice*)cvt_this->obj;
    return_value = this->AssertionURIRef;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Advice_assertionUriRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Advice* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Advice*)cvt_this->obj;
    set_list_of_pygobject(&this->AssertionURIRef, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Advice_assertion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Advice* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Advice*)cvt_this->obj;
    return_value = this->Assertion;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Advice_assertion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Advice* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Advice*)cvt_this->obj;
    set_list_of_pygobject(&this->Assertion, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Advice_encryptedAssertion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Advice* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Advice*)cvt_this->obj;
    return_value = this->EncryptedAssertion;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Advice_encryptedAssertion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Advice* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Advice*)cvt_this->obj;
    set_list_of_pygobject(&this->EncryptedAssertion, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Action_content_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Action* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Action*)cvt_this->obj;
    return_value = this->content;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2Action_content_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Action* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2Action*)cvt_this->obj;
    lasso_assign_string(this->content, value);
    return noneRef();
}

static PyObject*
Saml2Action_namespace_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Action* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Action*)cvt_this->obj;
    return_value = this->Namespace;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2Action_namespace_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Action* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2Action*)cvt_this->obj;
    lasso_assign_string(this->Namespace, value);
    return noneRef();
}

static PyObject*
Saml2EncryptedElement_encryptedData_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    xmlNode* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2EncryptedElement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2EncryptedElement*)cvt_this->obj;
    return_value = this->EncryptedData;
    if (return_value) {
        return_pyvalue = get_pystring_from_xml_node(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2EncryptedElement_encryptedData_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2EncryptedElement* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2EncryptedElement*)cvt_this->obj;
    if (this->EncryptedData) xmlFreeNode(this->EncryptedData);
    this->EncryptedData = get_xml_node_from_pystring(cvt_value);
    return noneRef();
}

static PyObject*
Saml2EncryptedElement_encryptedKey_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2EncryptedElement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2EncryptedElement*)cvt_this->obj;
    return_value = this->EncryptedKey;
    return_pyvalue = get_list_of_xml_nodes(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2EncryptedElement_encryptedKey_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2EncryptedElement* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2EncryptedElement*)cvt_this->obj;
    set_list_of_xml_nodes(&this->EncryptedKey, cvt_value);
    return noneRef();
}

static PyObject*
Saml2EncryptedElement_originalData_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoNode* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2EncryptedElement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2EncryptedElement*)cvt_this->obj;
    return_value = this->original_data;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2EncryptedElement_originalData_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2EncryptedElement* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2EncryptedElement*)cvt_this->obj;
    set_object_field((GObject**)&this->original_data, cvt_value);
    return noneRef();
}

static PyObject*
saml2_encrypted_element_server_decrypt(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2EncryptedElement* encrypted_element = NULL;
    PyGObjectPtr *cvt_encrypted_element = NULL;
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    PyObject *out_pyvalue = NULL;
    LassoNode* decrypted_node = NULL;
    PyObject *cvt_decrypted_node_out = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OOO", &cvt_encrypted_element, &cvt_server, &cvt_decrypted_node_out)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_encrypted_element, &PyGObjectPtrType)) {
        encrypted_element = (LassoSaml2EncryptedElement*)cvt_encrypted_element->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_encrypted_element_server_decrypt(encrypted_element, server, &decrypted_node);
    return_pyvalue = PyInt_FromLong(return_value);
    if (decrypted_node) {
        out_pyvalue = PyGObjectPtr_New(G_OBJECT(decrypted_node));
    } else {
        out_pyvalue = noneRef();
    }

    PyList_SetItem(cvt_decrypted_node_out, 0, out_pyvalue);
    return return_pyvalue;
}

static PyObject*
Saml2Assertion_issuer_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2NameID* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    return_value = this->Issuer;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2Assertion_issuer_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    set_object_field((GObject**)&this->Issuer, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Assertion_subject_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Subject* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    return_value = this->Subject;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2Assertion_subject_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    set_object_field((GObject**)&this->Subject, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Assertion_conditions_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Conditions* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    return_value = this->Conditions;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2Assertion_conditions_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    set_object_field((GObject**)&this->Conditions, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Assertion_advice_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Advice* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    return_value = this->Advice;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2Assertion_advice_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    set_object_field((GObject**)&this->Advice, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Assertion_statement_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    return_value = this->Statement;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Assertion_statement_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    set_list_of_pygobject(&this->Statement, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Assertion_authnStatement_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    return_value = this->AuthnStatement;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Assertion_authnStatement_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    set_list_of_pygobject(&this->AuthnStatement, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Assertion_authzDecisionStatement_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    return_value = this->AuthzDecisionStatement;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Assertion_authzDecisionStatement_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    set_list_of_pygobject(&this->AuthzDecisionStatement, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Assertion_attributeStatement_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    return_value = this->AttributeStatement;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Assertion_attributeStatement_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    set_list_of_pygobject(&this->AttributeStatement, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Assertion_version_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    return_value = this->Version;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2Assertion_version_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    lasso_assign_string(this->Version, value);
    return noneRef();
}

static PyObject*
Saml2Assertion_id_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    return_value = this->ID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2Assertion_id_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    lasso_assign_string(this->ID, value);
    return noneRef();
}

static PyObject*
Saml2Assertion_issueInstant_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    return_value = this->IssueInstant;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2Assertion_issueInstant_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Assertion* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2Assertion*)cvt_this->obj;
    lasso_assign_string(this->IssueInstant, value);
    return noneRef();
}

static PyObject*
saml2_assertion_has_audience_restriction(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_saml2_assertion)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_has_audience_restriction(saml2_assertion);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
saml2_assertion_is_audience_restricted(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    char* providerID = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_saml2_assertion, &providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_is_audience_restricted(saml2_assertion, providerID);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
saml2_assertion_set_subject_name_id(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_saml2_assertion, &cvt_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_saml2_assertion_set_subject_name_id(saml2_assertion, node);
    return noneRef();
}

static PyObject*
saml2_assertion_set_subject_confirmation_name_id(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    LassoNode* node = NULL;
    PyGObjectPtr *cvt_node = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_saml2_assertion, &cvt_node)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_node, &PyGObjectPtrType)) {
        node = (LassoNode*)cvt_node->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_saml2_assertion_set_subject_confirmation_name_id(saml2_assertion, node);
    return noneRef();
}

static PyObject*
saml2_assertion_set_subject_confirmation_data(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    time_t tolerance;
    time_t length;
    const char* Recipient = NULL;
    const char* InResponseTo = NULL;
    const char* Address = NULL;

    if (! PyArg_ParseTuple(args, "Oiisss", &cvt_saml2_assertion, &tolerance, &length, &Recipient, &InResponseTo, &Address)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_saml2_assertion_set_subject_confirmation_data(saml2_assertion, tolerance, length, Recipient, InResponseTo, Address);
    return noneRef();
}

static PyObject*
saml2_assertion_set_basic_conditions(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    time_t tolerance = -1;
    time_t length = -1;
    gboolean one_time_use = FALSE;

    if (! PyArg_ParseTuple(args, "O|iii", &cvt_saml2_assertion, &tolerance, &length, &one_time_use)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_saml2_assertion_set_basic_conditions(saml2_assertion, tolerance, length, one_time_use);
    return noneRef();
}

static PyObject*
saml2_assertion_add_audience_restriction(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    const char* providerID = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_saml2_assertion, &providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_saml2_assertion_add_audience_restriction(saml2_assertion, providerID);
    return noneRef();
}

static PyObject*
saml2_assertion_add_proxy_limit(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    int proxy_count = -1;
    GList* proxy_audiences = NULL;
    PyObject *cvt_proxy_audiences = NULL;

    if (! PyArg_ParseTuple(args, "O|iO", &cvt_saml2_assertion, &proxy_count, &cvt_proxy_audiences)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    set_list_of_strings(&proxy_audiences, cvt_proxy_audiences);
    lasso_saml2_assertion_add_proxy_limit(saml2_assertion, proxy_count, proxy_audiences);
    return noneRef();
}

static PyObject*
saml2_assertion_validate_conditions(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    const char* relaying_party_providerID = NULL;
    LassoSaml2AssertionValidationState return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|z", &cvt_saml2_assertion, &relaying_party_providerID)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_validate_conditions(saml2_assertion, relaying_party_providerID);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_assertion_get_issuer_provider(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    const LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    const LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    LassoProvider* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_saml2_assertion, &cvt_server)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (const LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (const LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_get_issuer_provider(saml2_assertion, server);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_assertion_add_attribute_with_node(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* assertion = NULL;
    PyGObjectPtr *cvt_assertion = NULL;
    const char* name = NULL;
    const char* nameformat = NULL;
    LassoNode* content = NULL;
    PyGObjectPtr *cvt_content = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OssO", &cvt_assertion, &name, &nameformat, &cvt_content)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_assertion, &PyGObjectPtrType)) {
        assertion = (LassoSaml2Assertion*)cvt_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_content, &PyGObjectPtrType)) {
        content = (LassoNode*)cvt_content->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_add_attribute_with_node(assertion, name, nameformat, content);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_assertion_get_subject_confirmation_data(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    gboolean create;
    LassoSaml2SubjectConfirmationData* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Oi", &cvt_saml2_assertion, &create)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_get_subject_confirmation_data(saml2_assertion, create);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    if (return_value) g_object_unref(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_assertion_get_in_response_to(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* assertion = NULL;
    PyGObjectPtr *cvt_assertion = NULL;
    const char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_assertion)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_assertion, &PyGObjectPtrType)) {
        assertion = (LassoSaml2Assertion*)cvt_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_get_in_response_to(assertion);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
saml2_assertion_decrypt_subject(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* assertion = NULL;
    PyGObjectPtr *cvt_assertion = NULL;
    LassoServer* server = NULL;
    PyGObjectPtr *cvt_server = NULL;
    lasso_error_t return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_assertion, &cvt_server)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_assertion, &PyGObjectPtrType)) {
        assertion = (LassoSaml2Assertion*)cvt_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_server, &PyGObjectPtrType)) {
        server = (LassoServer*)cvt_server->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_decrypt_subject(assertion, server);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_assertion_validate_time_checks(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    unsigned int tolerance;
    time_t now = 0;
    LassoSaml2AssertionValidationState return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Oi|i", &cvt_saml2_assertion, &tolerance, &now)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_validate_time_checks(saml2_assertion, tolerance, now);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_assertion_validate_audience(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    const gchar* audience = NULL;
    LassoSaml2AssertionValidationState return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_saml2_assertion, &audience)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_validate_audience(saml2_assertion, audience);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_assertion_has_one_time_use(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_saml2_assertion)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_has_one_time_use(saml2_assertion);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
saml2_assertion_allows_proxying(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    LassoSaml2AssertionValidationState return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_saml2_assertion)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_allows_proxying(saml2_assertion);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_assertion_allows_proxying_to(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    const char* audience = NULL;
    LassoSaml2AssertionValidationState return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O|z", &cvt_saml2_assertion, &audience)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_assertion_allows_proxying_to(saml2_assertion, audience);
    return_pyvalue = PyInt_FromLong(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_assertion_set_one_time_use(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Assertion* saml2_assertion = NULL;
    PyGObjectPtr *cvt_saml2_assertion = NULL;
    gboolean one_time_use;

    if (! PyArg_ParseTuple(args, "Oi", &cvt_saml2_assertion, &one_time_use)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_saml2_assertion, &PyGObjectPtrType)) {
        saml2_assertion = (LassoSaml2Assertion*)cvt_saml2_assertion->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_saml2_assertion_set_one_time_use(saml2_assertion, one_time_use);
    return noneRef();
}

static PyObject*
Samlp2SubjectQueryAbstract_subject_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Subject* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2SubjectQueryAbstract* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2SubjectQueryAbstract*)cvt_this->obj;
    return_value = this->Subject;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2SubjectQueryAbstract_subject_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2SubjectQueryAbstract* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2SubjectQueryAbstract*)cvt_this->obj;
    set_object_field((GObject**)&this->Subject, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2AuthnQuery_requestedAuthnContext_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2RequestedAuthnContext* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnQuery* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnQuery*)cvt_this->obj;
    return_value = this->RequestedAuthnContext;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2AuthnQuery_requestedAuthnContext_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnQuery* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2AuthnQuery*)cvt_this->obj;
    set_object_field((GObject**)&this->RequestedAuthnContext, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2AuthnQuery_sessionIndex_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnQuery* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthnQuery*)cvt_this->obj;
    return_value = this->SessionIndex;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2AuthnQuery_sessionIndex_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthnQuery* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2AuthnQuery*)cvt_this->obj;
    lasso_assign_string(this->SessionIndex, value);
    return noneRef();
}

static PyObject*
Saml2AttributeStatement_attribute_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AttributeStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AttributeStatement*)cvt_this->obj;
    return_value = this->Attribute;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2AttributeStatement_attribute_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AttributeStatement* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2AttributeStatement*)cvt_this->obj;
    set_list_of_pygobject(&this->Attribute, cvt_value);
    return noneRef();
}

static PyObject*
Saml2AttributeStatement_encryptedAttribute_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AttributeStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AttributeStatement*)cvt_this->obj;
    return_value = this->EncryptedAttribute;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2AttributeStatement_encryptedAttribute_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AttributeStatement* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2AttributeStatement*)cvt_this->obj;
    set_list_of_pygobject(&this->EncryptedAttribute, cvt_value);
    return noneRef();
}

static PyObject*
Saml2AuthzDecisionStatement_action_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Action* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthzDecisionStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AuthzDecisionStatement*)cvt_this->obj;
    return_value = this->Action;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2AuthzDecisionStatement_action_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthzDecisionStatement* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2AuthzDecisionStatement*)cvt_this->obj;
    set_object_field((GObject**)&this->Action, cvt_value);
    return noneRef();
}

static PyObject*
Saml2AuthzDecisionStatement_evidence_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Evidence* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthzDecisionStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AuthzDecisionStatement*)cvt_this->obj;
    return_value = this->Evidence;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2AuthzDecisionStatement_evidence_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthzDecisionStatement* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2AuthzDecisionStatement*)cvt_this->obj;
    set_object_field((GObject**)&this->Evidence, cvt_value);
    return noneRef();
}

static PyObject*
Saml2AuthzDecisionStatement_resource_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthzDecisionStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AuthzDecisionStatement*)cvt_this->obj;
    return_value = this->Resource;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2AuthzDecisionStatement_resource_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthzDecisionStatement* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2AuthzDecisionStatement*)cvt_this->obj;
    lasso_assign_string(this->Resource, value);
    return noneRef();
}

static PyObject*
Saml2AuthzDecisionStatement_decision_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthzDecisionStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AuthzDecisionStatement*)cvt_this->obj;
    return_value = this->Decision;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2AuthzDecisionStatement_decision_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthzDecisionStatement* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2AuthzDecisionStatement*)cvt_this->obj;
    lasso_assign_string(this->Decision, value);
    return noneRef();
}

static PyObject*
Saml2AuthnStatement_subjectLocality_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2SubjectLocality* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AuthnStatement*)cvt_this->obj;
    return_value = this->SubjectLocality;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2AuthnStatement_subjectLocality_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnStatement* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2AuthnStatement*)cvt_this->obj;
    set_object_field((GObject**)&this->SubjectLocality, cvt_value);
    return noneRef();
}

static PyObject*
Saml2AuthnStatement_authnContext_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2AuthnContext* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AuthnStatement*)cvt_this->obj;
    return_value = this->AuthnContext;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Saml2AuthnStatement_authnContext_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnStatement* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2AuthnStatement*)cvt_this->obj;
    set_object_field((GObject**)&this->AuthnContext, cvt_value);
    return noneRef();
}

static PyObject*
Saml2AuthnStatement_authnInstant_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AuthnStatement*)cvt_this->obj;
    return_value = this->AuthnInstant;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2AuthnStatement_authnInstant_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnStatement* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2AuthnStatement*)cvt_this->obj;
    lasso_assign_string(this->AuthnInstant, value);
    return noneRef();
}

static PyObject*
Saml2AuthnStatement_sessionIndex_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AuthnStatement*)cvt_this->obj;
    return_value = this->SessionIndex;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2AuthnStatement_sessionIndex_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnStatement* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2AuthnStatement*)cvt_this->obj;
    lasso_assign_string(this->SessionIndex, value);
    return noneRef();
}

static PyObject*
Saml2AuthnStatement_sessionNotOnOrAfter_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnStatement* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AuthnStatement*)cvt_this->obj;
    return_value = this->SessionNotOnOrAfter;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2AuthnStatement_sessionNotOnOrAfter_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AuthnStatement* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2AuthnStatement*)cvt_this->obj;
    lasso_assign_string(this->SessionNotOnOrAfter, value);
    return noneRef();
}

static PyObject*
Samlp2Scoping_idpList_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2IDPList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2Scoping* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2Scoping*)cvt_this->obj;
    return_value = this->IDPList;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2Scoping_idpList_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2Scoping* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2Scoping*)cvt_this->obj;
    set_object_field((GObject**)&this->IDPList, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2Scoping_requesterId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2Scoping* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2Scoping*)cvt_this->obj;
    return_value = this->RequesterID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2Scoping_requesterId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2Scoping* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2Scoping*)cvt_this->obj;
    lasso_assign_string(this->RequesterID, value);
    return noneRef();
}

static PyObject*
Samlp2Scoping_proxyCount_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2Scoping* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2Scoping*)cvt_this->obj;
    return_value = this->ProxyCount;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2Scoping_proxyCount_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2Scoping* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2Scoping*)cvt_this->obj;
    lasso_assign_string(this->ProxyCount, value);
    return noneRef();
}

static PyObject*
Saml2AttributeValue_any_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2AttributeValue* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2AttributeValue*)cvt_this->obj;
    return_value = this->any;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2AttributeValue_any_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2AttributeValue* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2AttributeValue*)cvt_this->obj;
    set_list_of_pygobject(&this->any, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2IDPEntry_providerId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2IDPEntry* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2IDPEntry*)cvt_this->obj;
    return_value = this->ProviderID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2IDPEntry_providerId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2IDPEntry* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2IDPEntry*)cvt_this->obj;
    lasso_assign_string(this->ProviderID, value);
    return noneRef();
}

static PyObject*
Samlp2IDPEntry_name_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2IDPEntry* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2IDPEntry*)cvt_this->obj;
    return_value = this->Name;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2IDPEntry_name_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2IDPEntry* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2IDPEntry*)cvt_this->obj;
    lasso_assign_string(this->Name, value);
    return noneRef();
}

static PyObject*
Samlp2IDPEntry_loc_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2IDPEntry* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2IDPEntry*)cvt_this->obj;
    return_value = this->Loc;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2IDPEntry_loc_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2IDPEntry* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2IDPEntry*)cvt_this->obj;
    lasso_assign_string(this->Loc, value);
    return noneRef();
}

static PyObject*
Samlp2Response_assertion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2Response* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2Response*)cvt_this->obj;
    return_value = this->Assertion;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Samlp2Response_assertion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2Response* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2Response*)cvt_this->obj;
    set_list_of_pygobject(&this->Assertion, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2Response_encryptedAssertion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2Response* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2Response*)cvt_this->obj;
    return_value = this->EncryptedAssertion;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Samlp2Response_encryptedAssertion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2Response* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2Response*)cvt_this->obj;
    set_list_of_pygobject(&this->EncryptedAssertion, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Conditions_condition_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Conditions* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Conditions*)cvt_this->obj;
    return_value = this->Condition;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Conditions_condition_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Conditions* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Conditions*)cvt_this->obj;
    set_list_of_pygobject(&this->Condition, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Conditions_audienceRestriction_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Conditions* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Conditions*)cvt_this->obj;
    return_value = this->AudienceRestriction;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Conditions_audienceRestriction_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Conditions* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Conditions*)cvt_this->obj;
    set_list_of_pygobject(&this->AudienceRestriction, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Conditions_oneTimeUse_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Conditions* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Conditions*)cvt_this->obj;
    return_value = this->OneTimeUse;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Conditions_oneTimeUse_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Conditions* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Conditions*)cvt_this->obj;
    set_list_of_pygobject(&this->OneTimeUse, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Conditions_proxyRestriction_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Conditions* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Conditions*)cvt_this->obj;
    return_value = this->ProxyRestriction;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Conditions_proxyRestriction_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Conditions* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Conditions*)cvt_this->obj;
    set_list_of_pygobject(&this->ProxyRestriction, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Conditions_notBefore_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Conditions* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Conditions*)cvt_this->obj;
    return_value = this->NotBefore;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2Conditions_notBefore_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Conditions* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2Conditions*)cvt_this->obj;
    lasso_assign_string(this->NotBefore, value);
    return noneRef();
}

static PyObject*
Saml2Conditions_notOnOrAfter_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Conditions* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Conditions*)cvt_this->obj;
    return_value = this->NotOnOrAfter;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2Conditions_notOnOrAfter_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Conditions* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2Conditions*)cvt_this->obj;
    lasso_assign_string(this->NotOnOrAfter, value);
    return noneRef();
}

static PyObject*
Samlp2ArtifactResolve_artifact_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2ArtifactResolve* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2ArtifactResolve*)cvt_this->obj;
    return_value = this->Artifact;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2ArtifactResolve_artifact_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2ArtifactResolve* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2ArtifactResolve*)cvt_this->obj;
    lasso_assign_string(this->Artifact, value);
    return noneRef();
}

static PyObject*
Samlp2LogoutRequest_baseId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2BaseIDAbstract* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2LogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2LogoutRequest*)cvt_this->obj;
    return_value = this->BaseID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2LogoutRequest_baseId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2LogoutRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2LogoutRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->BaseID, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2LogoutRequest_nameId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2NameID* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2LogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2LogoutRequest*)cvt_this->obj;
    return_value = this->NameID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2LogoutRequest_nameId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2LogoutRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2LogoutRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->NameID, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2LogoutRequest_encryptedId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2EncryptedElement* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2LogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2LogoutRequest*)cvt_this->obj;
    return_value = this->EncryptedID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2LogoutRequest_encryptedId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2LogoutRequest* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2LogoutRequest*)cvt_this->obj;
    set_object_field((GObject**)&this->EncryptedID, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2LogoutRequest_sessionIndex_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2LogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2LogoutRequest*)cvt_this->obj;
    return_value = this->SessionIndex;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2LogoutRequest_sessionIndex_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2LogoutRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2LogoutRequest*)cvt_this->obj;
    lasso_assign_string(this->SessionIndex, value);
    return noneRef();
}

static PyObject*
Samlp2LogoutRequest_reason_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2LogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2LogoutRequest*)cvt_this->obj;
    return_value = this->Reason;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2LogoutRequest_reason_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2LogoutRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2LogoutRequest*)cvt_this->obj;
    lasso_assign_string(this->Reason, value);
    return noneRef();
}

static PyObject*
Samlp2LogoutRequest_notOnOrAfter_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2LogoutRequest* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2LogoutRequest*)cvt_this->obj;
    return_value = this->NotOnOrAfter;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2LogoutRequest_notOnOrAfter_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2LogoutRequest* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2LogoutRequest*)cvt_this->obj;
    lasso_assign_string(this->NotOnOrAfter, value);
    return noneRef();
}

static PyObject*
samlp2_logout_request_get_session_indexes(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2LogoutRequest* logout_request = NULL;
    PyGObjectPtr *cvt_logout_request = NULL;
    GList* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_logout_request)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_logout_request, &PyGObjectPtrType)) {
        logout_request = (LassoSamlp2LogoutRequest*)cvt_logout_request->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_samlp2_logout_request_get_session_indexes(logout_request);
    return_pyvalue = get_list_of_strings(return_value);
    lasso_release_list_of_strings(return_value);
    return return_pyvalue;
}

static PyObject*
samlp2_logout_request_set_session_indexes(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSamlp2LogoutRequest* logout_request = NULL;
    PyGObjectPtr *cvt_logout_request = NULL;
    GList* session_index = NULL;
    PyObject *cvt_session_index = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_logout_request, &cvt_session_index)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_logout_request, &PyGObjectPtrType)) {
        logout_request = (LassoSamlp2LogoutRequest*)cvt_logout_request->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    set_list_of_strings(&session_index, cvt_session_index);
    lasso_samlp2_logout_request_set_session_indexes(logout_request, session_index);
    free_list(&session_index, (GFunc)g_free);
    return noneRef();
}

static PyObject*
Samlp2AttributeQuery_attribute_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AttributeQuery* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AttributeQuery*)cvt_this->obj;
    return_value = this->Attribute;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Samlp2AttributeQuery_attribute_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AttributeQuery* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2AttributeQuery*)cvt_this->obj;
    set_list_of_pygobject(&this->Attribute, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Evidence_assertionIdRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Evidence* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Evidence*)cvt_this->obj;
    return_value = this->AssertionIDRef;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Evidence_assertionIdRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Evidence* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Evidence*)cvt_this->obj;
    set_list_of_pygobject(&this->AssertionIDRef, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Evidence_assertionUriRef_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Evidence* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Evidence*)cvt_this->obj;
    return_value = this->AssertionURIRef;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Evidence_assertionUriRef_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Evidence* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Evidence*)cvt_this->obj;
    set_list_of_pygobject(&this->AssertionURIRef, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Evidence_assertion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Evidence* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Evidence*)cvt_this->obj;
    return_value = this->Assertion;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Evidence_assertion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Evidence* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Evidence*)cvt_this->obj;
    set_list_of_pygobject(&this->Assertion, cvt_value);
    return noneRef();
}

static PyObject*
Saml2Evidence_encryptedAssertion_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    GList* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2Evidence* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2Evidence*)cvt_this->obj;
    return_value = this->EncryptedAssertion;
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
Saml2Evidence_encryptedAssertion_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2Evidence* this;
    PyObject *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSaml2Evidence*)cvt_this->obj;
    set_list_of_pygobject(&this->EncryptedAssertion, cvt_value);
    return noneRef();
}

static PyObject*
Saml2NameID_content_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2NameID* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2NameID*)cvt_this->obj;
    return_value = this->content;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2NameID_content_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2NameID* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2NameID*)cvt_this->obj;
    lasso_assign_string(this->content, value);
    return noneRef();
}

static PyObject*
Saml2NameID_format_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2NameID* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2NameID*)cvt_this->obj;
    return_value = this->Format;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2NameID_format_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2NameID* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2NameID*)cvt_this->obj;
    lasso_assign_string(this->Format, value);
    return noneRef();
}

static PyObject*
Saml2NameID_spProvidedId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2NameID* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2NameID*)cvt_this->obj;
    return_value = this->SPProvidedID;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2NameID_spProvidedId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2NameID* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2NameID*)cvt_this->obj;
    lasso_assign_string(this->SPProvidedID, value);
    return noneRef();
}

static PyObject*
Saml2NameID_nameQualifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2NameID* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2NameID*)cvt_this->obj;
    return_value = this->NameQualifier;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2NameID_nameQualifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2NameID* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2NameID*)cvt_this->obj;
    lasso_assign_string(this->NameQualifier, value);
    return noneRef();
}

static PyObject*
Saml2NameID_spNameQualifier_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2NameID* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2NameID*)cvt_this->obj;
    return_value = this->SPNameQualifier;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2NameID_spNameQualifier_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2NameID* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2NameID*)cvt_this->obj;
    lasso_assign_string(this->SPNameQualifier, value);
    return noneRef();
}

static PyObject*
saml2_name_id_equals(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2NameID* name_id = NULL;
    PyGObjectPtr *cvt_name_id = NULL;
    LassoSaml2NameID* other_name_id = NULL;
    PyGObjectPtr *cvt_other_name_id = NULL;
    gboolean return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_name_id, &cvt_other_name_id)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_name_id, &PyGObjectPtrType)) {
        name_id = (LassoSaml2NameID*)cvt_name_id->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_other_name_id, &PyGObjectPtrType)) {
        other_name_id = (LassoSaml2NameID*)cvt_other_name_id->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_name_id_equals(name_id, other_name_id);
    if (return_value) {
        Py_INCREF(Py_True);
        return_pyvalue = Py_True;
    } else {
        Py_INCREF(Py_False);
        return_pyvalue = Py_False;
    }
    return return_pyvalue;
}

static PyObject*
Samlp2NameIDMappingResponse_nameId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2NameID* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDMappingResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2NameIDMappingResponse*)cvt_this->obj;
    return_value = this->NameID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2NameIDMappingResponse_nameId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDMappingResponse* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2NameIDMappingResponse*)cvt_this->obj;
    set_object_field((GObject**)&this->NameID, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2NameIDMappingResponse_encryptedId_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2EncryptedElement* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDMappingResponse* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2NameIDMappingResponse*)cvt_this->obj;
    return_value = this->EncryptedID;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2NameIDMappingResponse_encryptedId_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2NameIDMappingResponse* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2NameIDMappingResponse*)cvt_this->obj;
    set_object_field((GObject**)&this->EncryptedID, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2AuthzDecisionQuery_action_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Action* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthzDecisionQuery* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthzDecisionQuery*)cvt_this->obj;
    return_value = this->Action;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2AuthzDecisionQuery_action_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthzDecisionQuery* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2AuthzDecisionQuery*)cvt_this->obj;
    set_object_field((GObject**)&this->Action, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2AuthzDecisionQuery_evidence_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2Evidence* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthzDecisionQuery* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthzDecisionQuery*)cvt_this->obj;
    return_value = this->Evidence;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
Samlp2AuthzDecisionQuery_evidence_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthzDecisionQuery* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoSamlp2AuthzDecisionQuery*)cvt_this->obj;
    set_object_field((GObject**)&this->Evidence, cvt_value);
    return noneRef();
}

static PyObject*
Samlp2AuthzDecisionQuery_resource_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthzDecisionQuery* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSamlp2AuthzDecisionQuery*)cvt_this->obj;
    return_value = this->Resource;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Samlp2AuthzDecisionQuery_resource_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSamlp2AuthzDecisionQuery* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSamlp2AuthzDecisionQuery*)cvt_this->obj;
    lasso_assign_string(this->Resource, value);
    return noneRef();
}

static PyObject*
Saml2SubjectConfirmationData_notBefore_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmationData* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2SubjectConfirmationData*)cvt_this->obj;
    return_value = this->NotBefore;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2SubjectConfirmationData_notBefore_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmationData* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2SubjectConfirmationData*)cvt_this->obj;
    lasso_assign_string(this->NotBefore, value);
    return noneRef();
}

static PyObject*
Saml2SubjectConfirmationData_notOnOrAfter_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmationData* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2SubjectConfirmationData*)cvt_this->obj;
    return_value = this->NotOnOrAfter;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2SubjectConfirmationData_notOnOrAfter_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmationData* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2SubjectConfirmationData*)cvt_this->obj;
    lasso_assign_string(this->NotOnOrAfter, value);
    return noneRef();
}

static PyObject*
Saml2SubjectConfirmationData_recipient_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmationData* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2SubjectConfirmationData*)cvt_this->obj;
    return_value = this->Recipient;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2SubjectConfirmationData_recipient_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmationData* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2SubjectConfirmationData*)cvt_this->obj;
    lasso_assign_string(this->Recipient, value);
    return noneRef();
}

static PyObject*
Saml2SubjectConfirmationData_inResponseTo_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmationData* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2SubjectConfirmationData*)cvt_this->obj;
    return_value = this->InResponseTo;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2SubjectConfirmationData_inResponseTo_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmationData* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2SubjectConfirmationData*)cvt_this->obj;
    lasso_assign_string(this->InResponseTo, value);
    return noneRef();
}

static PyObject*
Saml2SubjectConfirmationData_address_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmationData* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoSaml2SubjectConfirmationData*)cvt_this->obj;
    return_value = this->Address;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
Saml2SubjectConfirmationData_address_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoSaml2SubjectConfirmationData* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoSaml2SubjectConfirmationData*)cvt_this->obj;
    lasso_assign_string(this->Address, value);
    return noneRef();
}

static PyObject*
saml2_key_info_confirmation_data_type_get_key_info(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2KeyInfoConfirmationDataType* kicdt = NULL;
    PyGObjectPtr *cvt_kicdt = NULL;
    GList* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_kicdt)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_kicdt, &PyGObjectPtrType)) {
        kicdt = (LassoSaml2KeyInfoConfirmationDataType*)cvt_kicdt->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_saml2_key_info_confirmation_data_type_get_key_info(kicdt);
    return_pyvalue = get_list_of_pygobject(return_value);
    return return_pyvalue;
}

static PyObject*
saml2_key_info_confirmation_data_type_set_key_info(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoSaml2KeyInfoConfirmationDataType* kicdt = NULL;
    PyGObjectPtr *cvt_kicdt = NULL;
    GList* key_infos = NULL;
    PyObject *cvt_key_infos = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_kicdt, &cvt_key_infos)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_kicdt, &PyGObjectPtrType)) {
        kicdt = (LassoSaml2KeyInfoConfirmationDataType*)cvt_kicdt->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    set_list_of_pygobject(&key_infos, cvt_key_infos);
    lasso_saml2_key_info_confirmation_data_type_set_key_info(kicdt, key_infos);
    return noneRef();
}

static PyObject*
DsKeyInfo_keyValue_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsKeyValue* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoDsKeyInfo* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoDsKeyInfo*)cvt_this->obj;
    return_value = this->KeyValue;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
DsKeyInfo_keyValue_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoDsKeyInfo* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoDsKeyInfo*)cvt_this->obj;
    set_object_field((GObject**)&this->KeyValue, cvt_value);
    return noneRef();
}

static PyObject*
DsRsaKeyValue_modulus_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoDsRsaKeyValue* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoDsRsaKeyValue*)cvt_this->obj;
    return_value = this->Modulus;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
DsRsaKeyValue_modulus_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoDsRsaKeyValue* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoDsRsaKeyValue*)cvt_this->obj;
    lasso_assign_string(this->Modulus, value);
    return noneRef();
}

static PyObject*
DsRsaKeyValue_exponent_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    char* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoDsRsaKeyValue* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoDsRsaKeyValue*)cvt_this->obj;
    return_value = this->Exponent;
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
DsRsaKeyValue_exponent_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoDsRsaKeyValue* this;
    char* value;
    if (! PyArg_ParseTuple(args, "Oz", &cvt_this, &value)) return NULL;
    this = (LassoDsRsaKeyValue*)cvt_this->obj;
    lasso_assign_string(this->Exponent, value);
    return noneRef();
}

static PyObject*
ds_x509_data_get_certificate(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsX509Data* x509_data = NULL;
    PyGObjectPtr *cvt_x509_data = NULL;
    const char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_x509_data)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_x509_data, &PyGObjectPtrType)) {
        x509_data = (LassoDsX509Data*)cvt_x509_data->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_ds_x509_data_get_certificate(x509_data);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
ds_x509_data_set_certificate(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsX509Data* x509_data = NULL;
    PyGObjectPtr *cvt_x509_data = NULL;
    const char* certificate = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_x509_data, &certificate)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_x509_data, &PyGObjectPtrType)) {
        x509_data = (LassoDsX509Data*)cvt_x509_data->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_ds_x509_data_set_certificate(x509_data, certificate);
    return noneRef();
}

static PyObject*
ds_x509_data_get_subject_name(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsX509Data* x509_data = NULL;
    PyGObjectPtr *cvt_x509_data = NULL;
    const char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_x509_data)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_x509_data, &PyGObjectPtrType)) {
        x509_data = (LassoDsX509Data*)cvt_x509_data->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_ds_x509_data_get_subject_name(x509_data);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
ds_x509_data_set_subject_name(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsX509Data* x509_data = NULL;
    PyGObjectPtr *cvt_x509_data = NULL;
    const char* subject_name = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_x509_data, &subject_name)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_x509_data, &PyGObjectPtrType)) {
        x509_data = (LassoDsX509Data*)cvt_x509_data->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_ds_x509_data_set_subject_name(x509_data, subject_name);
    return noneRef();
}

static PyObject*
ds_x509_data_get_crl(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsX509Data* x509_data = NULL;
    PyGObjectPtr *cvt_x509_data = NULL;
    const char* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_x509_data)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_x509_data, &PyGObjectPtrType)) {
        x509_data = (LassoDsX509Data*)cvt_x509_data->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_ds_x509_data_get_crl(x509_data);
    if (return_value) {
        return_pyvalue = PyString_FromString(return_value);
    } else {
        return_pyvalue = noneRef();
    }
    return return_pyvalue;
}

static PyObject*
ds_x509_data_set_crl(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsX509Data* x509_data = NULL;
    PyGObjectPtr *cvt_x509_data = NULL;
    const char* crl = NULL;

    if (! PyArg_ParseTuple(args, "Os", &cvt_x509_data, &crl)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_x509_data, &PyGObjectPtrType)) {
        x509_data = (LassoDsX509Data*)cvt_x509_data->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_ds_x509_data_set_crl(x509_data, crl);
    return noneRef();
}

static PyObject*
DsKeyValue_rsaKeyValue_get(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsRsaKeyValue* return_value;
    PyObject* return_pyvalue;
    PyGObjectPtr* cvt_this;
    LassoDsKeyValue* this;

    if (! PyArg_ParseTuple(args, "O", &cvt_this)) return NULL;
    this = (LassoDsKeyValue*)cvt_this->obj;
    return_value = this->RSAKeyValue;
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
DsKeyValue_rsaKeyValue_set(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    PyGObjectPtr* cvt_this;
    LassoDsKeyValue* this;
    PyGObjectPtr *cvt_value;
    if (! PyArg_ParseTuple(args, "OO", &cvt_this, &cvt_value)) return NULL;
    this = (LassoDsKeyValue*)cvt_this->obj;
    set_object_field((GObject**)&this->RSAKeyValue, cvt_value);
    return noneRef();
}

static PyObject*
ds_key_value_get_x509_data(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsKeyValue* key_value = NULL;
    PyGObjectPtr *cvt_key_value = NULL;
    LassoDsX509Data* return_value;
    PyObject* return_pyvalue = NULL;

    if (! PyArg_ParseTuple(args, "O", &cvt_key_value)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_key_value, &PyGObjectPtrType)) {
        key_value = (LassoDsKeyValue*)cvt_key_value->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    return_value = lasso_ds_key_value_get_x509_data(key_value);
    if (return_value) {
        return_pyvalue = PyGObjectPtr_New(G_OBJECT(return_value));
    } else {
        return_pyvalue = noneRef();
    }

    return return_pyvalue;
}

static PyObject*
ds_key_value_set_x509_data(G_GNUC_UNUSED PyObject *self, PyObject *args)
{
    LassoDsKeyValue* key_value = NULL;
    PyGObjectPtr *cvt_key_value = NULL;
    LassoDsX509Data* x509_data = NULL;
    PyGObjectPtr *cvt_x509_data = NULL;

    if (! PyArg_ParseTuple(args, "OO", &cvt_key_value, &cvt_x509_data)) return NULL;
    if (PyObject_TypeCheck((PyObject*)cvt_key_value, &PyGObjectPtrType)) {
        key_value = (LassoDsKeyValue*)cvt_key_value->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    if (PyObject_TypeCheck((PyObject*)cvt_x509_data, &PyGObjectPtrType)) {
        x509_data = (LassoDsX509Data*)cvt_x509_data->obj;
    } else {
        PyErr_SetString(PyExc_TypeError, "value should be a PyGObject");
        return NULL;
    }
    lasso_ds_key_value_set_x509_data(key_value, x509_data);
    return noneRef();
}


static PyMethodDef lasso_methods[] = {
    {"key_new_for_signature_from_base64_string", key_new_for_signature_from_base64_string, METH_VARARGS, NULL},
    {"key_new_for_signature_from_file", key_new_for_signature_from_file, METH_VARARGS, NULL},
    {"strError", strError, METH_VARARGS, NULL},
    {"init", init, METH_VARARGS, NULL},
    {"shutdown", shutdown, METH_VARARGS, NULL},
    {"check_version", check_version, METH_VARARGS, NULL},
    {"set_flag", set_flag, METH_VARARGS, NULL},
    {"federation_new", federation_new, METH_VARARGS, NULL},
    {"get_request_type_from_soap_msg", get_request_type_from_soap_msg, METH_VARARGS, NULL},
    {"isLibertyQuery", isLibertyQuery, METH_VARARGS, NULL},
    {"server_new", server_new, METH_VARARGS, NULL},
    {"server_new_from_buffers", server_new_from_buffers, METH_VARARGS, NULL},
    {"server_new_from_dump", server_new_from_dump, METH_VARARGS, NULL},
    {"session_new", session_new, METH_VARARGS, NULL},
    {"session_new_from_dump", session_new_from_dump, METH_VARARGS, NULL},
    {"provider_new", provider_new, METH_VARARGS, NULL},
    {"provider_new_from_buffer", provider_new_from_buffer, METH_VARARGS, NULL},
    {"provider_new_from_dump", provider_new_from_dump, METH_VARARGS, NULL},
    {"name_registration_new", name_registration_new, METH_VARARGS, NULL},
    {"name_registration_new_from_dump", name_registration_new_from_dump, METH_VARARGS, NULL},
    {"login_new", login_new, METH_VARARGS, NULL},
    {"login_new_from_dump", login_new_from_dump, METH_VARARGS, NULL},
    {"lecp_new", lecp_new, METH_VARARGS, NULL},
    {"identity_new", identity_new, METH_VARARGS, NULL},
    {"identity_new_from_dump", identity_new_from_dump, METH_VARARGS, NULL},
    {"logout_new", logout_new, METH_VARARGS, NULL},
    {"logout_new_from_dump", logout_new_from_dump, METH_VARARGS, NULL},
    {"name_identifier_mapping_new", name_identifier_mapping_new, METH_VARARGS, NULL},
    {"defederation_new", defederation_new, METH_VARARGS, NULL},
    {"isSamlQuery", isSamlQuery, METH_VARARGS, NULL},
    {"ecp_new", ecp_new, METH_VARARGS, NULL},
    {"assertion_query_new", assertion_query_new, METH_VARARGS, NULL},
    {"name_id_management_new", name_id_management_new, METH_VARARGS, NULL},
    {"name_id_management_new_from_dump", name_id_management_new_from_dump, METH_VARARGS, NULL},
    {"saml2_name_id_build_persistent", saml2_name_id_build_persistent, METH_VARARGS, NULL},
    {"saml2_encrypted_element_build_encrypted_persistent_name_id", saml2_encrypted_element_build_encrypted_persistent_name_id, METH_VARARGS, NULL},
    {"saml2_name_id_new_with_persistent_format", saml2_name_id_new_with_persistent_format, METH_VARARGS, NULL},
    {"lib_authn_request_envelope_new", lib_authn_request_envelope_new, METH_VARARGS, NULL},
    {"lib_authn_request_envelope_new_full", lib_authn_request_envelope_new_full, METH_VARARGS, NULL},
    {"lib_authn_response_envelope_new", lib_authn_response_envelope_new, METH_VARARGS, NULL},
    {"saml_attribute_statement_new", saml_attribute_statement_new, METH_VARARGS, NULL},
    {"lib_name_identifier_mapping_request_new", lib_name_identifier_mapping_request_new, METH_VARARGS, NULL},
    {"lib_name_identifier_mapping_request_new_full", lib_name_identifier_mapping_request_new_full, METH_VARARGS, NULL},
    {"build_unique_id", build_unique_id, METH_VARARGS, NULL},
    {"lib_subject_new", lib_subject_new, METH_VARARGS, NULL},
    {"saml_attribute_value_new", saml_attribute_value_new, METH_VARARGS, NULL},
    {"lib_authentication_statement_new", lib_authentication_statement_new, METH_VARARGS, NULL},
    {"lib_authentication_statement_new_full", lib_authentication_statement_new_full, METH_VARARGS, NULL},
    {"saml_subject_statement_new", saml_subject_statement_new, METH_VARARGS, NULL},
    {"lib_register_name_identifier_response_new", lib_register_name_identifier_response_new, METH_VARARGS, NULL},
    {"lib_register_name_identifier_response_new_full", lib_register_name_identifier_response_new_full, METH_VARARGS, NULL},
    {"lib_federation_termination_notification_new", lib_federation_termination_notification_new, METH_VARARGS, NULL},
    {"lib_federation_termination_notification_new_full", lib_federation_termination_notification_new_full, METH_VARARGS, NULL},
    {"lib_authn_context_new", lib_authn_context_new, METH_VARARGS, NULL},
    {"saml_attribute_designator_new", saml_attribute_designator_new, METH_VARARGS, NULL},
    {"lib_request_authn_context_new", lib_request_authn_context_new, METH_VARARGS, NULL},
    {"saml_subject_new", saml_subject_new, METH_VARARGS, NULL},
    {"lib_authn_response_new", lib_authn_response_new, METH_VARARGS, NULL},
    {"lib_name_identifier_mapping_response_new", lib_name_identifier_mapping_response_new, METH_VARARGS, NULL},
    {"lib_name_identifier_mapping_response_new_full", lib_name_identifier_mapping_response_new_full, METH_VARARGS, NULL},
    {"saml_attribute_new", saml_attribute_new, METH_VARARGS, NULL},
    {"saml_assertion_new", saml_assertion_new, METH_VARARGS, NULL},
    {"lib_idp_list_new", lib_idp_list_new, METH_VARARGS, NULL},
    {"lib_authn_request_new", lib_authn_request_new, METH_VARARGS, NULL},
    {"lib_idp_entry_new", lib_idp_entry_new, METH_VARARGS, NULL},
    {"samlp_status_code_new", samlp_status_code_new, METH_VARARGS, NULL},
    {"saml_advice_new", saml_advice_new, METH_VARARGS, NULL},
    {"samlp_response_new", samlp_response_new, METH_VARARGS, NULL},
    {"saml_conditions_new", saml_conditions_new, METH_VARARGS, NULL},
    {"lib_logout_request_new", lib_logout_request_new, METH_VARARGS, NULL},
    {"lib_logout_request_new_full", lib_logout_request_new_full, METH_VARARGS, NULL},
    {"lib_assertion_new", lib_assertion_new, METH_VARARGS, NULL},
    {"lib_assertion_new_full", lib_assertion_new_full, METH_VARARGS, NULL},
    {"samlp_request_new", samlp_request_new, METH_VARARGS, NULL},
    {"saml_subject_locality_new", saml_subject_locality_new, METH_VARARGS, NULL},
    {"lib_idp_entries_new", lib_idp_entries_new, METH_VARARGS, NULL},
    {"lib_logout_response_new", lib_logout_response_new, METH_VARARGS, NULL},
    {"lib_logout_response_new_full", lib_logout_response_new_full, METH_VARARGS, NULL},
    {"lib_status_response_new", lib_status_response_new, METH_VARARGS, NULL},
    {"saml_name_identifier_new", saml_name_identifier_new, METH_VARARGS, NULL},
    {"saml_name_identifier_new_from_xmlNode", saml_name_identifier_new_from_xmlNode, METH_VARARGS, NULL},
    {"misc_text_node_new", misc_text_node_new, METH_VARARGS, NULL},
    {"misc_text_node_new_with_string", misc_text_node_new_with_string, METH_VARARGS, NULL},
    {"misc_text_node_new_with_xml_node", misc_text_node_new_with_xml_node, METH_VARARGS, NULL},
    {"saml_audience_restriction_condition_new", saml_audience_restriction_condition_new, METH_VARARGS, NULL},
    {"saml_audience_restriction_condition_new_full", saml_audience_restriction_condition_new_full, METH_VARARGS, NULL},
    {"saml_subject_confirmation_new", saml_subject_confirmation_new, METH_VARARGS, NULL},
    {"saml_authority_binding_new", saml_authority_binding_new, METH_VARARGS, NULL},
    {"lib_scoping_new", lib_scoping_new, METH_VARARGS, NULL},
    {"samlp_status_new", samlp_status_new, METH_VARARGS, NULL},
    {"lib_register_name_identifier_request_new", lib_register_name_identifier_request_new, METH_VARARGS, NULL},
    {"lib_register_name_identifier_request_new_full", lib_register_name_identifier_request_new_full, METH_VARARGS, NULL},
    {"saml_authentication_statement_new", saml_authentication_statement_new, METH_VARARGS, NULL},
    {"node_new", node_new, METH_VARARGS, NULL},
    {"node_new_from_dump", node_new_from_dump, METH_VARARGS, NULL},
    {"node_new_from_soap", node_new_from_soap, METH_VARARGS, NULL},
    {"node_new_from_xmlNode", node_new_from_xmlNode, METH_VARARGS, NULL},
    {"register_dst_service", register_dst_service, METH_VARARGS, NULL},
    {"get_prefix_for_dst_service_href", get_prefix_for_dst_service_href, METH_VARARGS, NULL},
    {"register_idwsf2_dst_service", register_idwsf2_dst_service, METH_VARARGS, NULL},
    {"get_prefix_for_idwsf2_dst_service_href", get_prefix_for_idwsf2_dst_service_href, METH_VARARGS, NULL},
    {"soap_fault_new", soap_fault_new, METH_VARARGS, NULL},
    {"soap_fault_new_from_message", soap_fault_new_from_message, METH_VARARGS, NULL},
    {"soap_fault_new_full", soap_fault_new_full, METH_VARARGS, NULL},
    {"soap_envelope_new", soap_envelope_new, METH_VARARGS, NULL},
    {"soap_envelope_new_from_message", soap_envelope_new_from_message, METH_VARARGS, NULL},
    {"soap_header_new", soap_header_new, METH_VARARGS, NULL},
    {"soap_header_new_from_message", soap_header_new_from_message, METH_VARARGS, NULL},
    {"soap_body_new", soap_body_new, METH_VARARGS, NULL},
    {"soap_body_new_from_message", soap_body_new_from_message, METH_VARARGS, NULL},
    {"soap_detail_new", soap_detail_new, METH_VARARGS, NULL},
    {"soap_detail_new_from_message", soap_detail_new_from_message, METH_VARARGS, NULL},
    {"saml2_attribute_statement_new", saml2_attribute_statement_new, METH_VARARGS, NULL},
    {"saml2_subject_confirmation_new", saml2_subject_confirmation_new, METH_VARARGS, NULL},
    {"saml2_authz_decision_statement_new", saml2_authz_decision_statement_new, METH_VARARGS, NULL},
    {"samlp2_idp_list_new", samlp2_idp_list_new, METH_VARARGS, NULL},
    {"saml2_condition_abstract_new", saml2_condition_abstract_new, METH_VARARGS, NULL},
    {"samlp2_assertion_id_request_new", samlp2_assertion_id_request_new, METH_VARARGS, NULL},
    {"saml2_key_info_confirmation_data_new", saml2_key_info_confirmation_data_new, METH_VARARGS, NULL},
    {"saml2_authn_context_new", saml2_authn_context_new, METH_VARARGS, NULL},
    {"saml2_key_info_confirmation_data_type_new", saml2_key_info_confirmation_data_type_new, METH_VARARGS, NULL},
    {"samlp2_extensions_new", samlp2_extensions_new, METH_VARARGS, NULL},
    {"saml2_base_idabstract_new", saml2_base_idabstract_new, METH_VARARGS, NULL},
    {"saml2_subject_locality_new", saml2_subject_locality_new, METH_VARARGS, NULL},
    {"saml2_authn_statement_new", saml2_authn_statement_new, METH_VARARGS, NULL},
    {"samlp2_status_new", samlp2_status_new, METH_VARARGS, NULL},
    {"samlp2_request_abstract_new", samlp2_request_abstract_new, METH_VARARGS, NULL},
    {"samlp2_name_id_policy_new", samlp2_name_id_policy_new, METH_VARARGS, NULL},
    {"samlp2_manage_name_id_response_new", samlp2_manage_name_id_response_new, METH_VARARGS, NULL},
    {"samlp2_artifact_response_new", samlp2_artifact_response_new, METH_VARARGS, NULL},
    {"samlp2_authn_request_new", samlp2_authn_request_new, METH_VARARGS, NULL},
    {"samlp2_authn_query_new", samlp2_authn_query_new, METH_VARARGS, NULL},
    {"saml2_proxy_restriction_new", saml2_proxy_restriction_new, METH_VARARGS, NULL},
    {"saml2_subject_new", saml2_subject_new, METH_VARARGS, NULL},
    {"saml2_one_time_use_new", saml2_one_time_use_new, METH_VARARGS, NULL},
    {"samlp2_requested_authn_context_new", samlp2_requested_authn_context_new, METH_VARARGS, NULL},
    {"samlp2_status_response_new", samlp2_status_response_new, METH_VARARGS, NULL},
    {"saml2_audience_restriction_new", saml2_audience_restriction_new, METH_VARARGS, NULL},
    {"saml2_attribute_new", saml2_attribute_new, METH_VARARGS, NULL},
    {"samlp2_name_id_mapping_request_new", samlp2_name_id_mapping_request_new, METH_VARARGS, NULL},
    {"samlp2_manage_name_id_request_new", samlp2_manage_name_id_request_new, METH_VARARGS, NULL},
    {"samlp2_status_code_new", samlp2_status_code_new, METH_VARARGS, NULL},
    {"samlp2_terminate_new", samlp2_terminate_new, METH_VARARGS, NULL},
    {"saml2_advice_new", saml2_advice_new, METH_VARARGS, NULL},
    {"saml2_action_new", saml2_action_new, METH_VARARGS, NULL},
    {"saml2_action_new_with_string", saml2_action_new_with_string, METH_VARARGS, NULL},
    {"samlp2_status_detail_new", samlp2_status_detail_new, METH_VARARGS, NULL},
    {"saml2_encrypted_element_new", saml2_encrypted_element_new, METH_VARARGS, NULL},
    {"saml2_assertion_new", saml2_assertion_new, METH_VARARGS, NULL},
    {"samlp2_subject_query_abstract_new", samlp2_subject_query_abstract_new, METH_VARARGS, NULL},
    {"samlp2_logout_response_new", samlp2_logout_response_new, METH_VARARGS, NULL},
    {"saml2_statement_abstract_new", saml2_statement_abstract_new, METH_VARARGS, NULL},
    {"samlp2_scoping_new", samlp2_scoping_new, METH_VARARGS, NULL},
    {"saml2_attribute_value_new", saml2_attribute_value_new, METH_VARARGS, NULL},
    {"samlp2_idp_entry_new", samlp2_idp_entry_new, METH_VARARGS, NULL},
    {"samlp2_response_new", samlp2_response_new, METH_VARARGS, NULL},
    {"saml2_conditions_new", saml2_conditions_new, METH_VARARGS, NULL},
    {"samlp2_artifact_resolve_new", samlp2_artifact_resolve_new, METH_VARARGS, NULL},
    {"samlp2_logout_request_new", samlp2_logout_request_new, METH_VARARGS, NULL},
    {"samlp2_attribute_query_new", samlp2_attribute_query_new, METH_VARARGS, NULL},
    {"saml2_evidence_new", saml2_evidence_new, METH_VARARGS, NULL},
    {"saml2_name_id_new", saml2_name_id_new, METH_VARARGS, NULL},
    {"saml2_name_id_new_with_string", saml2_name_id_new_with_string, METH_VARARGS, NULL},
    {"samlp2_name_id_mapping_response_new", samlp2_name_id_mapping_response_new, METH_VARARGS, NULL},
    {"samlp2_authz_decision_query_new", samlp2_authz_decision_query_new, METH_VARARGS, NULL},
    {"saml2_subject_confirmation_data_new", saml2_subject_confirmation_data_new, METH_VARARGS, NULL},
    {"ds_key_info_new", ds_key_info_new, METH_VARARGS, NULL},
    {"ds_rsa_key_value_new", ds_rsa_key_value_new, METH_VARARGS, NULL},
    {"ds_x509_data_new", ds_x509_data_new, METH_VARARGS, NULL},
    {"ds_key_value_new", ds_key_value_new, METH_VARARGS, NULL},
    {"node_cleanup_original_xmlnodes", node_cleanup_original_xmlnodes, METH_VARARGS, NULL},
    {"node_dump", node_dump, METH_VARARGS, NULL},
    {"node_export_to_base64", node_export_to_base64, METH_VARARGS, NULL},
    {"node_export_to_query", node_export_to_query, METH_VARARGS, NULL},
    {"node_export_to_query_with_password", node_export_to_query_with_password, METH_VARARGS, NULL},
    {"node_export_to_soap", node_export_to_soap, METH_VARARGS, NULL},
    {"node_export_to_xml", node_export_to_xml, METH_VARARGS, NULL},
    {"node_export_to_paos_request", node_export_to_paos_request, METH_VARARGS, NULL},
    {"node_export_to_ecp_soap_response", node_export_to_ecp_soap_response, METH_VARARGS, NULL},
    {"node_get_xmlNode", node_get_xmlNode, METH_VARARGS, NULL},
    {"node_get_original_xmlnode", node_get_original_xmlnode, METH_VARARGS, NULL},
    {"node_set_original_xmlnode", node_set_original_xmlnode, METH_VARARGS, NULL},
    {"node_set_custom_namespace", node_set_custom_namespace, METH_VARARGS, NULL},
    {"node_set_custom_nodename", node_set_custom_nodename, METH_VARARGS, NULL},
    {"node_get_name", node_get_name, METH_VARARGS, NULL},
    {"node_get_namespace", node_get_namespace, METH_VARARGS, NULL},
    {"node_init_from_message", node_init_from_message, METH_VARARGS, NULL},
    {"node_init_from_query", node_init_from_query, METH_VARARGS, NULL},
    {"node_init_from_xml", node_init_from_xml, METH_VARARGS, NULL},
    {"node_debug", node_debug, METH_VARARGS, NULL},
    {"key_query_verify", key_query_verify, METH_VARARGS, NULL},
    {"key_query_sign", key_query_sign, METH_VARARGS, NULL},
    {"key_saml2_xml_verify", key_saml2_xml_verify, METH_VARARGS, NULL},
    {"key_saml2_xml_sign", key_saml2_xml_sign, METH_VARARGS, NULL},
    {"Federation_remoteProviderId_get", Federation_remoteProviderId_get, METH_VARARGS, NULL},
    {"Federation_remoteProviderId_set", Federation_remoteProviderId_set, METH_VARARGS, NULL},
    {"Federation_localNameIdentifier_get", Federation_localNameIdentifier_get, METH_VARARGS, NULL},
    {"Federation_localNameIdentifier_set", Federation_localNameIdentifier_set, METH_VARARGS, NULL},
    {"Federation_remoteNameIdentifier_get", Federation_remoteNameIdentifier_get, METH_VARARGS, NULL},
    {"Federation_remoteNameIdentifier_set", Federation_remoteNameIdentifier_set, METH_VARARGS, NULL},
    {"federation_build_local_name_identifier", federation_build_local_name_identifier, METH_VARARGS, NULL},
    {"federation_verify_name_identifier", federation_verify_name_identifier, METH_VARARGS, NULL},
    {"Profile_server_get", Profile_server_get, METH_VARARGS, NULL},
    {"Profile_server_set", Profile_server_set, METH_VARARGS, NULL},
    {"Profile_request_get", Profile_request_get, METH_VARARGS, NULL},
    {"Profile_request_set", Profile_request_set, METH_VARARGS, NULL},
    {"Profile_response_get", Profile_response_get, METH_VARARGS, NULL},
    {"Profile_response_set", Profile_response_set, METH_VARARGS, NULL},
    {"Profile_nameIdentifier_get", Profile_nameIdentifier_get, METH_VARARGS, NULL},
    {"Profile_nameIdentifier_set", Profile_nameIdentifier_set, METH_VARARGS, NULL},
    {"Profile_remoteProviderId_get", Profile_remoteProviderId_get, METH_VARARGS, NULL},
    {"Profile_remoteProviderId_set", Profile_remoteProviderId_set, METH_VARARGS, NULL},
    {"Profile_msgUrl_get", Profile_msgUrl_get, METH_VARARGS, NULL},
    {"Profile_msgUrl_set", Profile_msgUrl_set, METH_VARARGS, NULL},
    {"Profile_msgBody_get", Profile_msgBody_get, METH_VARARGS, NULL},
    {"Profile_msgBody_set", Profile_msgBody_set, METH_VARARGS, NULL},
    {"Profile_msgRelayState_get", Profile_msgRelayState_get, METH_VARARGS, NULL},
    {"Profile_msgRelayState_set", Profile_msgRelayState_set, METH_VARARGS, NULL},
    {"profile_get_identity", profile_get_identity, METH_VARARGS, NULL},
    {"profile_get_session", profile_get_session, METH_VARARGS, NULL},
    {"profile_has_dirty_identity", profile_has_dirty_identity, METH_VARARGS, NULL},
    {"profile_has_dirty_session", profile_has_dirty_session, METH_VARARGS, NULL},
    {"profile_set_identity_from_dump", profile_set_identity_from_dump, METH_VARARGS, NULL},
    {"profile_set_session_from_dump", profile_set_session_from_dump, METH_VARARGS, NULL},
    {"profile_get_federation_nameIdentifier", profile_get_federation_nameIdentifier, METH_VARARGS, NULL},
    {"profile_get_artifact", profile_get_artifact, METH_VARARGS, NULL},
    {"profile_get_artifact_message", profile_get_artifact_message, METH_VARARGS, NULL},
    {"profile_set_artifact_message", profile_set_artifact_message, METH_VARARGS, NULL},
    {"profile_set_signature_hint", profile_set_signature_hint, METH_VARARGS, NULL},
    {"profile_get_signature_hint", profile_get_signature_hint, METH_VARARGS, NULL},
    {"profile_set_soap_fault_response", profile_set_soap_fault_response, METH_VARARGS, NULL},
    {"profile_set_signature_verify_hint", profile_set_signature_verify_hint, METH_VARARGS, NULL},
    {"profile_get_signature_verify_hint", profile_get_signature_verify_hint, METH_VARARGS, NULL},
    {"profile_sso_role_with", profile_sso_role_with, METH_VARARGS, NULL},
    {"profile_get_signature_status", profile_get_signature_status, METH_VARARGS, NULL},
    {"Session_assertions_get", Session_assertions_get, METH_VARARGS, NULL},
    {"Session_assertions_set", Session_assertions_set, METH_VARARGS, NULL},
    {"Session_isDirty_get", Session_isDirty_get, METH_VARARGS, NULL},
    {"Session_isDirty_set", Session_isDirty_set, METH_VARARGS, NULL},
    {"session_dump", session_dump, METH_VARARGS, NULL},
    {"session_get_assertions", session_get_assertions, METH_VARARGS, NULL},
    {"session_get_provider_index", session_get_provider_index, METH_VARARGS, NULL},
    {"session_is_empty", session_is_empty, METH_VARARGS, NULL},
    {"session_remove_assertion", session_remove_assertion, METH_VARARGS, NULL},
    {"session_get_assertion", session_get_assertion, METH_VARARGS, NULL},
    {"session_add_assertion", session_add_assertion, METH_VARARGS, NULL},
    {"session_get_name_ids", session_get_name_ids, METH_VARARGS, NULL},
    {"session_get_session_indexes", session_get_session_indexes, METH_VARARGS, NULL},
    {"session_get_assertion_ids", session_get_assertion_ids, METH_VARARGS, NULL},
    {"Provider_providerId_get", Provider_providerId_get, METH_VARARGS, NULL},
    {"Provider_providerId_set", Provider_providerId_set, METH_VARARGS, NULL},
    {"Provider_role_get", Provider_role_get, METH_VARARGS, NULL},
    {"Provider_role_set", Provider_role_set, METH_VARARGS, NULL},
    {"Provider_metadataFilename_get", Provider_metadataFilename_get, METH_VARARGS, NULL},
    {"Provider_metadataFilename_set", Provider_metadataFilename_set, METH_VARARGS, NULL},
    {"Provider_publicKey_get", Provider_publicKey_get, METH_VARARGS, NULL},
    {"Provider_publicKey_set", Provider_publicKey_set, METH_VARARGS, NULL},
    {"Provider_caCertChain_get", Provider_caCertChain_get, METH_VARARGS, NULL},
    {"Provider_caCertChain_set", Provider_caCertChain_set, METH_VARARGS, NULL},
    {"provider_get_assertion_consumer_service_url", provider_get_assertion_consumer_service_url, METH_VARARGS, NULL},
    {"provider_get_metadata_one", provider_get_metadata_one, METH_VARARGS, NULL},
    {"provider_get_metadata_list", provider_get_metadata_list, METH_VARARGS, NULL},
    {"provider_get_first_http_method", provider_get_first_http_method, METH_VARARGS, NULL},
    {"provider_accept_http_method", provider_accept_http_method, METH_VARARGS, NULL},
    {"provider_has_protocol_profile", provider_has_protocol_profile, METH_VARARGS, NULL},
    {"provider_get_base64_succinct_id", provider_get_base64_succinct_id, METH_VARARGS, NULL},
    {"provider_get_organization", provider_get_organization, METH_VARARGS, NULL},
    {"provider_get_protocol_conformance", provider_get_protocol_conformance, METH_VARARGS, NULL},
    {"provider_set_encryption_mode", provider_set_encryption_mode, METH_VARARGS, NULL},
    {"provider_get_encryption_mode", provider_get_encryption_mode, METH_VARARGS, NULL},
    {"provider_set_encryption_sym_key_type", provider_set_encryption_sym_key_type, METH_VARARGS, NULL},
    {"provider_get_default_name_id_format", provider_get_default_name_id_format, METH_VARARGS, NULL},
    {"provider_get_sp_name_qualifier", provider_get_sp_name_qualifier, METH_VARARGS, NULL},
    {"provider_verify_single_node_signature", provider_verify_single_node_signature, METH_VARARGS, NULL},
    {"provider_get_idp_supported_attributes", provider_get_idp_supported_attributes, METH_VARARGS, NULL},
    {"provider_get_valid_until", provider_get_valid_until, METH_VARARGS, NULL},
    {"provider_get_cache_duration", provider_get_cache_duration, METH_VARARGS, NULL},
    {"provider_get_metadata_one_for_role", provider_get_metadata_one_for_role, METH_VARARGS, NULL},
    {"provider_get_metadata_list_for_role", provider_get_metadata_list_for_role, METH_VARARGS, NULL},
    {"provider_get_metadata_keys_for_role", provider_get_metadata_keys_for_role, METH_VARARGS, NULL},
    {"provider_get_roles", provider_get_roles, METH_VARARGS, NULL},
    {"provider_match_conformance", provider_match_conformance, METH_VARARGS, NULL},
    {"provider_set_server_signing_key", provider_set_server_signing_key, METH_VARARGS, NULL},
    {"provider_add_key", provider_add_key, METH_VARARGS, NULL},
    {"provider_verify_signature", provider_verify_signature, METH_VARARGS, NULL},
    {"provider_saml2_node_encrypt", provider_saml2_node_encrypt, METH_VARARGS, NULL},
    {"Server_providers_get", Server_providers_get, METH_VARARGS, NULL},
    {"Server_providers_set", Server_providers_set, METH_VARARGS, NULL},
    {"Server_privateKey_get", Server_privateKey_get, METH_VARARGS, NULL},
    {"Server_privateKey_set", Server_privateKey_set, METH_VARARGS, NULL},
    {"Server_privateKeyPassword_get", Server_privateKeyPassword_get, METH_VARARGS, NULL},
    {"Server_privateKeyPassword_set", Server_privateKeyPassword_set, METH_VARARGS, NULL},
    {"Server_certificate_get", Server_certificate_get, METH_VARARGS, NULL},
    {"Server_certificate_set", Server_certificate_set, METH_VARARGS, NULL},
    {"Server_signatureMethod_get", Server_signatureMethod_get, METH_VARARGS, NULL},
    {"Server_signatureMethod_set", Server_signatureMethod_set, METH_VARARGS, NULL},
    {"server_add_provider", server_add_provider, METH_VARARGS, NULL},
    {"server_add_provider_from_buffer", server_add_provider_from_buffer, METH_VARARGS, NULL},
    {"server_dump", server_dump, METH_VARARGS, NULL},
    {"server_get_provider", server_get_provider, METH_VARARGS, NULL},
    {"server_set_encryption_private_key", server_set_encryption_private_key, METH_VARARGS, NULL},
    {"server_load_affiliation", server_load_affiliation, METH_VARARGS, NULL},
    {"server_set_encryption_private_key_with_password", server_set_encryption_private_key_with_password, METH_VARARGS, NULL},
    {"server_load_metadata", server_load_metadata, METH_VARARGS, NULL},
    {"server_add_provider2", server_add_provider2, METH_VARARGS, NULL},
    {"server_saml2_assertion_setup_signature", server_saml2_assertion_setup_signature, METH_VARARGS, NULL},
    {"NameRegistration_oldNameIdentifier_get", NameRegistration_oldNameIdentifier_get, METH_VARARGS, NULL},
    {"NameRegistration_oldNameIdentifier_set", NameRegistration_oldNameIdentifier_set, METH_VARARGS, NULL},
    {"name_registration_build_request_msg", name_registration_build_request_msg, METH_VARARGS, NULL},
    {"name_registration_build_response_msg", name_registration_build_response_msg, METH_VARARGS, NULL},
    {"name_registration_dump", name_registration_dump, METH_VARARGS, NULL},
    {"name_registration_init_request", name_registration_init_request, METH_VARARGS, NULL},
    {"name_registration_process_request_msg", name_registration_process_request_msg, METH_VARARGS, NULL},
    {"name_registration_process_response_msg", name_registration_process_response_msg, METH_VARARGS, NULL},
    {"name_registration_validate_request", name_registration_validate_request, METH_VARARGS, NULL},
    {"Login_protocolProfile_get", Login_protocolProfile_get, METH_VARARGS, NULL},
    {"Login_protocolProfile_set", Login_protocolProfile_set, METH_VARARGS, NULL},
    {"Login_assertionArtifact_get", Login_assertionArtifact_get, METH_VARARGS, NULL},
    {"Login_assertionArtifact_set", Login_assertionArtifact_set, METH_VARARGS, NULL},
    {"login_accept_sso", login_accept_sso, METH_VARARGS, NULL},
    {"login_build_artifact_msg", login_build_artifact_msg, METH_VARARGS, NULL},
    {"login_build_assertion", login_build_assertion, METH_VARARGS, NULL},
    {"login_build_authn_request_msg", login_build_authn_request_msg, METH_VARARGS, NULL},
    {"login_build_authn_response_msg", login_build_authn_response_msg, METH_VARARGS, NULL},
    {"login_build_request_msg", login_build_request_msg, METH_VARARGS, NULL},
    {"login_build_response_msg", login_build_response_msg, METH_VARARGS, NULL},
    {"login_dump", login_dump, METH_VARARGS, NULL},
    {"login_init_authn_request", login_init_authn_request, METH_VARARGS, NULL},
    {"login_init_request", login_init_request, METH_VARARGS, NULL},
    {"login_init_idp_initiated_authn_request", login_init_idp_initiated_authn_request, METH_VARARGS, NULL},
    {"login_must_ask_for_consent", login_must_ask_for_consent, METH_VARARGS, NULL},
    {"login_must_authenticate", login_must_authenticate, METH_VARARGS, NULL},
    {"login_process_authn_request_msg", login_process_authn_request_msg, METH_VARARGS, NULL},
    {"login_process_authn_response_msg", login_process_authn_response_msg, METH_VARARGS, NULL},
    {"login_process_request_msg", login_process_request_msg, METH_VARARGS, NULL},
    {"login_process_response_msg", login_process_response_msg, METH_VARARGS, NULL},
    {"login_validate_request_msg", login_validate_request_msg, METH_VARARGS, NULL},
    {"login_process_paos_response_msg", login_process_paos_response_msg, METH_VARARGS, NULL},
    {"login_get_assertion", login_get_assertion, METH_VARARGS, NULL},
    {"Lecp_authnRequestEnvelope_get", Lecp_authnRequestEnvelope_get, METH_VARARGS, NULL},
    {"Lecp_authnRequestEnvelope_set", Lecp_authnRequestEnvelope_set, METH_VARARGS, NULL},
    {"Lecp_authnResponseEnvelope_get", Lecp_authnResponseEnvelope_get, METH_VARARGS, NULL},
    {"Lecp_authnResponseEnvelope_set", Lecp_authnResponseEnvelope_set, METH_VARARGS, NULL},
    {"Lecp_assertionConsumerServiceURL_get", Lecp_assertionConsumerServiceURL_get, METH_VARARGS, NULL},
    {"Lecp_assertionConsumerServiceURL_set", Lecp_assertionConsumerServiceURL_set, METH_VARARGS, NULL},
    {"lecp_build_authn_request_envelope_msg", lecp_build_authn_request_envelope_msg, METH_VARARGS, NULL},
    {"lecp_build_authn_request_msg", lecp_build_authn_request_msg, METH_VARARGS, NULL},
    {"lecp_build_authn_response_msg", lecp_build_authn_response_msg, METH_VARARGS, NULL},
    {"lecp_build_authn_response_envelope_msg", lecp_build_authn_response_envelope_msg, METH_VARARGS, NULL},
    {"lecp_init_authn_request", lecp_init_authn_request, METH_VARARGS, NULL},
    {"lecp_process_authn_request_msg", lecp_process_authn_request_msg, METH_VARARGS, NULL},
    {"lecp_process_authn_request_envelope_msg", lecp_process_authn_request_envelope_msg, METH_VARARGS, NULL},
    {"lecp_process_authn_response_envelope_msg", lecp_process_authn_response_envelope_msg, METH_VARARGS, NULL},
    {"Identity_federations_get", Identity_federations_get, METH_VARARGS, NULL},
    {"Identity_federations_set", Identity_federations_set, METH_VARARGS, NULL},
    {"Identity_isDirty_get", Identity_isDirty_get, METH_VARARGS, NULL},
    {"Identity_isDirty_set", Identity_isDirty_set, METH_VARARGS, NULL},
    {"identity_get_federation", identity_get_federation, METH_VARARGS, NULL},
    {"identity_dump", identity_dump, METH_VARARGS, NULL},
    {"logout_build_request_msg", logout_build_request_msg, METH_VARARGS, NULL},
    {"logout_build_response_msg", logout_build_response_msg, METH_VARARGS, NULL},
    {"logout_dump", logout_dump, METH_VARARGS, NULL},
    {"logout_get_next_providerID", logout_get_next_providerID, METH_VARARGS, NULL},
    {"logout_init_request", logout_init_request, METH_VARARGS, NULL},
    {"logout_process_request_msg", logout_process_request_msg, METH_VARARGS, NULL},
    {"logout_process_response_msg", logout_process_response_msg, METH_VARARGS, NULL},
    {"logout_reset_providerID_index", logout_reset_providerID_index, METH_VARARGS, NULL},
    {"logout_validate_request", logout_validate_request, METH_VARARGS, NULL},
    {"NameIdentifierMapping_targetNameIdentifier_get", NameIdentifierMapping_targetNameIdentifier_get, METH_VARARGS, NULL},
    {"NameIdentifierMapping_targetNameIdentifier_set", NameIdentifierMapping_targetNameIdentifier_set, METH_VARARGS, NULL},
    {"name_identifier_mapping_build_request_msg", name_identifier_mapping_build_request_msg, METH_VARARGS, NULL},
    {"name_identifier_mapping_build_response_msg", name_identifier_mapping_build_response_msg, METH_VARARGS, NULL},
    {"name_identifier_mapping_init_request", name_identifier_mapping_init_request, METH_VARARGS, NULL},
    {"name_identifier_mapping_process_request_msg", name_identifier_mapping_process_request_msg, METH_VARARGS, NULL},
    {"name_identifier_mapping_process_response_msg", name_identifier_mapping_process_response_msg, METH_VARARGS, NULL},
    {"name_identifier_mapping_validate_request", name_identifier_mapping_validate_request, METH_VARARGS, NULL},
    {"defederation_build_notification_msg", defederation_build_notification_msg, METH_VARARGS, NULL},
    {"defederation_init_notification", defederation_init_notification, METH_VARARGS, NULL},
    {"defederation_process_notification_msg", defederation_process_notification_msg, METH_VARARGS, NULL},
    {"defederation_validate_notification", defederation_validate_notification, METH_VARARGS, NULL},
    {"Ecp_assertionConsumerURL_get", Ecp_assertionConsumerURL_get, METH_VARARGS, NULL},
    {"Ecp_assertionConsumerURL_set", Ecp_assertionConsumerURL_set, METH_VARARGS, NULL},
    {"ecp_process_authn_request_msg", ecp_process_authn_request_msg, METH_VARARGS, NULL},
    {"ecp_process_response_msg", ecp_process_response_msg, METH_VARARGS, NULL},
    {"assertion_query_init_request", assertion_query_init_request, METH_VARARGS, NULL},
    {"assertion_query_validate_request", assertion_query_validate_request, METH_VARARGS, NULL},
    {"assertion_query_build_request_msg", assertion_query_build_request_msg, METH_VARARGS, NULL},
    {"assertion_query_process_request_msg", assertion_query_process_request_msg, METH_VARARGS, NULL},
    {"assertion_query_build_response_msg", assertion_query_build_response_msg, METH_VARARGS, NULL},
    {"assertion_query_process_response_msg", assertion_query_process_response_msg, METH_VARARGS, NULL},
    {"assertion_query_add_attribute_request", assertion_query_add_attribute_request, METH_VARARGS, NULL},
    {"assertion_query_get_request_type", assertion_query_get_request_type, METH_VARARGS, NULL},
    {"name_id_management_dump", name_id_management_dump, METH_VARARGS, NULL},
    {"name_id_management_init_request", name_id_management_init_request, METH_VARARGS, NULL},
    {"name_id_management_build_request_msg", name_id_management_build_request_msg, METH_VARARGS, NULL},
    {"name_id_management_process_request_msg", name_id_management_process_request_msg, METH_VARARGS, NULL},
    {"name_id_management_validate_request", name_id_management_validate_request, METH_VARARGS, NULL},
    {"name_id_management_build_response_msg", name_id_management_build_response_msg, METH_VARARGS, NULL},
    {"name_id_management_process_response_msg", name_id_management_process_response_msg, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_extension_get", LibAuthnRequestEnvelope_extension_get, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_extension_set", LibAuthnRequestEnvelope_extension_set, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_authnRequest_get", LibAuthnRequestEnvelope_authnRequest_get, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_authnRequest_set", LibAuthnRequestEnvelope_authnRequest_set, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_providerId_get", LibAuthnRequestEnvelope_providerId_get, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_providerId_set", LibAuthnRequestEnvelope_providerId_set, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_providerName_get", LibAuthnRequestEnvelope_providerName_get, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_providerName_set", LibAuthnRequestEnvelope_providerName_set, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_assertionConsumerServiceUrl_get", LibAuthnRequestEnvelope_assertionConsumerServiceUrl_get, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_assertionConsumerServiceUrl_set", LibAuthnRequestEnvelope_assertionConsumerServiceUrl_set, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_idpList_get", LibAuthnRequestEnvelope_idpList_get, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_idpList_set", LibAuthnRequestEnvelope_idpList_set, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_isPassive_get", LibAuthnRequestEnvelope_isPassive_get, METH_VARARGS, NULL},
    {"LibAuthnRequestEnvelope_isPassive_set", LibAuthnRequestEnvelope_isPassive_set, METH_VARARGS, NULL},
    {"LibAuthnResponseEnvelope_extension_get", LibAuthnResponseEnvelope_extension_get, METH_VARARGS, NULL},
    {"LibAuthnResponseEnvelope_extension_set", LibAuthnResponseEnvelope_extension_set, METH_VARARGS, NULL},
    {"LibAuthnResponseEnvelope_authnResponse_get", LibAuthnResponseEnvelope_authnResponse_get, METH_VARARGS, NULL},
    {"LibAuthnResponseEnvelope_authnResponse_set", LibAuthnResponseEnvelope_authnResponse_set, METH_VARARGS, NULL},
    {"LibAuthnResponseEnvelope_assertionConsumerServiceUrl_get", LibAuthnResponseEnvelope_assertionConsumerServiceUrl_get, METH_VARARGS, NULL},
    {"LibAuthnResponseEnvelope_assertionConsumerServiceUrl_set", LibAuthnResponseEnvelope_assertionConsumerServiceUrl_set, METH_VARARGS, NULL},
    {"SamlAttributeValue_any_get", SamlAttributeValue_any_get, METH_VARARGS, NULL},
    {"SamlAttributeValue_any_set", SamlAttributeValue_any_set, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_responseId_get", SamlpResponseAbstract_responseId_get, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_responseId_set", SamlpResponseAbstract_responseId_set, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_inResponseTo_get", SamlpResponseAbstract_inResponseTo_get, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_inResponseTo_set", SamlpResponseAbstract_inResponseTo_set, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_majorVersion_get", SamlpResponseAbstract_majorVersion_get, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_majorVersion_set", SamlpResponseAbstract_majorVersion_set, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_minorVersion_get", SamlpResponseAbstract_minorVersion_get, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_minorVersion_set", SamlpResponseAbstract_minorVersion_set, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_issueInstant_get", SamlpResponseAbstract_issueInstant_get, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_issueInstant_set", SamlpResponseAbstract_issueInstant_set, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_recipient_get", SamlpResponseAbstract_recipient_get, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_recipient_set", SamlpResponseAbstract_recipient_set, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_signType_get", SamlpResponseAbstract_signType_get, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_signType_set", SamlpResponseAbstract_signType_set, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_signMethod_get", SamlpResponseAbstract_signMethod_get, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_signMethod_set", SamlpResponseAbstract_signMethod_set, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_privateKeyFile_get", SamlpResponseAbstract_privateKeyFile_get, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_privateKeyFile_set", SamlpResponseAbstract_privateKeyFile_set, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_certificateFile_get", SamlpResponseAbstract_certificateFile_get, METH_VARARGS, NULL},
    {"SamlpResponseAbstract_certificateFile_set", SamlpResponseAbstract_certificateFile_set, METH_VARARGS, NULL},
    {"samlp_response_abstract_fill", samlp_response_abstract_fill, METH_VARARGS, NULL},
    {"LibAuthnContext_authnContextClassRef_get", LibAuthnContext_authnContextClassRef_get, METH_VARARGS, NULL},
    {"LibAuthnContext_authnContextClassRef_set", LibAuthnContext_authnContextClassRef_set, METH_VARARGS, NULL},
    {"LibAuthnContext_authnContextStatementRef_get", LibAuthnContext_authnContextStatementRef_get, METH_VARARGS, NULL},
    {"LibAuthnContext_authnContextStatementRef_set", LibAuthnContext_authnContextStatementRef_set, METH_VARARGS, NULL},
    {"SamlAttributeDesignator_attributeName_get", SamlAttributeDesignator_attributeName_get, METH_VARARGS, NULL},
    {"SamlAttributeDesignator_attributeName_set", SamlAttributeDesignator_attributeName_set, METH_VARARGS, NULL},
    {"SamlAttributeDesignator_attributeNamespace_get", SamlAttributeDesignator_attributeNamespace_get, METH_VARARGS, NULL},
    {"SamlAttributeDesignator_attributeNamespace_set", SamlAttributeDesignator_attributeNamespace_set, METH_VARARGS, NULL},
    {"LibRequestAuthnContext_authnContextClassRef_get", LibRequestAuthnContext_authnContextClassRef_get, METH_VARARGS, NULL},
    {"LibRequestAuthnContext_authnContextClassRef_set", LibRequestAuthnContext_authnContextClassRef_set, METH_VARARGS, NULL},
    {"LibRequestAuthnContext_authnContextStatementRef_get", LibRequestAuthnContext_authnContextStatementRef_get, METH_VARARGS, NULL},
    {"LibRequestAuthnContext_authnContextStatementRef_set", LibRequestAuthnContext_authnContextStatementRef_set, METH_VARARGS, NULL},
    {"LibRequestAuthnContext_authnContextComparison_get", LibRequestAuthnContext_authnContextComparison_get, METH_VARARGS, NULL},
    {"LibRequestAuthnContext_authnContextComparison_set", LibRequestAuthnContext_authnContextComparison_set, METH_VARARGS, NULL},
    {"SamlSubject_nameIdentifier_get", SamlSubject_nameIdentifier_get, METH_VARARGS, NULL},
    {"SamlSubject_nameIdentifier_set", SamlSubject_nameIdentifier_set, METH_VARARGS, NULL},
    {"SamlSubject_subjectConfirmation_get", SamlSubject_subjectConfirmation_get, METH_VARARGS, NULL},
    {"SamlSubject_subjectConfirmation_set", SamlSubject_subjectConfirmation_set, METH_VARARGS, NULL},
    {"SamlSubject_encryptedNameIdentifier_get", SamlSubject_encryptedNameIdentifier_get, METH_VARARGS, NULL},
    {"SamlSubject_encryptedNameIdentifier_set", SamlSubject_encryptedNameIdentifier_set, METH_VARARGS, NULL},
    {"LibSubject_idpProvidedNameIdentifier_get", LibSubject_idpProvidedNameIdentifier_get, METH_VARARGS, NULL},
    {"LibSubject_idpProvidedNameIdentifier_set", LibSubject_idpProvidedNameIdentifier_set, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingResponse_extension_get", LibNameIdentifierMappingResponse_extension_get, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingResponse_extension_set", LibNameIdentifierMappingResponse_extension_set, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingResponse_providerId_get", LibNameIdentifierMappingResponse_providerId_get, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingResponse_providerId_set", LibNameIdentifierMappingResponse_providerId_set, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingResponse_status_get", LibNameIdentifierMappingResponse_status_get, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingResponse_status_set", LibNameIdentifierMappingResponse_status_set, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingResponse_nameIdentifier_get", LibNameIdentifierMappingResponse_nameIdentifier_get, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingResponse_nameIdentifier_set", LibNameIdentifierMappingResponse_nameIdentifier_set, METH_VARARGS, NULL},
    {"SamlAttribute_attributeName_get", SamlAttribute_attributeName_get, METH_VARARGS, NULL},
    {"SamlAttribute_attributeName_set", SamlAttribute_attributeName_set, METH_VARARGS, NULL},
    {"SamlAttribute_attributeNameSpace_get", SamlAttribute_attributeNameSpace_get, METH_VARARGS, NULL},
    {"SamlAttribute_attributeNameSpace_set", SamlAttribute_attributeNameSpace_set, METH_VARARGS, NULL},
    {"SamlAttribute_attributeValue_get", SamlAttribute_attributeValue_get, METH_VARARGS, NULL},
    {"SamlAttribute_attributeValue_set", SamlAttribute_attributeValue_set, METH_VARARGS, NULL},
    {"SamlAssertion_conditions_get", SamlAssertion_conditions_get, METH_VARARGS, NULL},
    {"SamlAssertion_conditions_set", SamlAssertion_conditions_set, METH_VARARGS, NULL},
    {"SamlAssertion_advice_get", SamlAssertion_advice_get, METH_VARARGS, NULL},
    {"SamlAssertion_advice_set", SamlAssertion_advice_set, METH_VARARGS, NULL},
    {"SamlAssertion_subjectStatement_get", SamlAssertion_subjectStatement_get, METH_VARARGS, NULL},
    {"SamlAssertion_subjectStatement_set", SamlAssertion_subjectStatement_set, METH_VARARGS, NULL},
    {"SamlAssertion_authenticationStatement_get", SamlAssertion_authenticationStatement_get, METH_VARARGS, NULL},
    {"SamlAssertion_authenticationStatement_set", SamlAssertion_authenticationStatement_set, METH_VARARGS, NULL},
    {"SamlAssertion_attributeStatement_get", SamlAssertion_attributeStatement_get, METH_VARARGS, NULL},
    {"SamlAssertion_attributeStatement_set", SamlAssertion_attributeStatement_set, METH_VARARGS, NULL},
    {"SamlAssertion_majorVersion_get", SamlAssertion_majorVersion_get, METH_VARARGS, NULL},
    {"SamlAssertion_majorVersion_set", SamlAssertion_majorVersion_set, METH_VARARGS, NULL},
    {"SamlAssertion_minorVersion_get", SamlAssertion_minorVersion_get, METH_VARARGS, NULL},
    {"SamlAssertion_minorVersion_set", SamlAssertion_minorVersion_set, METH_VARARGS, NULL},
    {"SamlAssertion_assertionId_get", SamlAssertion_assertionId_get, METH_VARARGS, NULL},
    {"SamlAssertion_assertionId_set", SamlAssertion_assertionId_set, METH_VARARGS, NULL},
    {"SamlAssertion_issuer_get", SamlAssertion_issuer_get, METH_VARARGS, NULL},
    {"SamlAssertion_issuer_set", SamlAssertion_issuer_set, METH_VARARGS, NULL},
    {"SamlAssertion_issueInstant_get", SamlAssertion_issueInstant_get, METH_VARARGS, NULL},
    {"SamlAssertion_issueInstant_set", SamlAssertion_issueInstant_set, METH_VARARGS, NULL},
    {"SamlAssertion_signType_get", SamlAssertion_signType_get, METH_VARARGS, NULL},
    {"SamlAssertion_signType_set", SamlAssertion_signType_set, METH_VARARGS, NULL},
    {"SamlAssertion_signMethod_get", SamlAssertion_signMethod_get, METH_VARARGS, NULL},
    {"SamlAssertion_signMethod_set", SamlAssertion_signMethod_set, METH_VARARGS, NULL},
    {"SamlAssertion_privateKeyFile_get", SamlAssertion_privateKeyFile_get, METH_VARARGS, NULL},
    {"SamlAssertion_privateKeyFile_set", SamlAssertion_privateKeyFile_set, METH_VARARGS, NULL},
    {"SamlAssertion_certificateFile_get", SamlAssertion_certificateFile_get, METH_VARARGS, NULL},
    {"SamlAssertion_certificateFile_set", SamlAssertion_certificateFile_set, METH_VARARGS, NULL},
    {"LibIDPList_idpEntries_get", LibIDPList_idpEntries_get, METH_VARARGS, NULL},
    {"LibIDPList_idpEntries_set", LibIDPList_idpEntries_set, METH_VARARGS, NULL},
    {"LibIDPList_getComplete_get", LibIDPList_getComplete_get, METH_VARARGS, NULL},
    {"LibIDPList_getComplete_set", LibIDPList_getComplete_set, METH_VARARGS, NULL},
    {"LibIDPEntry_providerId_get", LibIDPEntry_providerId_get, METH_VARARGS, NULL},
    {"LibIDPEntry_providerId_set", LibIDPEntry_providerId_set, METH_VARARGS, NULL},
    {"LibIDPEntry_providerName_get", LibIDPEntry_providerName_get, METH_VARARGS, NULL},
    {"LibIDPEntry_providerName_set", LibIDPEntry_providerName_set, METH_VARARGS, NULL},
    {"LibIDPEntry_loc_get", LibIDPEntry_loc_get, METH_VARARGS, NULL},
    {"LibIDPEntry_loc_set", LibIDPEntry_loc_set, METH_VARARGS, NULL},
    {"SamlpStatusCode_statusCode_get", SamlpStatusCode_statusCode_get, METH_VARARGS, NULL},
    {"SamlpStatusCode_statusCode_set", SamlpStatusCode_statusCode_set, METH_VARARGS, NULL},
    {"SamlpStatusCode_value_get", SamlpStatusCode_value_get, METH_VARARGS, NULL},
    {"SamlpStatusCode_value_set", SamlpStatusCode_value_set, METH_VARARGS, NULL},
    {"SamlAdvice_assertionIdReference_get", SamlAdvice_assertionIdReference_get, METH_VARARGS, NULL},
    {"SamlAdvice_assertionIdReference_set", SamlAdvice_assertionIdReference_set, METH_VARARGS, NULL},
    {"SamlAdvice_assertion_get", SamlAdvice_assertion_get, METH_VARARGS, NULL},
    {"SamlAdvice_assertion_set", SamlAdvice_assertion_set, METH_VARARGS, NULL},
    {"SamlpResponse_status_get", SamlpResponse_status_get, METH_VARARGS, NULL},
    {"SamlpResponse_status_set", SamlpResponse_status_set, METH_VARARGS, NULL},
    {"SamlpResponse_assertion_get", SamlpResponse_assertion_get, METH_VARARGS, NULL},
    {"SamlpResponse_assertion_set", SamlpResponse_assertion_set, METH_VARARGS, NULL},
    {"LibAuthnResponse_extension_get", LibAuthnResponse_extension_get, METH_VARARGS, NULL},
    {"LibAuthnResponse_extension_set", LibAuthnResponse_extension_set, METH_VARARGS, NULL},
    {"LibAuthnResponse_providerId_get", LibAuthnResponse_providerId_get, METH_VARARGS, NULL},
    {"LibAuthnResponse_providerId_set", LibAuthnResponse_providerId_set, METH_VARARGS, NULL},
    {"LibAuthnResponse_relayState_get", LibAuthnResponse_relayState_get, METH_VARARGS, NULL},
    {"LibAuthnResponse_relayState_set", LibAuthnResponse_relayState_set, METH_VARARGS, NULL},
    {"LibAuthnResponse_consent_get", LibAuthnResponse_consent_get, METH_VARARGS, NULL},
    {"LibAuthnResponse_consent_set", LibAuthnResponse_consent_set, METH_VARARGS, NULL},
    {"SamlConditions_condition_get", SamlConditions_condition_get, METH_VARARGS, NULL},
    {"SamlConditions_condition_set", SamlConditions_condition_set, METH_VARARGS, NULL},
    {"SamlConditions_audienceRestrictionCondition_get", SamlConditions_audienceRestrictionCondition_get, METH_VARARGS, NULL},
    {"SamlConditions_audienceRestrictionCondition_set", SamlConditions_audienceRestrictionCondition_set, METH_VARARGS, NULL},
    {"SamlConditions_notBefore_get", SamlConditions_notBefore_get, METH_VARARGS, NULL},
    {"SamlConditions_notBefore_set", SamlConditions_notBefore_set, METH_VARARGS, NULL},
    {"SamlConditions_notOnOrAfter_get", SamlConditions_notOnOrAfter_get, METH_VARARGS, NULL},
    {"SamlConditions_notOnOrAfter_set", SamlConditions_notOnOrAfter_set, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_respondWith_get", SamlpRequestAbstract_respondWith_get, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_respondWith_set", SamlpRequestAbstract_respondWith_set, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_requestId_get", SamlpRequestAbstract_requestId_get, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_requestId_set", SamlpRequestAbstract_requestId_set, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_majorVersion_get", SamlpRequestAbstract_majorVersion_get, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_majorVersion_set", SamlpRequestAbstract_majorVersion_set, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_minorVersion_get", SamlpRequestAbstract_minorVersion_get, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_minorVersion_set", SamlpRequestAbstract_minorVersion_set, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_issueInstant_get", SamlpRequestAbstract_issueInstant_get, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_issueInstant_set", SamlpRequestAbstract_issueInstant_set, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_signType_get", SamlpRequestAbstract_signType_get, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_signType_set", SamlpRequestAbstract_signType_set, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_signMethod_get", SamlpRequestAbstract_signMethod_get, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_signMethod_set", SamlpRequestAbstract_signMethod_set, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_privateKeyFile_get", SamlpRequestAbstract_privateKeyFile_get, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_privateKeyFile_set", SamlpRequestAbstract_privateKeyFile_set, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_certificateFile_get", SamlpRequestAbstract_certificateFile_get, METH_VARARGS, NULL},
    {"SamlpRequestAbstract_certificateFile_set", SamlpRequestAbstract_certificateFile_set, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingRequest_extension_get", LibNameIdentifierMappingRequest_extension_get, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingRequest_extension_set", LibNameIdentifierMappingRequest_extension_set, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingRequest_providerId_get", LibNameIdentifierMappingRequest_providerId_get, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingRequest_providerId_set", LibNameIdentifierMappingRequest_providerId_set, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingRequest_nameIdentifier_get", LibNameIdentifierMappingRequest_nameIdentifier_get, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingRequest_nameIdentifier_set", LibNameIdentifierMappingRequest_nameIdentifier_set, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingRequest_targetNamespace_get", LibNameIdentifierMappingRequest_targetNamespace_get, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingRequest_targetNamespace_set", LibNameIdentifierMappingRequest_targetNamespace_set, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingRequest_consent_get", LibNameIdentifierMappingRequest_consent_get, METH_VARARGS, NULL},
    {"LibNameIdentifierMappingRequest_consent_set", LibNameIdentifierMappingRequest_consent_set, METH_VARARGS, NULL},
    {"LibFederationTerminationNotification_extension_get", LibFederationTerminationNotification_extension_get, METH_VARARGS, NULL},
    {"LibFederationTerminationNotification_extension_set", LibFederationTerminationNotification_extension_set, METH_VARARGS, NULL},
    {"LibFederationTerminationNotification_providerId_get", LibFederationTerminationNotification_providerId_get, METH_VARARGS, NULL},
    {"LibFederationTerminationNotification_providerId_set", LibFederationTerminationNotification_providerId_set, METH_VARARGS, NULL},
    {"LibFederationTerminationNotification_nameIdentifier_get", LibFederationTerminationNotification_nameIdentifier_get, METH_VARARGS, NULL},
    {"LibFederationTerminationNotification_nameIdentifier_set", LibFederationTerminationNotification_nameIdentifier_set, METH_VARARGS, NULL},
    {"LibFederationTerminationNotification_consent_get", LibFederationTerminationNotification_consent_get, METH_VARARGS, NULL},
    {"LibFederationTerminationNotification_consent_set", LibFederationTerminationNotification_consent_set, METH_VARARGS, NULL},
    {"LibFederationTerminationNotification_relayState_get", LibFederationTerminationNotification_relayState_get, METH_VARARGS, NULL},
    {"LibFederationTerminationNotification_relayState_set", LibFederationTerminationNotification_relayState_set, METH_VARARGS, NULL},
    {"LibAuthnRequest_extension_get", LibAuthnRequest_extension_get, METH_VARARGS, NULL},
    {"LibAuthnRequest_extension_set", LibAuthnRequest_extension_set, METH_VARARGS, NULL},
    {"LibAuthnRequest_providerId_get", LibAuthnRequest_providerId_get, METH_VARARGS, NULL},
    {"LibAuthnRequest_providerId_set", LibAuthnRequest_providerId_set, METH_VARARGS, NULL},
    {"LibAuthnRequest_affiliationId_get", LibAuthnRequest_affiliationId_get, METH_VARARGS, NULL},
    {"LibAuthnRequest_affiliationId_set", LibAuthnRequest_affiliationId_set, METH_VARARGS, NULL},
    {"LibAuthnRequest_nameIdPolicy_get", LibAuthnRequest_nameIdPolicy_get, METH_VARARGS, NULL},
    {"LibAuthnRequest_nameIdPolicy_set", LibAuthnRequest_nameIdPolicy_set, METH_VARARGS, NULL},
    {"LibAuthnRequest_forceAuthn_get", LibAuthnRequest_forceAuthn_get, METH_VARARGS, NULL},
    {"LibAuthnRequest_forceAuthn_set", LibAuthnRequest_forceAuthn_set, METH_VARARGS, NULL},
    {"LibAuthnRequest_isPassive_get", LibAuthnRequest_isPassive_get, METH_VARARGS, NULL},
    {"LibAuthnRequest_isPassive_set", LibAuthnRequest_isPassive_set, METH_VARARGS, NULL},
    {"LibAuthnRequest_protocolProfile_get", LibAuthnRequest_protocolProfile_get, METH_VARARGS, NULL},
    {"LibAuthnRequest_protocolProfile_set", LibAuthnRequest_protocolProfile_set, METH_VARARGS, NULL},
    {"LibAuthnRequest_assertionConsumerServiceId_get", LibAuthnRequest_assertionConsumerServiceId_get, METH_VARARGS, NULL},
    {"LibAuthnRequest_assertionConsumerServiceId_set", LibAuthnRequest_assertionConsumerServiceId_set, METH_VARARGS, NULL},
    {"LibAuthnRequest_requestAuthnContext_get", LibAuthnRequest_requestAuthnContext_get, METH_VARARGS, NULL},
    {"LibAuthnRequest_requestAuthnContext_set", LibAuthnRequest_requestAuthnContext_set, METH_VARARGS, NULL},
    {"LibAuthnRequest_relayState_get", LibAuthnRequest_relayState_get, METH_VARARGS, NULL},
    {"LibAuthnRequest_relayState_set", LibAuthnRequest_relayState_set, METH_VARARGS, NULL},
    {"LibAuthnRequest_scoping_get", LibAuthnRequest_scoping_get, METH_VARARGS, NULL},
    {"LibAuthnRequest_scoping_set", LibAuthnRequest_scoping_set, METH_VARARGS, NULL},
    {"LibAuthnRequest_consent_get", LibAuthnRequest_consent_get, METH_VARARGS, NULL},
    {"LibAuthnRequest_consent_set", LibAuthnRequest_consent_set, METH_VARARGS, NULL},
    {"LibLogoutRequest_extension_get", LibLogoutRequest_extension_get, METH_VARARGS, NULL},
    {"LibLogoutRequest_extension_set", LibLogoutRequest_extension_set, METH_VARARGS, NULL},
    {"LibLogoutRequest_providerId_get", LibLogoutRequest_providerId_get, METH_VARARGS, NULL},
    {"LibLogoutRequest_providerId_set", LibLogoutRequest_providerId_set, METH_VARARGS, NULL},
    {"LibLogoutRequest_nameIdentifier_get", LibLogoutRequest_nameIdentifier_get, METH_VARARGS, NULL},
    {"LibLogoutRequest_nameIdentifier_set", LibLogoutRequest_nameIdentifier_set, METH_VARARGS, NULL},
    {"LibLogoutRequest_sessionIndex_get", LibLogoutRequest_sessionIndex_get, METH_VARARGS, NULL},
    {"LibLogoutRequest_sessionIndex_set", LibLogoutRequest_sessionIndex_set, METH_VARARGS, NULL},
    {"LibLogoutRequest_relayState_get", LibLogoutRequest_relayState_get, METH_VARARGS, NULL},
    {"LibLogoutRequest_relayState_set", LibLogoutRequest_relayState_set, METH_VARARGS, NULL},
    {"LibLogoutRequest_consent_get", LibLogoutRequest_consent_get, METH_VARARGS, NULL},
    {"LibLogoutRequest_consent_set", LibLogoutRequest_consent_set, METH_VARARGS, NULL},
    {"LibLogoutRequest_notOnOrAfter_get", LibLogoutRequest_notOnOrAfter_get, METH_VARARGS, NULL},
    {"LibLogoutRequest_notOnOrAfter_set", LibLogoutRequest_notOnOrAfter_set, METH_VARARGS, NULL},
    {"lib_logout_request_set_session_indexes", lib_logout_request_set_session_indexes, METH_VARARGS, NULL},
    {"lib_logout_request_get_session_indexes", lib_logout_request_get_session_indexes, METH_VARARGS, NULL},
    {"LibAssertion_inResponseTo_get", LibAssertion_inResponseTo_get, METH_VARARGS, NULL},
    {"LibAssertion_inResponseTo_set", LibAssertion_inResponseTo_set, METH_VARARGS, NULL},
    {"SamlpRequest_assertionArtifact_get", SamlpRequest_assertionArtifact_get, METH_VARARGS, NULL},
    {"SamlpRequest_assertionArtifact_set", SamlpRequest_assertionArtifact_set, METH_VARARGS, NULL},
    {"SamlSubjectStatementAbstract_subject_get", SamlSubjectStatementAbstract_subject_get, METH_VARARGS, NULL},
    {"SamlSubjectStatementAbstract_subject_set", SamlSubjectStatementAbstract_subject_set, METH_VARARGS, NULL},
    {"SamlAttributeStatement_attribute_get", SamlAttributeStatement_attribute_get, METH_VARARGS, NULL},
    {"SamlAttributeStatement_attribute_set", SamlAttributeStatement_attribute_set, METH_VARARGS, NULL},
    {"SamlSubjectLocality_ipAddress_get", SamlSubjectLocality_ipAddress_get, METH_VARARGS, NULL},
    {"SamlSubjectLocality_ipAddress_set", SamlSubjectLocality_ipAddress_set, METH_VARARGS, NULL},
    {"SamlSubjectLocality_dnsAddress_get", SamlSubjectLocality_dnsAddress_get, METH_VARARGS, NULL},
    {"SamlSubjectLocality_dnsAddress_set", SamlSubjectLocality_dnsAddress_set, METH_VARARGS, NULL},
    {"LibIDPEntries_idpEntry_get", LibIDPEntries_idpEntry_get, METH_VARARGS, NULL},
    {"LibIDPEntries_idpEntry_set", LibIDPEntries_idpEntry_set, METH_VARARGS, NULL},
    {"LibStatusResponse_extension_get", LibStatusResponse_extension_get, METH_VARARGS, NULL},
    {"LibStatusResponse_extension_set", LibStatusResponse_extension_set, METH_VARARGS, NULL},
    {"LibStatusResponse_providerId_get", LibStatusResponse_providerId_get, METH_VARARGS, NULL},
    {"LibStatusResponse_providerId_set", LibStatusResponse_providerId_set, METH_VARARGS, NULL},
    {"LibStatusResponse_status_get", LibStatusResponse_status_get, METH_VARARGS, NULL},
    {"LibStatusResponse_status_set", LibStatusResponse_status_set, METH_VARARGS, NULL},
    {"LibStatusResponse_relayState_get", LibStatusResponse_relayState_get, METH_VARARGS, NULL},
    {"LibStatusResponse_relayState_set", LibStatusResponse_relayState_set, METH_VARARGS, NULL},
    {"SamlNameIdentifier_nameQualifier_get", SamlNameIdentifier_nameQualifier_get, METH_VARARGS, NULL},
    {"SamlNameIdentifier_nameQualifier_set", SamlNameIdentifier_nameQualifier_set, METH_VARARGS, NULL},
    {"SamlNameIdentifier_format_get", SamlNameIdentifier_format_get, METH_VARARGS, NULL},
    {"SamlNameIdentifier_format_set", SamlNameIdentifier_format_set, METH_VARARGS, NULL},
    {"SamlNameIdentifier_content_get", SamlNameIdentifier_content_get, METH_VARARGS, NULL},
    {"SamlNameIdentifier_content_set", SamlNameIdentifier_content_set, METH_VARARGS, NULL},
    {"saml_name_identifier_equals", saml_name_identifier_equals, METH_VARARGS, NULL},
    {"MiscTextNode_content_get", MiscTextNode_content_get, METH_VARARGS, NULL},
    {"MiscTextNode_content_set", MiscTextNode_content_set, METH_VARARGS, NULL},
    {"MiscTextNode_name_get", MiscTextNode_name_get, METH_VARARGS, NULL},
    {"MiscTextNode_name_set", MiscTextNode_name_set, METH_VARARGS, NULL},
    {"MiscTextNode_nsHref_get", MiscTextNode_nsHref_get, METH_VARARGS, NULL},
    {"MiscTextNode_nsHref_set", MiscTextNode_nsHref_set, METH_VARARGS, NULL},
    {"MiscTextNode_nsPrefix_get", MiscTextNode_nsPrefix_get, METH_VARARGS, NULL},
    {"MiscTextNode_nsPrefix_set", MiscTextNode_nsPrefix_set, METH_VARARGS, NULL},
    {"MiscTextNode_textChild_get", MiscTextNode_textChild_get, METH_VARARGS, NULL},
    {"MiscTextNode_textChild_set", MiscTextNode_textChild_set, METH_VARARGS, NULL},
    {"misc_text_node_set_xml_content", misc_text_node_set_xml_content, METH_VARARGS, NULL},
    {"misc_text_node_get_xml_content", misc_text_node_get_xml_content, METH_VARARGS, NULL},
    {"SamlAudienceRestrictionCondition_audience_get", SamlAudienceRestrictionCondition_audience_get, METH_VARARGS, NULL},
    {"SamlAudienceRestrictionCondition_audience_set", SamlAudienceRestrictionCondition_audience_set, METH_VARARGS, NULL},
    {"SamlSubjectConfirmation_confirmationMethod_get", SamlSubjectConfirmation_confirmationMethod_get, METH_VARARGS, NULL},
    {"SamlSubjectConfirmation_confirmationMethod_set", SamlSubjectConfirmation_confirmationMethod_set, METH_VARARGS, NULL},
    {"SamlSubjectConfirmation_subjectConfirmationData_get", SamlSubjectConfirmation_subjectConfirmationData_get, METH_VARARGS, NULL},
    {"SamlSubjectConfirmation_subjectConfirmationData_set", SamlSubjectConfirmation_subjectConfirmationData_set, METH_VARARGS, NULL},
    {"SamlSubjectConfirmation_keyInfo_get", SamlSubjectConfirmation_keyInfo_get, METH_VARARGS, NULL},
    {"SamlSubjectConfirmation_keyInfo_set", SamlSubjectConfirmation_keyInfo_set, METH_VARARGS, NULL},
    {"SamlAuthorityBinding_authorityKind_get", SamlAuthorityBinding_authorityKind_get, METH_VARARGS, NULL},
    {"SamlAuthorityBinding_authorityKind_set", SamlAuthorityBinding_authorityKind_set, METH_VARARGS, NULL},
    {"SamlAuthorityBinding_location_get", SamlAuthorityBinding_location_get, METH_VARARGS, NULL},
    {"SamlAuthorityBinding_location_set", SamlAuthorityBinding_location_set, METH_VARARGS, NULL},
    {"SamlAuthorityBinding_binding_get", SamlAuthorityBinding_binding_get, METH_VARARGS, NULL},
    {"SamlAuthorityBinding_binding_set", SamlAuthorityBinding_binding_set, METH_VARARGS, NULL},
    {"LibScoping_proxyCount_get", LibScoping_proxyCount_get, METH_VARARGS, NULL},
    {"LibScoping_proxyCount_set", LibScoping_proxyCount_set, METH_VARARGS, NULL},
    {"LibScoping_idpList_get", LibScoping_idpList_get, METH_VARARGS, NULL},
    {"LibScoping_idpList_set", LibScoping_idpList_set, METH_VARARGS, NULL},
    {"SamlpStatus_statusCode_get", SamlpStatus_statusCode_get, METH_VARARGS, NULL},
    {"SamlpStatus_statusCode_set", SamlpStatus_statusCode_set, METH_VARARGS, NULL},
    {"SamlpStatus_statusMessage_get", SamlpStatus_statusMessage_get, METH_VARARGS, NULL},
    {"SamlpStatus_statusMessage_set", SamlpStatus_statusMessage_set, METH_VARARGS, NULL},
    {"LibRegisterNameIdentifierRequest_extension_get", LibRegisterNameIdentifierRequest_extension_get, METH_VARARGS, NULL},
    {"LibRegisterNameIdentifierRequest_extension_set", LibRegisterNameIdentifierRequest_extension_set, METH_VARARGS, NULL},
    {"LibRegisterNameIdentifierRequest_providerId_get", LibRegisterNameIdentifierRequest_providerId_get, METH_VARARGS, NULL},
    {"LibRegisterNameIdentifierRequest_providerId_set", LibRegisterNameIdentifierRequest_providerId_set, METH_VARARGS, NULL},
    {"LibRegisterNameIdentifierRequest_idpProvidedNameIdentifier_get", LibRegisterNameIdentifierRequest_idpProvidedNameIdentifier_get, METH_VARARGS, NULL},
    {"LibRegisterNameIdentifierRequest_idpProvidedNameIdentifier_set", LibRegisterNameIdentifierRequest_idpProvidedNameIdentifier_set, METH_VARARGS, NULL},
    {"LibRegisterNameIdentifierRequest_spProvidedNameIdentifier_get", LibRegisterNameIdentifierRequest_spProvidedNameIdentifier_get, METH_VARARGS, NULL},
    {"LibRegisterNameIdentifierRequest_spProvidedNameIdentifier_set", LibRegisterNameIdentifierRequest_spProvidedNameIdentifier_set, METH_VARARGS, NULL},
    {"LibRegisterNameIdentifierRequest_oldProvidedNameIdentifier_get", LibRegisterNameIdentifierRequest_oldProvidedNameIdentifier_get, METH_VARARGS, NULL},
    {"LibRegisterNameIdentifierRequest_oldProvidedNameIdentifier_set", LibRegisterNameIdentifierRequest_oldProvidedNameIdentifier_set, METH_VARARGS, NULL},
    {"LibRegisterNameIdentifierRequest_relayState_get", LibRegisterNameIdentifierRequest_relayState_get, METH_VARARGS, NULL},
    {"LibRegisterNameIdentifierRequest_relayState_set", LibRegisterNameIdentifierRequest_relayState_set, METH_VARARGS, NULL},
    {"SamlAuthenticationStatement_subjectLocality_get", SamlAuthenticationStatement_subjectLocality_get, METH_VARARGS, NULL},
    {"SamlAuthenticationStatement_subjectLocality_set", SamlAuthenticationStatement_subjectLocality_set, METH_VARARGS, NULL},
    {"SamlAuthenticationStatement_authorityBinding_get", SamlAuthenticationStatement_authorityBinding_get, METH_VARARGS, NULL},
    {"SamlAuthenticationStatement_authorityBinding_set", SamlAuthenticationStatement_authorityBinding_set, METH_VARARGS, NULL},
    {"SamlAuthenticationStatement_authenticationMethod_get", SamlAuthenticationStatement_authenticationMethod_get, METH_VARARGS, NULL},
    {"SamlAuthenticationStatement_authenticationMethod_set", SamlAuthenticationStatement_authenticationMethod_set, METH_VARARGS, NULL},
    {"SamlAuthenticationStatement_authenticationInstant_get", SamlAuthenticationStatement_authenticationInstant_get, METH_VARARGS, NULL},
    {"SamlAuthenticationStatement_authenticationInstant_set", SamlAuthenticationStatement_authenticationInstant_set, METH_VARARGS, NULL},
    {"LibAuthenticationStatement_authnContext_get", LibAuthenticationStatement_authnContext_get, METH_VARARGS, NULL},
    {"LibAuthenticationStatement_authnContext_set", LibAuthenticationStatement_authnContext_set, METH_VARARGS, NULL},
    {"LibAuthenticationStatement_reauthenticateOnOrAfter_get", LibAuthenticationStatement_reauthenticateOnOrAfter_get, METH_VARARGS, NULL},
    {"LibAuthenticationStatement_reauthenticateOnOrAfter_set", LibAuthenticationStatement_reauthenticateOnOrAfter_set, METH_VARARGS, NULL},
    {"LibAuthenticationStatement_sessionIndex_get", LibAuthenticationStatement_sessionIndex_get, METH_VARARGS, NULL},
    {"LibAuthenticationStatement_sessionIndex_set", LibAuthenticationStatement_sessionIndex_set, METH_VARARGS, NULL},
    {"SoapFault_faultcode_get", SoapFault_faultcode_get, METH_VARARGS, NULL},
    {"SoapFault_faultcode_set", SoapFault_faultcode_set, METH_VARARGS, NULL},
    {"SoapFault_faultstring_get", SoapFault_faultstring_get, METH_VARARGS, NULL},
    {"SoapFault_faultstring_set", SoapFault_faultstring_set, METH_VARARGS, NULL},
    {"SoapFault_faultactor_get", SoapFault_faultactor_get, METH_VARARGS, NULL},
    {"SoapFault_faultactor_set", SoapFault_faultactor_set, METH_VARARGS, NULL},
    {"SoapFault_detail_get", SoapFault_detail_get, METH_VARARGS, NULL},
    {"SoapFault_detail_set", SoapFault_detail_set, METH_VARARGS, NULL},
    {"SoapEnvelope_header_get", SoapEnvelope_header_get, METH_VARARGS, NULL},
    {"SoapEnvelope_header_set", SoapEnvelope_header_set, METH_VARARGS, NULL},
    {"SoapEnvelope_body_get", SoapEnvelope_body_get, METH_VARARGS, NULL},
    {"SoapEnvelope_body_set", SoapEnvelope_body_set, METH_VARARGS, NULL},
    {"SoapHeader_other_get", SoapHeader_other_get, METH_VARARGS, NULL},
    {"SoapHeader_other_set", SoapHeader_other_set, METH_VARARGS, NULL},
    {"SoapBody_any_get", SoapBody_any_get, METH_VARARGS, NULL},
    {"SoapBody_any_set", SoapBody_any_set, METH_VARARGS, NULL},
    {"SoapBody_id_get", SoapBody_id_get, METH_VARARGS, NULL},
    {"SoapBody_id_set", SoapBody_id_set, METH_VARARGS, NULL},
    {"SoapDetail_any_get", SoapDetail_any_get, METH_VARARGS, NULL},
    {"SoapDetail_any_set", SoapDetail_any_set, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmation_baseId_get", Saml2SubjectConfirmation_baseId_get, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmation_baseId_set", Saml2SubjectConfirmation_baseId_set, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmation_nameId_get", Saml2SubjectConfirmation_nameId_get, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmation_nameId_set", Saml2SubjectConfirmation_nameId_set, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmation_encryptedId_get", Saml2SubjectConfirmation_encryptedId_get, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmation_encryptedId_set", Saml2SubjectConfirmation_encryptedId_set, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmation_subjectConfirmationData_get", Saml2SubjectConfirmation_subjectConfirmationData_get, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmation_subjectConfirmationData_set", Saml2SubjectConfirmation_subjectConfirmationData_set, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmation_method_get", Saml2SubjectConfirmation_method_get, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmation_method_set", Saml2SubjectConfirmation_method_set, METH_VARARGS, NULL},
    {"Samlp2IDPList_idpEntry_get", Samlp2IDPList_idpEntry_get, METH_VARARGS, NULL},
    {"Samlp2IDPList_idpEntry_set", Samlp2IDPList_idpEntry_set, METH_VARARGS, NULL},
    {"Samlp2IDPList_getComplete_get", Samlp2IDPList_getComplete_get, METH_VARARGS, NULL},
    {"Samlp2IDPList_getComplete_set", Samlp2IDPList_getComplete_set, METH_VARARGS, NULL},
    {"Saml2AuthnContext_authnContextClassRef_get", Saml2AuthnContext_authnContextClassRef_get, METH_VARARGS, NULL},
    {"Saml2AuthnContext_authnContextClassRef_set", Saml2AuthnContext_authnContextClassRef_set, METH_VARARGS, NULL},
    {"Saml2AuthnContext_authnContextDeclRef_get", Saml2AuthnContext_authnContextDeclRef_get, METH_VARARGS, NULL},
    {"Saml2AuthnContext_authnContextDeclRef_set", Saml2AuthnContext_authnContextDeclRef_set, METH_VARARGS, NULL},
    {"Saml2AuthnContext_authenticatingAuthority_get", Saml2AuthnContext_authenticatingAuthority_get, METH_VARARGS, NULL},
    {"Saml2AuthnContext_authenticatingAuthority_set", Saml2AuthnContext_authenticatingAuthority_set, METH_VARARGS, NULL},
    {"Saml2BaseIDAbstract_nameQualifier_get", Saml2BaseIDAbstract_nameQualifier_get, METH_VARARGS, NULL},
    {"Saml2BaseIDAbstract_nameQualifier_set", Saml2BaseIDAbstract_nameQualifier_set, METH_VARARGS, NULL},
    {"Saml2BaseIDAbstract_spNameQualifier_get", Saml2BaseIDAbstract_spNameQualifier_get, METH_VARARGS, NULL},
    {"Saml2BaseIDAbstract_spNameQualifier_set", Saml2BaseIDAbstract_spNameQualifier_set, METH_VARARGS, NULL},
    {"Saml2SubjectLocality_address_get", Saml2SubjectLocality_address_get, METH_VARARGS, NULL},
    {"Saml2SubjectLocality_address_set", Saml2SubjectLocality_address_set, METH_VARARGS, NULL},
    {"Saml2SubjectLocality_dnsName_get", Saml2SubjectLocality_dnsName_get, METH_VARARGS, NULL},
    {"Saml2SubjectLocality_dnsName_set", Saml2SubjectLocality_dnsName_set, METH_VARARGS, NULL},
    {"Samlp2Status_statusCode_get", Samlp2Status_statusCode_get, METH_VARARGS, NULL},
    {"Samlp2Status_statusCode_set", Samlp2Status_statusCode_set, METH_VARARGS, NULL},
    {"Samlp2Status_statusMessage_get", Samlp2Status_statusMessage_get, METH_VARARGS, NULL},
    {"Samlp2Status_statusMessage_set", Samlp2Status_statusMessage_set, METH_VARARGS, NULL},
    {"Samlp2Status_statusDetail_get", Samlp2Status_statusDetail_get, METH_VARARGS, NULL},
    {"Samlp2Status_statusDetail_set", Samlp2Status_statusDetail_set, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_issuer_get", Samlp2RequestAbstract_issuer_get, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_issuer_set", Samlp2RequestAbstract_issuer_set, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_extensions_get", Samlp2RequestAbstract_extensions_get, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_extensions_set", Samlp2RequestAbstract_extensions_set, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_id_get", Samlp2RequestAbstract_id_get, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_id_set", Samlp2RequestAbstract_id_set, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_version_get", Samlp2RequestAbstract_version_get, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_version_set", Samlp2RequestAbstract_version_set, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_issueInstant_get", Samlp2RequestAbstract_issueInstant_get, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_issueInstant_set", Samlp2RequestAbstract_issueInstant_set, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_destination_get", Samlp2RequestAbstract_destination_get, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_destination_set", Samlp2RequestAbstract_destination_set, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_consent_get", Samlp2RequestAbstract_consent_get, METH_VARARGS, NULL},
    {"Samlp2RequestAbstract_consent_set", Samlp2RequestAbstract_consent_set, METH_VARARGS, NULL},
    {"Samlp2AssertionIDRequest_assertionIdRef_get", Samlp2AssertionIDRequest_assertionIdRef_get, METH_VARARGS, NULL},
    {"Samlp2AssertionIDRequest_assertionIdRef_set", Samlp2AssertionIDRequest_assertionIdRef_set, METH_VARARGS, NULL},
    {"Samlp2NameIDPolicy_format_get", Samlp2NameIDPolicy_format_get, METH_VARARGS, NULL},
    {"Samlp2NameIDPolicy_format_set", Samlp2NameIDPolicy_format_set, METH_VARARGS, NULL},
    {"Samlp2NameIDPolicy_spNameQualifier_get", Samlp2NameIDPolicy_spNameQualifier_get, METH_VARARGS, NULL},
    {"Samlp2NameIDPolicy_spNameQualifier_set", Samlp2NameIDPolicy_spNameQualifier_set, METH_VARARGS, NULL},
    {"Samlp2NameIDPolicy_allowCreate_get", Samlp2NameIDPolicy_allowCreate_get, METH_VARARGS, NULL},
    {"Samlp2NameIDPolicy_allowCreate_set", Samlp2NameIDPolicy_allowCreate_set, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_subject_get", Samlp2AuthnRequest_subject_get, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_subject_set", Samlp2AuthnRequest_subject_set, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_nameIdPolicy_get", Samlp2AuthnRequest_nameIdPolicy_get, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_nameIdPolicy_set", Samlp2AuthnRequest_nameIdPolicy_set, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_conditions_get", Samlp2AuthnRequest_conditions_get, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_conditions_set", Samlp2AuthnRequest_conditions_set, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_requestedAuthnContext_get", Samlp2AuthnRequest_requestedAuthnContext_get, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_requestedAuthnContext_set", Samlp2AuthnRequest_requestedAuthnContext_set, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_scoping_get", Samlp2AuthnRequest_scoping_get, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_scoping_set", Samlp2AuthnRequest_scoping_set, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_forceAuthn_get", Samlp2AuthnRequest_forceAuthn_get, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_forceAuthn_set", Samlp2AuthnRequest_forceAuthn_set, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_isPassive_get", Samlp2AuthnRequest_isPassive_get, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_isPassive_set", Samlp2AuthnRequest_isPassive_set, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_protocolBinding_get", Samlp2AuthnRequest_protocolBinding_get, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_protocolBinding_set", Samlp2AuthnRequest_protocolBinding_set, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_assertionConsumerServiceIndex_get", Samlp2AuthnRequest_assertionConsumerServiceIndex_get, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_assertionConsumerServiceIndex_set", Samlp2AuthnRequest_assertionConsumerServiceIndex_set, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_assertionConsumerServiceUrl_get", Samlp2AuthnRequest_assertionConsumerServiceUrl_get, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_assertionConsumerServiceUrl_set", Samlp2AuthnRequest_assertionConsumerServiceUrl_set, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_attributeConsumingServiceIndex_get", Samlp2AuthnRequest_attributeConsumingServiceIndex_get, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_attributeConsumingServiceIndex_set", Samlp2AuthnRequest_attributeConsumingServiceIndex_set, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_providerName_get", Samlp2AuthnRequest_providerName_get, METH_VARARGS, NULL},
    {"Samlp2AuthnRequest_providerName_set", Samlp2AuthnRequest_providerName_set, METH_VARARGS, NULL},
    {"Saml2ProxyRestriction_audience_get", Saml2ProxyRestriction_audience_get, METH_VARARGS, NULL},
    {"Saml2ProxyRestriction_audience_set", Saml2ProxyRestriction_audience_set, METH_VARARGS, NULL},
    {"Saml2ProxyRestriction_count_get", Saml2ProxyRestriction_count_get, METH_VARARGS, NULL},
    {"Saml2ProxyRestriction_count_set", Saml2ProxyRestriction_count_set, METH_VARARGS, NULL},
    {"Saml2Subject_baseId_get", Saml2Subject_baseId_get, METH_VARARGS, NULL},
    {"Saml2Subject_baseId_set", Saml2Subject_baseId_set, METH_VARARGS, NULL},
    {"Saml2Subject_nameId_get", Saml2Subject_nameId_get, METH_VARARGS, NULL},
    {"Saml2Subject_nameId_set", Saml2Subject_nameId_set, METH_VARARGS, NULL},
    {"Saml2Subject_encryptedId_get", Saml2Subject_encryptedId_get, METH_VARARGS, NULL},
    {"Saml2Subject_encryptedId_set", Saml2Subject_encryptedId_set, METH_VARARGS, NULL},
    {"Saml2Subject_subjectConfirmation_get", Saml2Subject_subjectConfirmation_get, METH_VARARGS, NULL},
    {"Saml2Subject_subjectConfirmation_set", Saml2Subject_subjectConfirmation_set, METH_VARARGS, NULL},
    {"Samlp2RequestedAuthnContext_authnContextClassRef_get", Samlp2RequestedAuthnContext_authnContextClassRef_get, METH_VARARGS, NULL},
    {"Samlp2RequestedAuthnContext_authnContextClassRef_set", Samlp2RequestedAuthnContext_authnContextClassRef_set, METH_VARARGS, NULL},
    {"Samlp2RequestedAuthnContext_authnContextDeclRef_get", Samlp2RequestedAuthnContext_authnContextDeclRef_get, METH_VARARGS, NULL},
    {"Samlp2RequestedAuthnContext_authnContextDeclRef_set", Samlp2RequestedAuthnContext_authnContextDeclRef_set, METH_VARARGS, NULL},
    {"Samlp2RequestedAuthnContext_comparison_get", Samlp2RequestedAuthnContext_comparison_get, METH_VARARGS, NULL},
    {"Samlp2RequestedAuthnContext_comparison_set", Samlp2RequestedAuthnContext_comparison_set, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_issuer_get", Samlp2StatusResponse_issuer_get, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_issuer_set", Samlp2StatusResponse_issuer_set, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_extensions_get", Samlp2StatusResponse_extensions_get, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_extensions_set", Samlp2StatusResponse_extensions_set, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_status_get", Samlp2StatusResponse_status_get, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_status_set", Samlp2StatusResponse_status_set, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_id_get", Samlp2StatusResponse_id_get, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_id_set", Samlp2StatusResponse_id_set, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_inResponseTo_get", Samlp2StatusResponse_inResponseTo_get, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_inResponseTo_set", Samlp2StatusResponse_inResponseTo_set, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_version_get", Samlp2StatusResponse_version_get, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_version_set", Samlp2StatusResponse_version_set, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_issueInstant_get", Samlp2StatusResponse_issueInstant_get, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_issueInstant_set", Samlp2StatusResponse_issueInstant_set, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_destination_get", Samlp2StatusResponse_destination_get, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_destination_set", Samlp2StatusResponse_destination_set, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_consent_get", Samlp2StatusResponse_consent_get, METH_VARARGS, NULL},
    {"Samlp2StatusResponse_consent_set", Samlp2StatusResponse_consent_set, METH_VARARGS, NULL},
    {"Samlp2ArtifactResponse_any_get", Samlp2ArtifactResponse_any_get, METH_VARARGS, NULL},
    {"Samlp2ArtifactResponse_any_set", Samlp2ArtifactResponse_any_set, METH_VARARGS, NULL},
    {"Saml2AudienceRestriction_audience_get", Saml2AudienceRestriction_audience_get, METH_VARARGS, NULL},
    {"Saml2AudienceRestriction_audience_set", Saml2AudienceRestriction_audience_set, METH_VARARGS, NULL},
    {"Saml2Attribute_attributeValue_get", Saml2Attribute_attributeValue_get, METH_VARARGS, NULL},
    {"Saml2Attribute_attributeValue_set", Saml2Attribute_attributeValue_set, METH_VARARGS, NULL},
    {"Saml2Attribute_name_get", Saml2Attribute_name_get, METH_VARARGS, NULL},
    {"Saml2Attribute_name_set", Saml2Attribute_name_set, METH_VARARGS, NULL},
    {"Saml2Attribute_nameFormat_get", Saml2Attribute_nameFormat_get, METH_VARARGS, NULL},
    {"Saml2Attribute_nameFormat_set", Saml2Attribute_nameFormat_set, METH_VARARGS, NULL},
    {"Saml2Attribute_friendlyName_get", Saml2Attribute_friendlyName_get, METH_VARARGS, NULL},
    {"Saml2Attribute_friendlyName_set", Saml2Attribute_friendlyName_set, METH_VARARGS, NULL},
    {"Samlp2NameIDMappingRequest_baseId_get", Samlp2NameIDMappingRequest_baseId_get, METH_VARARGS, NULL},
    {"Samlp2NameIDMappingRequest_baseId_set", Samlp2NameIDMappingRequest_baseId_set, METH_VARARGS, NULL},
    {"Samlp2NameIDMappingRequest_nameId_get", Samlp2NameIDMappingRequest_nameId_get, METH_VARARGS, NULL},
    {"Samlp2NameIDMappingRequest_nameId_set", Samlp2NameIDMappingRequest_nameId_set, METH_VARARGS, NULL},
    {"Samlp2NameIDMappingRequest_encryptedId_get", Samlp2NameIDMappingRequest_encryptedId_get, METH_VARARGS, NULL},
    {"Samlp2NameIDMappingRequest_encryptedId_set", Samlp2NameIDMappingRequest_encryptedId_set, METH_VARARGS, NULL},
    {"Samlp2NameIDMappingRequest_nameIdPolicy_get", Samlp2NameIDMappingRequest_nameIdPolicy_get, METH_VARARGS, NULL},
    {"Samlp2NameIDMappingRequest_nameIdPolicy_set", Samlp2NameIDMappingRequest_nameIdPolicy_set, METH_VARARGS, NULL},
    {"Samlp2ManageNameIDRequest_nameId_get", Samlp2ManageNameIDRequest_nameId_get, METH_VARARGS, NULL},
    {"Samlp2ManageNameIDRequest_nameId_set", Samlp2ManageNameIDRequest_nameId_set, METH_VARARGS, NULL},
    {"Samlp2ManageNameIDRequest_encryptedId_get", Samlp2ManageNameIDRequest_encryptedId_get, METH_VARARGS, NULL},
    {"Samlp2ManageNameIDRequest_encryptedId_set", Samlp2ManageNameIDRequest_encryptedId_set, METH_VARARGS, NULL},
    {"Samlp2ManageNameIDRequest_newId_get", Samlp2ManageNameIDRequest_newId_get, METH_VARARGS, NULL},
    {"Samlp2ManageNameIDRequest_newId_set", Samlp2ManageNameIDRequest_newId_set, METH_VARARGS, NULL},
    {"Samlp2ManageNameIDRequest_newEncryptedId_get", Samlp2ManageNameIDRequest_newEncryptedId_get, METH_VARARGS, NULL},
    {"Samlp2ManageNameIDRequest_newEncryptedId_set", Samlp2ManageNameIDRequest_newEncryptedId_set, METH_VARARGS, NULL},
    {"Samlp2ManageNameIDRequest_terminate_get", Samlp2ManageNameIDRequest_terminate_get, METH_VARARGS, NULL},
    {"Samlp2ManageNameIDRequest_terminate_set", Samlp2ManageNameIDRequest_terminate_set, METH_VARARGS, NULL},
    {"Samlp2StatusCode_statusCode_get", Samlp2StatusCode_statusCode_get, METH_VARARGS, NULL},
    {"Samlp2StatusCode_statusCode_set", Samlp2StatusCode_statusCode_set, METH_VARARGS, NULL},
    {"Samlp2StatusCode_value_get", Samlp2StatusCode_value_get, METH_VARARGS, NULL},
    {"Samlp2StatusCode_value_set", Samlp2StatusCode_value_set, METH_VARARGS, NULL},
    {"Saml2Advice_assertionIdRef_get", Saml2Advice_assertionIdRef_get, METH_VARARGS, NULL},
    {"Saml2Advice_assertionIdRef_set", Saml2Advice_assertionIdRef_set, METH_VARARGS, NULL},
    {"Saml2Advice_assertionUriRef_get", Saml2Advice_assertionUriRef_get, METH_VARARGS, NULL},
    {"Saml2Advice_assertionUriRef_set", Saml2Advice_assertionUriRef_set, METH_VARARGS, NULL},
    {"Saml2Advice_assertion_get", Saml2Advice_assertion_get, METH_VARARGS, NULL},
    {"Saml2Advice_assertion_set", Saml2Advice_assertion_set, METH_VARARGS, NULL},
    {"Saml2Advice_encryptedAssertion_get", Saml2Advice_encryptedAssertion_get, METH_VARARGS, NULL},
    {"Saml2Advice_encryptedAssertion_set", Saml2Advice_encryptedAssertion_set, METH_VARARGS, NULL},
    {"Saml2Action_content_get", Saml2Action_content_get, METH_VARARGS, NULL},
    {"Saml2Action_content_set", Saml2Action_content_set, METH_VARARGS, NULL},
    {"Saml2Action_namespace_get", Saml2Action_namespace_get, METH_VARARGS, NULL},
    {"Saml2Action_namespace_set", Saml2Action_namespace_set, METH_VARARGS, NULL},
    {"Saml2EncryptedElement_encryptedData_get", Saml2EncryptedElement_encryptedData_get, METH_VARARGS, NULL},
    {"Saml2EncryptedElement_encryptedData_set", Saml2EncryptedElement_encryptedData_set, METH_VARARGS, NULL},
    {"Saml2EncryptedElement_encryptedKey_get", Saml2EncryptedElement_encryptedKey_get, METH_VARARGS, NULL},
    {"Saml2EncryptedElement_encryptedKey_set", Saml2EncryptedElement_encryptedKey_set, METH_VARARGS, NULL},
    {"Saml2EncryptedElement_originalData_get", Saml2EncryptedElement_originalData_get, METH_VARARGS, NULL},
    {"Saml2EncryptedElement_originalData_set", Saml2EncryptedElement_originalData_set, METH_VARARGS, NULL},
    {"saml2_encrypted_element_server_decrypt", saml2_encrypted_element_server_decrypt, METH_VARARGS, NULL},
    {"Saml2Assertion_issuer_get", Saml2Assertion_issuer_get, METH_VARARGS, NULL},
    {"Saml2Assertion_issuer_set", Saml2Assertion_issuer_set, METH_VARARGS, NULL},
    {"Saml2Assertion_subject_get", Saml2Assertion_subject_get, METH_VARARGS, NULL},
    {"Saml2Assertion_subject_set", Saml2Assertion_subject_set, METH_VARARGS, NULL},
    {"Saml2Assertion_conditions_get", Saml2Assertion_conditions_get, METH_VARARGS, NULL},
    {"Saml2Assertion_conditions_set", Saml2Assertion_conditions_set, METH_VARARGS, NULL},
    {"Saml2Assertion_advice_get", Saml2Assertion_advice_get, METH_VARARGS, NULL},
    {"Saml2Assertion_advice_set", Saml2Assertion_advice_set, METH_VARARGS, NULL},
    {"Saml2Assertion_statement_get", Saml2Assertion_statement_get, METH_VARARGS, NULL},
    {"Saml2Assertion_statement_set", Saml2Assertion_statement_set, METH_VARARGS, NULL},
    {"Saml2Assertion_authnStatement_get", Saml2Assertion_authnStatement_get, METH_VARARGS, NULL},
    {"Saml2Assertion_authnStatement_set", Saml2Assertion_authnStatement_set, METH_VARARGS, NULL},
    {"Saml2Assertion_authzDecisionStatement_get", Saml2Assertion_authzDecisionStatement_get, METH_VARARGS, NULL},
    {"Saml2Assertion_authzDecisionStatement_set", Saml2Assertion_authzDecisionStatement_set, METH_VARARGS, NULL},
    {"Saml2Assertion_attributeStatement_get", Saml2Assertion_attributeStatement_get, METH_VARARGS, NULL},
    {"Saml2Assertion_attributeStatement_set", Saml2Assertion_attributeStatement_set, METH_VARARGS, NULL},
    {"Saml2Assertion_version_get", Saml2Assertion_version_get, METH_VARARGS, NULL},
    {"Saml2Assertion_version_set", Saml2Assertion_version_set, METH_VARARGS, NULL},
    {"Saml2Assertion_id_get", Saml2Assertion_id_get, METH_VARARGS, NULL},
    {"Saml2Assertion_id_set", Saml2Assertion_id_set, METH_VARARGS, NULL},
    {"Saml2Assertion_issueInstant_get", Saml2Assertion_issueInstant_get, METH_VARARGS, NULL},
    {"Saml2Assertion_issueInstant_set", Saml2Assertion_issueInstant_set, METH_VARARGS, NULL},
    {"saml2_assertion_has_audience_restriction", saml2_assertion_has_audience_restriction, METH_VARARGS, NULL},
    {"saml2_assertion_is_audience_restricted", saml2_assertion_is_audience_restricted, METH_VARARGS, NULL},
    {"saml2_assertion_set_subject_name_id", saml2_assertion_set_subject_name_id, METH_VARARGS, NULL},
    {"saml2_assertion_set_subject_confirmation_name_id", saml2_assertion_set_subject_confirmation_name_id, METH_VARARGS, NULL},
    {"saml2_assertion_set_subject_confirmation_data", saml2_assertion_set_subject_confirmation_data, METH_VARARGS, NULL},
    {"saml2_assertion_set_basic_conditions", saml2_assertion_set_basic_conditions, METH_VARARGS, NULL},
    {"saml2_assertion_add_audience_restriction", saml2_assertion_add_audience_restriction, METH_VARARGS, NULL},
    {"saml2_assertion_add_proxy_limit", saml2_assertion_add_proxy_limit, METH_VARARGS, NULL},
    {"saml2_assertion_validate_conditions", saml2_assertion_validate_conditions, METH_VARARGS, NULL},
    {"saml2_assertion_get_issuer_provider", saml2_assertion_get_issuer_provider, METH_VARARGS, NULL},
    {"saml2_assertion_add_attribute_with_node", saml2_assertion_add_attribute_with_node, METH_VARARGS, NULL},
    {"saml2_assertion_get_subject_confirmation_data", saml2_assertion_get_subject_confirmation_data, METH_VARARGS, NULL},
    {"saml2_assertion_get_in_response_to", saml2_assertion_get_in_response_to, METH_VARARGS, NULL},
    {"saml2_assertion_decrypt_subject", saml2_assertion_decrypt_subject, METH_VARARGS, NULL},
    {"saml2_assertion_validate_time_checks", saml2_assertion_validate_time_checks, METH_VARARGS, NULL},
    {"saml2_assertion_validate_audience", saml2_assertion_validate_audience, METH_VARARGS, NULL},
    {"saml2_assertion_has_one_time_use", saml2_assertion_has_one_time_use, METH_VARARGS, NULL},
    {"saml2_assertion_allows_proxying", saml2_assertion_allows_proxying, METH_VARARGS, NULL},
    {"saml2_assertion_allows_proxying_to", saml2_assertion_allows_proxying_to, METH_VARARGS, NULL},
    {"saml2_assertion_set_one_time_use", saml2_assertion_set_one_time_use, METH_VARARGS, NULL},
    {"Samlp2SubjectQueryAbstract_subject_get", Samlp2SubjectQueryAbstract_subject_get, METH_VARARGS, NULL},
    {"Samlp2SubjectQueryAbstract_subject_set", Samlp2SubjectQueryAbstract_subject_set, METH_VARARGS, NULL},
    {"Samlp2AuthnQuery_requestedAuthnContext_get", Samlp2AuthnQuery_requestedAuthnContext_get, METH_VARARGS, NULL},
    {"Samlp2AuthnQuery_requestedAuthnContext_set", Samlp2AuthnQuery_requestedAuthnContext_set, METH_VARARGS, NULL},
    {"Samlp2AuthnQuery_sessionIndex_get", Samlp2AuthnQuery_sessionIndex_get, METH_VARARGS, NULL},
    {"Samlp2AuthnQuery_sessionIndex_set", Samlp2AuthnQuery_sessionIndex_set, METH_VARARGS, NULL},
    {"Saml2AttributeStatement_attribute_get", Saml2AttributeStatement_attribute_get, METH_VARARGS, NULL},
    {"Saml2AttributeStatement_attribute_set", Saml2AttributeStatement_attribute_set, METH_VARARGS, NULL},
    {"Saml2AttributeStatement_encryptedAttribute_get", Saml2AttributeStatement_encryptedAttribute_get, METH_VARARGS, NULL},
    {"Saml2AttributeStatement_encryptedAttribute_set", Saml2AttributeStatement_encryptedAttribute_set, METH_VARARGS, NULL},
    {"Saml2AuthzDecisionStatement_action_get", Saml2AuthzDecisionStatement_action_get, METH_VARARGS, NULL},
    {"Saml2AuthzDecisionStatement_action_set", Saml2AuthzDecisionStatement_action_set, METH_VARARGS, NULL},
    {"Saml2AuthzDecisionStatement_evidence_get", Saml2AuthzDecisionStatement_evidence_get, METH_VARARGS, NULL},
    {"Saml2AuthzDecisionStatement_evidence_set", Saml2AuthzDecisionStatement_evidence_set, METH_VARARGS, NULL},
    {"Saml2AuthzDecisionStatement_resource_get", Saml2AuthzDecisionStatement_resource_get, METH_VARARGS, NULL},
    {"Saml2AuthzDecisionStatement_resource_set", Saml2AuthzDecisionStatement_resource_set, METH_VARARGS, NULL},
    {"Saml2AuthzDecisionStatement_decision_get", Saml2AuthzDecisionStatement_decision_get, METH_VARARGS, NULL},
    {"Saml2AuthzDecisionStatement_decision_set", Saml2AuthzDecisionStatement_decision_set, METH_VARARGS, NULL},
    {"Saml2AuthnStatement_subjectLocality_get", Saml2AuthnStatement_subjectLocality_get, METH_VARARGS, NULL},
    {"Saml2AuthnStatement_subjectLocality_set", Saml2AuthnStatement_subjectLocality_set, METH_VARARGS, NULL},
    {"Saml2AuthnStatement_authnContext_get", Saml2AuthnStatement_authnContext_get, METH_VARARGS, NULL},
    {"Saml2AuthnStatement_authnContext_set", Saml2AuthnStatement_authnContext_set, METH_VARARGS, NULL},
    {"Saml2AuthnStatement_authnInstant_get", Saml2AuthnStatement_authnInstant_get, METH_VARARGS, NULL},
    {"Saml2AuthnStatement_authnInstant_set", Saml2AuthnStatement_authnInstant_set, METH_VARARGS, NULL},
    {"Saml2AuthnStatement_sessionIndex_get", Saml2AuthnStatement_sessionIndex_get, METH_VARARGS, NULL},
    {"Saml2AuthnStatement_sessionIndex_set", Saml2AuthnStatement_sessionIndex_set, METH_VARARGS, NULL},
    {"Saml2AuthnStatement_sessionNotOnOrAfter_get", Saml2AuthnStatement_sessionNotOnOrAfter_get, METH_VARARGS, NULL},
    {"Saml2AuthnStatement_sessionNotOnOrAfter_set", Saml2AuthnStatement_sessionNotOnOrAfter_set, METH_VARARGS, NULL},
    {"Samlp2Scoping_idpList_get", Samlp2Scoping_idpList_get, METH_VARARGS, NULL},
    {"Samlp2Scoping_idpList_set", Samlp2Scoping_idpList_set, METH_VARARGS, NULL},
    {"Samlp2Scoping_requesterId_get", Samlp2Scoping_requesterId_get, METH_VARARGS, NULL},
    {"Samlp2Scoping_requesterId_set", Samlp2Scoping_requesterId_set, METH_VARARGS, NULL},
    {"Samlp2Scoping_proxyCount_get", Samlp2Scoping_proxyCount_get, METH_VARARGS, NULL},
    {"Samlp2Scoping_proxyCount_set", Samlp2Scoping_proxyCount_set, METH_VARARGS, NULL},
    {"Saml2AttributeValue_any_get", Saml2AttributeValue_any_get, METH_VARARGS, NULL},
    {"Saml2AttributeValue_any_set", Saml2AttributeValue_any_set, METH_VARARGS, NULL},
    {"Samlp2IDPEntry_providerId_get", Samlp2IDPEntry_providerId_get, METH_VARARGS, NULL},
    {"Samlp2IDPEntry_providerId_set", Samlp2IDPEntry_providerId_set, METH_VARARGS, NULL},
    {"Samlp2IDPEntry_name_get", Samlp2IDPEntry_name_get, METH_VARARGS, NULL},
    {"Samlp2IDPEntry_name_set", Samlp2IDPEntry_name_set, METH_VARARGS, NULL},
    {"Samlp2IDPEntry_loc_get", Samlp2IDPEntry_loc_get, METH_VARARGS, NULL},
    {"Samlp2IDPEntry_loc_set", Samlp2IDPEntry_loc_set, METH_VARARGS, NULL},
    {"Samlp2Response_assertion_get", Samlp2Response_assertion_get, METH_VARARGS, NULL},
    {"Samlp2Response_assertion_set", Samlp2Response_assertion_set, METH_VARARGS, NULL},
    {"Samlp2Response_encryptedAssertion_get", Samlp2Response_encryptedAssertion_get, METH_VARARGS, NULL},
    {"Samlp2Response_encryptedAssertion_set", Samlp2Response_encryptedAssertion_set, METH_VARARGS, NULL},
    {"Saml2Conditions_condition_get", Saml2Conditions_condition_get, METH_VARARGS, NULL},
    {"Saml2Conditions_condition_set", Saml2Conditions_condition_set, METH_VARARGS, NULL},
    {"Saml2Conditions_audienceRestriction_get", Saml2Conditions_audienceRestriction_get, METH_VARARGS, NULL},
    {"Saml2Conditions_audienceRestriction_set", Saml2Conditions_audienceRestriction_set, METH_VARARGS, NULL},
    {"Saml2Conditions_oneTimeUse_get", Saml2Conditions_oneTimeUse_get, METH_VARARGS, NULL},
    {"Saml2Conditions_oneTimeUse_set", Saml2Conditions_oneTimeUse_set, METH_VARARGS, NULL},
    {"Saml2Conditions_proxyRestriction_get", Saml2Conditions_proxyRestriction_get, METH_VARARGS, NULL},
    {"Saml2Conditions_proxyRestriction_set", Saml2Conditions_proxyRestriction_set, METH_VARARGS, NULL},
    {"Saml2Conditions_notBefore_get", Saml2Conditions_notBefore_get, METH_VARARGS, NULL},
    {"Saml2Conditions_notBefore_set", Saml2Conditions_notBefore_set, METH_VARARGS, NULL},
    {"Saml2Conditions_notOnOrAfter_get", Saml2Conditions_notOnOrAfter_get, METH_VARARGS, NULL},
    {"Saml2Conditions_notOnOrAfter_set", Saml2Conditions_notOnOrAfter_set, METH_VARARGS, NULL},
    {"Samlp2ArtifactResolve_artifact_get", Samlp2ArtifactResolve_artifact_get, METH_VARARGS, NULL},
    {"Samlp2ArtifactResolve_artifact_set", Samlp2ArtifactResolve_artifact_set, METH_VARARGS, NULL},
    {"Samlp2LogoutRequest_baseId_get", Samlp2LogoutRequest_baseId_get, METH_VARARGS, NULL},
    {"Samlp2LogoutRequest_baseId_set", Samlp2LogoutRequest_baseId_set, METH_VARARGS, NULL},
    {"Samlp2LogoutRequest_nameId_get", Samlp2LogoutRequest_nameId_get, METH_VARARGS, NULL},
    {"Samlp2LogoutRequest_nameId_set", Samlp2LogoutRequest_nameId_set, METH_VARARGS, NULL},
    {"Samlp2LogoutRequest_encryptedId_get", Samlp2LogoutRequest_encryptedId_get, METH_VARARGS, NULL},
    {"Samlp2LogoutRequest_encryptedId_set", Samlp2LogoutRequest_encryptedId_set, METH_VARARGS, NULL},
    {"Samlp2LogoutRequest_sessionIndex_get", Samlp2LogoutRequest_sessionIndex_get, METH_VARARGS, NULL},
    {"Samlp2LogoutRequest_sessionIndex_set", Samlp2LogoutRequest_sessionIndex_set, METH_VARARGS, NULL},
    {"Samlp2LogoutRequest_reason_get", Samlp2LogoutRequest_reason_get, METH_VARARGS, NULL},
    {"Samlp2LogoutRequest_reason_set", Samlp2LogoutRequest_reason_set, METH_VARARGS, NULL},
    {"Samlp2LogoutRequest_notOnOrAfter_get", Samlp2LogoutRequest_notOnOrAfter_get, METH_VARARGS, NULL},
    {"Samlp2LogoutRequest_notOnOrAfter_set", Samlp2LogoutRequest_notOnOrAfter_set, METH_VARARGS, NULL},
    {"samlp2_logout_request_get_session_indexes", samlp2_logout_request_get_session_indexes, METH_VARARGS, NULL},
    {"samlp2_logout_request_set_session_indexes", samlp2_logout_request_set_session_indexes, METH_VARARGS, NULL},
    {"Samlp2AttributeQuery_attribute_get", Samlp2AttributeQuery_attribute_get, METH_VARARGS, NULL},
    {"Samlp2AttributeQuery_attribute_set", Samlp2AttributeQuery_attribute_set, METH_VARARGS, NULL},
    {"Saml2Evidence_assertionIdRef_get", Saml2Evidence_assertionIdRef_get, METH_VARARGS, NULL},
    {"Saml2Evidence_assertionIdRef_set", Saml2Evidence_assertionIdRef_set, METH_VARARGS, NULL},
    {"Saml2Evidence_assertionUriRef_get", Saml2Evidence_assertionUriRef_get, METH_VARARGS, NULL},
    {"Saml2Evidence_assertionUriRef_set", Saml2Evidence_assertionUriRef_set, METH_VARARGS, NULL},
    {"Saml2Evidence_assertion_get", Saml2Evidence_assertion_get, METH_VARARGS, NULL},
    {"Saml2Evidence_assertion_set", Saml2Evidence_assertion_set, METH_VARARGS, NULL},
    {"Saml2Evidence_encryptedAssertion_get", Saml2Evidence_encryptedAssertion_get, METH_VARARGS, NULL},
    {"Saml2Evidence_encryptedAssertion_set", Saml2Evidence_encryptedAssertion_set, METH_VARARGS, NULL},
    {"Saml2NameID_content_get", Saml2NameID_content_get, METH_VARARGS, NULL},
    {"Saml2NameID_content_set", Saml2NameID_content_set, METH_VARARGS, NULL},
    {"Saml2NameID_format_get", Saml2NameID_format_get, METH_VARARGS, NULL},
    {"Saml2NameID_format_set", Saml2NameID_format_set, METH_VARARGS, NULL},
    {"Saml2NameID_spProvidedId_get", Saml2NameID_spProvidedId_get, METH_VARARGS, NULL},
    {"Saml2NameID_spProvidedId_set", Saml2NameID_spProvidedId_set, METH_VARARGS, NULL},
    {"Saml2NameID_nameQualifier_get", Saml2NameID_nameQualifier_get, METH_VARARGS, NULL},
    {"Saml2NameID_nameQualifier_set", Saml2NameID_nameQualifier_set, METH_VARARGS, NULL},
    {"Saml2NameID_spNameQualifier_get", Saml2NameID_spNameQualifier_get, METH_VARARGS, NULL},
    {"Saml2NameID_spNameQualifier_set", Saml2NameID_spNameQualifier_set, METH_VARARGS, NULL},
    {"saml2_name_id_equals", saml2_name_id_equals, METH_VARARGS, NULL},
    {"Samlp2NameIDMappingResponse_nameId_get", Samlp2NameIDMappingResponse_nameId_get, METH_VARARGS, NULL},
    {"Samlp2NameIDMappingResponse_nameId_set", Samlp2NameIDMappingResponse_nameId_set, METH_VARARGS, NULL},
    {"Samlp2NameIDMappingResponse_encryptedId_get", Samlp2NameIDMappingResponse_encryptedId_get, METH_VARARGS, NULL},
    {"Samlp2NameIDMappingResponse_encryptedId_set", Samlp2NameIDMappingResponse_encryptedId_set, METH_VARARGS, NULL},
    {"Samlp2AuthzDecisionQuery_action_get", Samlp2AuthzDecisionQuery_action_get, METH_VARARGS, NULL},
    {"Samlp2AuthzDecisionQuery_action_set", Samlp2AuthzDecisionQuery_action_set, METH_VARARGS, NULL},
    {"Samlp2AuthzDecisionQuery_evidence_get", Samlp2AuthzDecisionQuery_evidence_get, METH_VARARGS, NULL},
    {"Samlp2AuthzDecisionQuery_evidence_set", Samlp2AuthzDecisionQuery_evidence_set, METH_VARARGS, NULL},
    {"Samlp2AuthzDecisionQuery_resource_get", Samlp2AuthzDecisionQuery_resource_get, METH_VARARGS, NULL},
    {"Samlp2AuthzDecisionQuery_resource_set", Samlp2AuthzDecisionQuery_resource_set, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmationData_notBefore_get", Saml2SubjectConfirmationData_notBefore_get, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmationData_notBefore_set", Saml2SubjectConfirmationData_notBefore_set, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmationData_notOnOrAfter_get", Saml2SubjectConfirmationData_notOnOrAfter_get, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmationData_notOnOrAfter_set", Saml2SubjectConfirmationData_notOnOrAfter_set, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmationData_recipient_get", Saml2SubjectConfirmationData_recipient_get, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmationData_recipient_set", Saml2SubjectConfirmationData_recipient_set, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmationData_inResponseTo_get", Saml2SubjectConfirmationData_inResponseTo_get, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmationData_inResponseTo_set", Saml2SubjectConfirmationData_inResponseTo_set, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmationData_address_get", Saml2SubjectConfirmationData_address_get, METH_VARARGS, NULL},
    {"Saml2SubjectConfirmationData_address_set", Saml2SubjectConfirmationData_address_set, METH_VARARGS, NULL},
    {"saml2_key_info_confirmation_data_type_get_key_info", saml2_key_info_confirmation_data_type_get_key_info, METH_VARARGS, NULL},
    {"saml2_key_info_confirmation_data_type_set_key_info", saml2_key_info_confirmation_data_type_set_key_info, METH_VARARGS, NULL},
    {"DsKeyInfo_keyValue_get", DsKeyInfo_keyValue_get, METH_VARARGS, NULL},
    {"DsKeyInfo_keyValue_set", DsKeyInfo_keyValue_set, METH_VARARGS, NULL},
    {"DsRsaKeyValue_modulus_get", DsRsaKeyValue_modulus_get, METH_VARARGS, NULL},
    {"DsRsaKeyValue_modulus_set", DsRsaKeyValue_modulus_set, METH_VARARGS, NULL},
    {"DsRsaKeyValue_exponent_get", DsRsaKeyValue_exponent_get, METH_VARARGS, NULL},
    {"DsRsaKeyValue_exponent_set", DsRsaKeyValue_exponent_set, METH_VARARGS, NULL},
    {"ds_x509_data_get_certificate", ds_x509_data_get_certificate, METH_VARARGS, NULL},
    {"ds_x509_data_set_certificate", ds_x509_data_set_certificate, METH_VARARGS, NULL},
    {"ds_x509_data_get_subject_name", ds_x509_data_get_subject_name, METH_VARARGS, NULL},
    {"ds_x509_data_set_subject_name", ds_x509_data_set_subject_name, METH_VARARGS, NULL},
    {"ds_x509_data_get_crl", ds_x509_data_get_crl, METH_VARARGS, NULL},
    {"ds_x509_data_set_crl", ds_x509_data_set_crl, METH_VARARGS, NULL},
    {"DsKeyValue_rsaKeyValue_get", DsKeyValue_rsaKeyValue_get, METH_VARARGS, NULL},
    {"DsKeyValue_rsaKeyValue_set", DsKeyValue_rsaKeyValue_set, METH_VARARGS, NULL},
    {"ds_key_value_get_x509_data", ds_key_value_get_x509_data, METH_VARARGS, NULL},
    {"ds_key_value_set_x509_data", ds_key_value_set_x509_data, METH_VARARGS, NULL},
    {NULL, NULL, 0, NULL}
};

PyMODINIT_FUNC
init_lasso(void)
{
	PyObject *m, *d;

	if (PyType_Ready(&PyGObjectPtrType) < 0)
		return;

	m = Py_InitModule3("_lasso", lasso_methods, "_lasso wrapper module");
        d = PyModule_GetDict(m);
        register_constants(d);

	lasso_wrapper_key = g_quark_from_static_string("PyLasso::wrapper");

	Py_INCREF(&PyGObjectPtrType);
	PyModule_AddObject(m, "PyGObjectPtr", (PyObject *)&PyGObjectPtrType);
	lasso_init();
	lasso_log_set_handler(G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION | G_LOG_LEVEL_MASK,
			lasso_python_log, NULL);
}


